/* 
 * Copyright (c) 1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 */
/*
 *	Machine independent stub for remote kgdb
 *
 *	These routines implement the remote side of the gdb protocol. See
 *	the file "remote.c" in the gdb sources for details. In fact, most of 
 *	this file is lifted from those sources.
 */

#ifndef KGDB_LAN
#include <mach/boolean.h>
#include <mach/vm_param.h>
#include <mach/vm_prot.h>
#include <sys/reboot.h>
#include <machine/kgdb.h>

#define PBUFSIZ 1024

/*
 *	buffers for input and output
 */
static char obuf[PBUFSIZ];
static char ibuf[PBUFSIZ];

/*
 *	status and debugging flags
 */
boolean_t	kgdb_enabled = FALSE;	/* are we catching traps with kgdb? */
boolean_t	kgdb_debug_init = TRUE;	/* != 0 waits for host at bootstrap */
int		kgdb_debug = 0;		/* > 0 prints command & errors */
					/* > 1 prints packet traffic */
/*
 *	machine independent routines
 */
extern char kgdb_getc();
extern void kgdb_putc();
extern void kgdb_break();
extern void kgdb_flush_cache();
extern void kgdb_fixup_pc();
extern void kgdb_continue();
extern void kgdb_single_step();
extern void kgdb_dump_core();
extern boolean_t kgdb_mem_access();

/*
 *	convert a hex digit to an inteteger
 */
static int
fromhex(a)
	int a;
{
	if (a >= '0' && a <= '9')
		return a - '0';
	else if (a >= 'a' && a <= 'f')
		return a - 'a' + 10;
	else
		printf("Reply contains invalid hex digit `%c'\n", a);
}

/*
 *	convert an integer into a hex digit
 */
static int
tohex(nib)
	int nib;
{
	if (nib < 10)
		return '0' + nib;
	else
		return 'a' + nib - 10;
}

/*
 *	convert a string of hex digits to memory
 */
static boolean_t
hextomem(ptr, addr, size)
	char *ptr;
	char *addr;
	int size;
{
	int i;

	for (i = 0; i < size; i++) {
		if (ptr[0] == 0 || ptr[1] == 0) {
			if (kgdb_debug)
				printf("hextomem : data buffer too short\n");
			return FALSE;
		}
		*addr++ = fromhex(ptr[0]) << 4 | fromhex(ptr[1]);
		ptr += 2;
	}

	return TRUE;
}

/*
 *	convert memory to a string of hex digits
 */
static void
memtohex(addr, ptr, size)
	char *addr;
	char *ptr;
	int size;
{
	int i;

	if (kgdb_debug > 1)
		printf("memtohex: ");
	
	for (i = 0; i < size; i++) {
		if (kgdb_debug > 1)
			printf("%x", addr[i]);
		
		*ptr++ = tohex(addr[i] >> 4 & 0xf);
		*ptr++ = tohex(addr[i] & 0xf);
	}
	if (kgdb_debug > 1)
		printf("\n");
	
	*ptr = 0;
}

/*
 *	convert a hex number into an integer, stop at first non hex character
 */
static char *hextoint(ptr, value)
	char *ptr;
	int *value;
{
	*value = 0;
	while((*ptr >= '0' && *ptr <= '9') || (*ptr >= 'a' && *ptr <= 'f')) {
		*value = *value << 4 | fromhex(*ptr);
		ptr++;
	}

	return ptr;
}

/*
 *	this routine is called by the mux initialization code as soon as the
 *	mux is configured.
 */
kgdb_mux_init()
{
	extern int boothowto;

	if (boothowto & RB_KDB) {
		boothowto &= ~RB_KDB;
		kgdb_enabled = TRUE;

		if (kgdb_debug_init) {
			printf("kgdb waiting...");
			kgdb_break();
			printf(" connected.\n");
		} else
			printf("kgdb enabled\n");
	} else
		kgdb_enabled = FALSE;
}

/*
 *	Send a packet to the remote machine, with error checking. The data of 
 *	the packet is in BUF.  
 */
static void
putpkt(buf)
	char *buf;
{
	unsigned char	csum = 0;
	char 		*p = buf;

	kgdb_putc('$');

	while (*p) {
		kgdb_putc(*p);
		csum += *p++;
	}

	kgdb_putc('#');
	kgdb_putc(tohex(csum >> 4 & 0xf));
	kgdb_putc(tohex(csum & 0xf));

	if (kgdb_debug > 1)
		printf("sent packet : $%s#%2.2x\n", buf, csum);
}


/*
 *	Read a packet from the remote machine, with error checking, and store 
 *	it in BUF.
 */
static void
getpkt(buf)
	char *buf;
{
	char *bp;
	unsigned char csum;
	int c;
	unsigned char c1, c2;
	int count;

	if (kgdb_debug)
		printf("in getpkt\n");

	c = kgdb_getc();

	while (1) {
		csum = 0;
		count = 0;

		while (c != '$')
			c = kgdb_getc();

		bp = buf;
		while (1) {
			c = kgdb_getc();

			if (c == '$')
				break;

			if (c == '#')
				break;

			*bp++ = c;
			csum += c;
			count++;
			if (count > PBUFSIZ - 1)
				break;
		}
		*bp = 0;

		if (c != '#')
			continue;

		c1 = fromhex(kgdb_getc());
		c2 = fromhex(kgdb_getc());

		if ((csum & 0xff) == (c1 << 4) + c2)
			break;
		if (kgdb_debug)
			printf("Bad checksum, sentsum=0x%x,csum=0x%x,buf=%s\n",
			       (c1 << 4) + c2, csum & 0xff, buf);

		kgdb_putc('-');
	}

	kgdb_putc('+');

	if (kgdb_debug > 1)
		printf("Packet received :%s\n", buf);
}

/*
 *	This routine is called from the trap or interrupt handler when
 *	a breakpoint instruction is encountered or a single step operation
 *	completes. The argument is a pointer to a machine dependent 
 *	saved_state structure that was built on the interrupt or kernel stack.
 */
void 
kgdb_trap(type, ssp)
	int	type;
	char	*ssp;
{
	char *		iptr;
	char *		optr;
	vm_offset_t	offset;
	vm_size_t	length;
	vm_size_t	size;
	vm_prot_t	access;
	int		reg;

	/*
	 *	If we are not debugging then return
	 */
	if (!kgdb_enabled)
		return;

	/*
	 *	let the machine independent routine fix up the PC
	 */
	kgdb_fixup_pc(type, ssp);

	/* 
	 *	tell the host that we have stopped, we always signal a SIGTRAP
	 */
	strcpy(obuf, "S05");
	putpkt(obuf);

	/*
	 *	kgdb command loop
	 */
	while (TRUE) {
		getpkt(ibuf);
		iptr = ibuf;
		access = VM_PROT_NONE;

		switch (*iptr++) {
		case '?':
			/*
			 *	return last signal, always SIGTRAP
			 */
			strcpy(obuf, "S05");
			putpkt(obuf);
			break;

		case 'g':
			/*
			 *	send registers to host
			 */
			memtohex(ssp, obuf, REGISTER_BYTES);
			putpkt(obuf);
			break;

		case 'G':
			/*
			 *	read registers from host
			 */
			if (hextomem(iptr, ssp, REGISTER_BYTES))
				strcpy(obuf, "OK");
			else
				strcpy(obuf, "E01");
			putpkt(obuf);
			break;
			
		case 'M':
			access |= VM_PROT_WRITE;
			/*
			 *	fall through...
			 */

		case 'm':
			access |= VM_PROT_READ;
			/*
			 *	read (m) or write (M) memory
			 */
			iptr = hextoint(iptr, &offset);
			if (*iptr++ != ',') {
				if (kgdb_debug)
					printf("missing ','\n");
				strcpy(obuf, "E01");
				putpkt(obuf);
				break;
			}
			iptr = hextoint(iptr, &length);

			if (access & VM_PROT_WRITE)
				if (*iptr++ != ':') {
					if (kgdb_debug)
						printf("missing ':'\n");
					strcpy(obuf, "E01");
					putpkt(obuf);
					break;
				}
			
			if (length <= 0 || length >= PBUFSIZ) {
				if (kgdb_debug)
					printf("bad length %x\n",length);
				strcpy(obuf, "E02");
				putpkt(obuf);
				break;
			}

			optr = obuf;

			/*
			 *	handle this in pages
			 */
			while(length > 0) {
				/*
				 *	get the number of bytes that we can
				 *	access in this page
				 */
				size = trunc_page(offset + PAGE_SIZE) - offset;

				if (kgdb_debug)
					printf("m: %x %x %x\n",
					       offset, length, size);
				
				if (length < size)
					size = length;

				if (!kgdb_mem_access(offset, access)) {
					if (kgdb_debug)
						printf("can't read memory\n");
					strcpy(obuf, "E03");
					access = VM_PROT_NONE;
					break;
				}

				/*
				 *	we can move anything on this page
				 */
				if (access & VM_PROT_WRITE) {
					if (!hextomem(iptr, offset, size)) {
						strcpy(obuf, "E01");
						access = VM_PROT_NONE;
						break;
					}
					iptr += size * 2;
					kgdb_flush_cache(offset, size);
				} else {
					memtohex(offset, optr, size);
					optr += size * 2;
				}
				offset += size;
				length -= size;
			}

			if (access & VM_PROT_WRITE)
				strcpy(obuf, "OK");

			putpkt(obuf);
			break;

		case 'c':
			/*
			 *	Continue
			 */
			if (*iptr == 0) {
				kgdb_continue(type, ssp, (vm_offset_t)0);
				return;
			}

			/*
			 *	there is a start address, get it
			 */
			iptr = hextoint(iptr, &offset);
			
			/*
			 *	see if we can really execute here
			 */
			if (!kgdb_mem_access(offset, VM_PROT_READ | 
					     VM_PROT_EXECUTE)){
				if (kgdb_debug)
					printf("can't execute at %x\n",offset);
				strcpy(obuf, "E03");
				putpkt(obuf);
				break;
			}
			kgdb_continue(type, ssp, offset);
			return;
				
		case 's':
			/*
			 * 	Single step
			 */
			if (*iptr == 0) {
				kgdb_single_step(type, ssp, (vm_offset_t)0);
				return;
			}

			/*
			 *	there is a start address
			 */
			iptr = hextoint(iptr, &offset);

			/*
			 *	see if we can really execute here
			 */
			if (!kgdb_mem_access(offset, VM_PROT_READ | 
					     VM_PROT_EXECUTE)){
				if (kgdb_debug)
					printf("can't execute at %x\n",offset);
				strcpy(obuf, "E03");
				putpkt(obuf);
				break;
			}
			kgdb_single_step(type, ssp, offset);
			return;

		case 'k':
			/*
			 *	Kill the kernel and dump core
			 */
			kgdb_dump_core();
			panic("can't dump core from kgdb");
			return;

		}
	}
}
#endif
