/*
 * Mach Operating System
 * Copyright (c) 1990,1991,1992,1993,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * Copyright (c) 1991,1987 Carnegie Mellon University.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation,
 * and that all advertising materials mentioning features or use of
 * this software display the following acknowledgement: ``This product
 * includes software developed by the Computer Systems Laboratory at
 * the University of Utah.''
 *
 * CARNEGIE MELLON, THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF
 * THIS SOFTWARE IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY
 * OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF
 * THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * Carnegie Mellon requests users of this software to return to
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 *
 * 	Utah $Hdr: pmap.c 1.49 94/12/15$
 *	Author: Mike Hibler, Bob Wheeler, University of Utah CSL, 10/90
 */
 
/*
 *	Manages physical address maps for hppa.
 *
 *	In addition to hardware address maps, this
 *	module is called upon to provide software-use-only
 *	maps which may or may not be stored in the same
 *	form as hardware maps.  These pseudo-maps are
 *	used to store intermediate results from copy
 *	operations to and from address spaces.
 *
 *	Since the information managed by this module is
 *	also stored by the logical address mapping module,
 *	this module may throw away valid virtual-to-physical
 *	mappings at almost any time.  However, invalidations
 *	of virtual-to-physical mappings must be done as
 *	requested.
 *
 *	In order to cope with hardware architectures which
 *	make virtual-to-physical map invalidates expensive,
 *	this module may delay invalidate or reduced protection
 *	operations until such time as they are actually
 *	necessary.  This module is given full information to
 *	when physical maps must be made correct.
 *	
 */

/*
 * CAVAETS:
 *
 *	PAGE_SIZE must equal PARISC_PGBYTES
 *	Needs more work for MP support
 */

#include <kern/thread.h>
#include <mach/vm_attributes.h>
#include <mach/vm_param.h>
#include <mach/vm_prot.h>
#include <vm/pmap.h>

#include <machine/regs.h>
#include <machine/machspl.h>
#include <machine/cpu.h>

#define       page_mask       (PAGE_SIZE-1)

#if defined(EQUIV_HACK) || defined(IO_HACK)
/*
 * We do not maintain mapping structures for equivalently mapped memory,
 * the theory being that we can easily construct TLB entries at miss time.
 * All we need to know is the virtual address and a protection.  The former
 * is part of the exception information, the latter is provided by the
 * following variables:
 *
 *	[ io_end - 0 )			KRW
 *	[ 0 - text_end )		KRX
 *	[ text_end - equiv_end )	KRW
 *
 * Note that this means we cannot multiply-map an address less than equiv_end
 * due to a lack of auxiliary information normally contained in the mapping
 * structure [XXX not entirely true anymore].
 */

extern vm_offset_t	avail_start;
extern vm_offset_t	avail_end;
extern vm_offset_t	virtual_avail;
extern vm_offset_t	virtual_end;
extern vm_offset_t	avail_next;
extern unsigned int	avail_remaining;

long text_start = 0;
long text_end = 0;
#endif
long equiv_end = 0;
long io_end = 0;

#ifdef PMAPDEBUG
int pmdebug = 0;
#define PDB_USER	0x01	/* exported functions */
#define PDB_MAPPING	0x02	/* low-level mapping routines */
#define PDB_ENTER	0x04	/* pmap_enter specifics */
#define PDB_COPY	0x08	/* copy page debugging */
#define PDB_ZERO	0x10	/* zero page debugging */
#define PDB_IO		0x20	/* IO map debugging */
#endif

struct pmap	kernel_pmap_store;
pmap_t		kernel_pmap;
struct zone	*pmap_zone;		/* zone of pmap structures */
boolean_t	pmap_initialized = FALSE;

#ifdef USEALIGNMENT
/*
 * Mask to determine if two VAs naturally align in the data cache:
 *	834/835:	0 (cannot do)
 *	720/730/750:	256k
 *	705/710:	64k
 *      715:            64k
 * Note that this is a mask only for the data cache since we are
 * only attempting to prevent excess cache flushing of RW data.
 */
vm_offset_t	pmap_alignmask;

#define pmap_aligned(va1, va2) \
	(((va1) & pmap_alignmask) == ((va2) & pmap_alignmask))
#endif

/*
 * Virtual-to-physical translations are handled with hashed lists of "mapping"
 * structures.  The hash table is vtop_table.
 */
struct vtop_entry *vtop_table;

/*
 * Physical-to-virtual translations are handled by an inverted page table
 * structure, phys_table.  Multiple mappings of a single page are handled
 * by linking the affected mapping structures.
 */
struct phys_entry *phys_table;

#ifdef	HPT
/*
 * Hashed (Hardware) Page Table, for use as a software cache, or as a
 * hardware accessible cache on machines with hardware TLB walkers.
 */
struct hpt_entry *hpt_table;
int	usehpt = 1;
#endif

/*
 * First and last physical addresses that we maintain any information for.
 * These provide the range of addresses represented in phys_table.
 * Note that they are long as vm_first_phys may be negative.
 */
long	vm_first_phys = 0;
long	vm_last_phys  = 0;

/*
 * Function to get index into phys_table for a given physical address
 */
#define PHYS_INDEX(pa)	(((pa) - (vm_offset_t)vm_first_phys) >> PARISC_PGSHIFT)

#define PHYS_INRANGE(pa) \
	((long)(pa) >= vm_first_phys && (long)(pa) < vm_last_phys)
#define pmap_find_physentry(pa)	\
	(PHYS_INRANGE(pa) ? &phys_table[PHYS_INDEX(pa)] : PHYS_NULL)

/*
 * Convert a TLB ready page number to a physical address and visa versa
 *	PA 1.0:		pn << 4 implies pa >> 7 since pageshift == 11
 *	PA 1.1:		pn << 5 implies pa >> 7 since pageshift == 12
 */
#define tlbtopa(tlbpn)	((vm_offset_t)(tlbpn) << 7)
#define patotlb(pa)	((vm_offset_t)(pa) >> 7)

/*
 * XXX use mpqueue_head_t's??
 * (no one else does, are they coming or going?)
 */
queue_head_t		free_mapping;	   /* list of free mapping structs */
simple_lock_data_t	free_mapping_lock; /* and lock */

pmap_t			free_pmap;	   /* list of free pmaps */
simple_lock_data_t	free_pmap_lock;	   /* and lock */

extern void pmap_enter(pmap_t pmap, vm_offset_t va, vm_offset_t pa,
		       vm_prot_t prot, boolean_t wired);

static struct mapping *pmap_enter_mapping(), *pmap_find_mapping();
static void pmap_free_mapping(), pmap_clear_mappings(), pmap_reap_mappings();
static void pmap_stash_init();
static void pmap_add_pmap(pmap_t pmap);
static void pmap_rem_pmap(pmap_t pmap);
static pmap_t pmap_get_pmap(pa_space_t space);
static void pmap_clear_cache(struct mapping *mp);
static void pmap_flush_writers(struct phys_entry *pp);

struct mapping *firstmap, *lastmap;	/* XXX for post-mortems */
int		pages_per_vm_page;
unsigned	kern_prot[8], user_prot[8];
int		pid_counter;

#define	pmap_sid(pmap, va) \
	(((va & 0xc0000000) != 0xc0000000) ? pmap->space : PARISC_SID_KERNEL)

#define pmap_prot(pmap, prot)	\
	(((pmap) == kernel_pmap ? kern_prot : user_prot)[prot])

/*
 * parisc_protection_init()
 *	Initialize the user/kern_prot_codes[] arrays which are 
 *	used to translate machine independent protection codes
 *	to HP-PA protection codes.
 */
static void
parisc_protection_init()
{
	register unsigned *kp, *up;
	register int prot;

	kp = kern_prot;
	up = user_prot;
	for (prot = 0; prot < 8; prot++) {
		switch (prot) {
		case VM_PROT_NONE | VM_PROT_NONE | VM_PROT_NONE:
			*kp++ = TLB_AR_NA;
			*up++ = TLB_AR_NA;
			break;
		case VM_PROT_READ | VM_PROT_NONE | VM_PROT_NONE:
#if 1
			*kp++ = TLB_AR_KRX;
			*up++ = TLB_AR_URX;
#else
			*kp++ = TLB_AR_KR;
			*up++ = TLB_AR_UR;
#endif
			break;
		case VM_PROT_READ | VM_PROT_NONE | VM_PROT_EXECUTE:
		case VM_PROT_NONE | VM_PROT_NONE | VM_PROT_EXECUTE:
			*kp++ = TLB_AR_KRX;
			*up++ = TLB_AR_URX;
			break;
		case VM_PROT_NONE | VM_PROT_WRITE | VM_PROT_NONE:
		case VM_PROT_READ | VM_PROT_WRITE | VM_PROT_NONE:
#if 1
			*kp++ = TLB_AR_KRWX;
			*up++ = TLB_AR_URWX;
#else
			*kp++ = TLB_AR_KRW;
			*up++ = TLB_AR_URW;
#endif
			break;
		case VM_PROT_NONE | VM_PROT_WRITE | VM_PROT_EXECUTE:
		case VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE:
			*kp++ = TLB_AR_KRWX;
			*up++ = TLB_AR_URWX;
			break;
		}
	}
}

/*
 * pmap_map(va, spa, epa, prot)
 *	is called during boot to map memory in the kernel's address map.
 *	A virtual address range starting at "va" is mapped to the physical
 *	address range "spa" to "epa" with machine independent protection
 *	"prot".
 *
 *	"va", "spa", and "epa" are byte addresses and must be on machine
 *	indpendent page boundaries.
 */
vm_offset_t
pmap_map(va, spa, epa, prot)
	vm_offset_t va;
	vm_offset_t spa;
	vm_offset_t epa;
	vm_prot_t prot;
{
#ifndef IO_HACK
	boolean_t wrapped = FALSE;

	if (spa == epa)
		return(va);
	if (epa == 0) {
		wrapped = TRUE;
		epa = trunc_page(~0);
	}
#endif
	while (spa < epa) {
#ifdef EQUIV_HACK
		if ((long)va >= (long)vm_first_phys && (long)va < equiv_end) {
			register struct phys_entry *pp;
			
			if (va != spa)
				panic("pmap_map: non-equivalent address");
			/*
			 * equivalently mapped, just add protection to page
			 */
			pp = pmap_find_physentry(spa);
#ifndef IO_HACK
			if (wrapped)
			pp->tlbprot = pmap_prot(kernel_pmap, prot) | TLB_DIRTY;
			else
#endif
			pp->tlbprot =
			    pmap_prot(kernel_pmap, prot) | kernel_pmap->pid;
		} else
#endif
			pmap_enter(kernel_pmap, va, spa, prot, TRUE);

		va += PAGE_SIZE;
		spa += PAGE_SIZE;
	}
#ifndef IO_HACK
	if (wrapped) {
		if ((long)va >= (long)vm_first_phys && (long)va < equiv_end) {
			register struct phys_entry *pp;
			
			if (va != spa)
				panic("pmap_map: non-equivalent address");
			/*
			 * equivalently mapped, just add protection to page
			 */
			pp = pmap_find_physentry(spa);
#if 1
			if (1)
			pp->tlbprot = pmap_prot(kernel_pmap, prot) | TLB_DIRTY;
			else
#endif
			pp->tlbprot =
			    pmap_prot(kernel_pmap, prot) | kernel_pmap->pid;
		} else
			pmap_enter(kernel_pmap, va, spa, prot, TRUE);
		va += PAGE_SIZE;
	}
#endif
	return(va);
}

#ifdef	BTLB
/*
 * pmap_block_map(pa, size, prot, entry, dtlb)
 *    Block map a physical region. Size must be a power of 2. Address must
 *    must be aligned to the size. Entry is the block TLB entry to use.
 *
 *    S-CHIP: Entries 0,2 and 1,3 must be the same size for each TLB type. 
 */
void
pmap_block_map(pa, size, prot, entry, tlbflag)
	vm_offset_t pa;
	vm_size_t size;
	vm_prot_t prot;
	int entry, tlbflag;
{
	unsigned tlbprot = pmap_prot(kernel_pmap, prot) | kernel_pmap->pid;
	
	switch (tlbflag) {
		
	case BLK_ICACHE:
  		insert_block_itlb(entry, pa, size, tlbprot);
		break;
	case BLK_DCACHE:
		insert_block_dtlb(entry, pa, size, tlbprot);
		break;
	case BLK_COMBINED:
		insert_block_ctlb(entry, pa, size, tlbprot);
		break;
	case BLK_LCOMBINED:
		insert_L_block_ctlb(entry, pa, size, tlbprot);
		break;
	default:
		printf("pmap_block_map routine: unknown flag %d\n",tlbflag);
	}
}
#endif

/*
 *	Bootstrap the system enough to run with virtual memory.
 *	Map the kernel's code and data, and allocate the system page table.
 *	Called with mapping OFF.  Page_size must already be set.
 *
 *	Parameters:
 *	avail_start	PA of first available physical page
 *	avail_end	PA of last available physical page
 */
void
pmap_bootstrap(avail_start, avail_end)
	vm_offset_t *avail_start, *avail_end;
{
	register struct mapping *mp;
	vm_offset_t addr, eaddr, naddr;
	vm_size_t size, asize;
	int i, num;
	extern int etext;
	extern unsigned int io_size;
#ifdef USEALIGNMENT
	extern int cputype, dcache_size;
#endif

	*avail_start = round_page(*avail_start);
	*avail_end = trunc_page(*avail_end);

#ifdef IO_HACK
	vm_first_phys = 0;
#else
	vm_first_phys = (long) trunc_page(io_size);
#endif
	vm_last_phys = (long) *avail_end;

	pages_per_vm_page = PAGE_SIZE / PARISC_PGBYTES;
	/* XXX for now */
	if (pages_per_vm_page != 1)
		panic("HP800 page != MACH page");

	parisc_protection_init();
	pmap_stash_init();

	/*
	 * Initialize kernel pmap
	 */
	kernel_pmap = &kernel_pmap_store;
#if	NCPUS > 1
	lock_init(&pmap_lock, FALSE);
#endif	NCPUS > 1
	simple_lock_init(&kernel_pmap->lock);
	kernel_pmap->ref_count = 1;
	kernel_pmap->space = PARISC_SID_KERNEL;
	kernel_pmap->pid = PARISC_PID_KERNEL;
	kernel_pmap->next = PMAP_NULL;

	/*
	 * Allocate various tables and structures.
	 */
	addr = *avail_start;

	/*
	 * Allocate the physical to virtual table.
	 */
	addr = cache_align(addr);
	phys_table = (struct phys_entry *) addr;

	num = atop(vm_last_phys - vm_first_phys);
	for (i = 0; i < num; i++)
		queue_init(&phys_table[i].phys_link);

	addr = (vm_offset_t) &phys_table[num];

	/*
	 * load cr24 with the address of the phys_table entry for virtual
	 * address 0.
	 */
	mtctl(CR_PTOV, (unsigned) &phys_table[PHYS_INDEX((vm_offset_t) 0)]);
	
	/*
	 * Allocate the virtual to physical table.
	 */
	addr = cache_align(addr);
	vtop_table = (struct vtop_entry *) addr;

	for (i = 0; i < PARISC_HASHSIZE; i++)
		queue_init(&vtop_table[i].hash_link);
	addr = (vm_offset_t) &vtop_table[PARISC_HASHSIZE];

	/*
	 * load cr25 with the address of the hash table
	 */
	mtctl(CR_VTOP, (unsigned) vtop_table);

	/*
	 * Allocate at least the required number of mapping entries.
	 * (XXX can't use a zone since zone package hasn't been initialized yet)
	 * We use any excess space (e.g. as required for alignment of the HPT)
	 * for mapping structures as well.
	 */
	addr = cache_align(addr);
	size = sizeof(struct mapping) * (PARISC_MIN_MPP * num);
#ifdef HPT
	/*
	 * HW requires HPT to be sligned to its size, so allocate mapping
	 * structs up to that boundary.
	 */
	if (usehpt)
		asize = sizeof(struct hpt_entry) * PARISC_HASHSIZE;
	else
#endif
	asize = PAGE_SIZE;
	size = (((addr + size + asize - 1) / asize) * asize) - addr;
	num  = size / sizeof(struct mapping);

	mp = (struct mapping *) addr;

	queue_init(&free_mapping);
	simple_lock_init(&free_mapping_lock);

	firstmap = mp;
	while (num--) {
		queue_enter(&free_mapping, mp, struct mapping *, hash_link);
		mp++;
	}
	lastmap = mp;
	addr = round_page((vm_offset_t)mp);

#ifdef	HPT
	/*
	 * Allocate the HPT.
	 * We set the software pointer in each entry to the equivalent entry
	 * in the VTOP table. This is for the benefit of the TLB handlers,
	 * which would have to compute the new address anyway on a HPT
	 * cache miss. It means reading a word from memory, but I think its
	 * a wash, since the HPT entry is going to be in the data cache.
	 */
	size = sizeof(struct hpt_entry) * PARISC_HASHSIZE;
	if (usehpt && (addr & (size - 1)))
		panic("HPT not aligned correctly!");

	hpt_table = (struct hpt_entry *) addr;
	for (i = 0; i < PARISC_HASHSIZE; i++) {
		hpt_table[i].valid      = 0;
		hpt_table[i].vpn        = 0;
		hpt_table[i].space      = -1;
		hpt_table[i].tlbpage    = 0;
		hpt_table[i].tlbprot    = 0;
		hpt_table[i].vtop_entry = (unsigned) &vtop_table[i];
	}
	addr = round_page((vm_offset_t)&hpt_table[PARISC_HASHSIZE]);
	
	/*
	 * load cr25 with the address of the HPT table
	 *
	 * NB: It sez CR_VTOP, but we (and the TLB handlers) know better ...
	 */
	mtctl(CR_VTOP, (unsigned) hpt_table);
#endif

	*avail_start = addr;
	avail_remaining = atop(*avail_end - *avail_start);
	avail_next = *avail_start;

#ifdef USEALIGNMENT
	/*
	 * If we can take advantage of natural alignments in the cache
	 * set that up now.
	 */
	switch (cputype) {
	case HP_705:
	case HP_710:
	case HP_720:
	case HP_730:
	case HP_750:
		pmap_alignmask = dcache_size - 1;
                disable_S_sid_hashing();   /* S-CHIP specific routine */
		break;
        case HP_715_33:
        case HP_715_50:
        case HP_715T_50:
        case HP_715S_50:
        case HP_715_75:
	case HP_725:
	case HP_735:
	case HP_735_125:	/* XXX is this right? */
	case HP_755:
		pmap_alignmask = dcache_size - 1;
		disable_T_sid_hashing();   /* T-CHIP specific routine */
		break;
	case HP_712_60:
	case HP_712_80:
	case HP_712_3:
	case HP_712_4:
		pmap_alignmask = dcache_size - 1;
		disable_L_sid_hashing();   /* L-CHIP specific routine */
		break;
	default:
		pmap_alignmask = 0;
		break;
	}
#endif

        /*
	 * Load the kernel's PID
	 */
	mtctl(CR_PIDR1, PARISC_PID_KERNEL);
}

/*
 * pmap_init(spa, epa)
 *	finishes the initialization of the pmap module.
 *	This procedure is called from vm_mem_init() in vm/vm_init.c
 *	to initialize any remaining data structures that the pmap module
 *	needs to map virtual memory (VM is already ON).
 */
/*ARGSUSED*/
void
pmap_init()
{
	vm_size_t s;

	s = sizeof(struct pmap);
	pmap_zone = zinit(s, 400*s, 4096, 0, "pmap");	/* XXX */

	pid_counter = PARISC_PID_KERNEL + 2;

	pmap_initialized = TRUE;
}

unsigned int pmap_free_pages()
{
	return avail_remaining;
}

boolean_t pmap_next_page(addrp)
	vm_offset_t *addrp;
{
	if (avail_next == avail_end)
		return FALSE;

	*addrp = avail_next;
	avail_next += PAGE_SIZE;
	avail_remaining--;
	return TRUE;
}

void pmap_virtual_space(startp, endp)
	vm_offset_t *startp;
	vm_offset_t *endp;
{
	*startp = virtual_avail;
	*endp = virtual_end;
}

/*
 * pmap_create
 *
 * Create and return a physical map.
 *
 * If the size specified for the map is zero, the map is an actual physical
 * map, and may be referenced by the hardware.
 *
 * If the size specified is non-zero, the map will be used in software 
 * only, and is bounded by that size.
 */
pmap_t
pmap_create(vm_size_t size)
{
	pmap_t pmap;
	int s, pid;
	pa_space_t sid;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_create(size=%x)%c", size, size ? '\n' : ' ');
#endif

	/*
	 * A software use-only map doesn't even need a pmap structure.
	 */
	if (size)
		return(PMAP_NULL);

	/* 
	 * If there is a pmap in the pmap free list, reuse it. 
	 */
	s = splvm();
	simple_lock(&free_pmap_lock);
	if (free_pmap != PMAP_NULL) {
		pmap = free_pmap;
		free_pmap = pmap->next;
	} else
		pmap = PMAP_NULL;
	simple_unlock(&free_pmap_lock);
	splx(s);

	/*
	 * Couldn't find a pmap on the free list, try to allocate a new one.
	 */
	if (pmap == PMAP_NULL) {
		pmap = (pmap_t) zalloc(pmap_zone);
		if (pmap == PMAP_NULL)
			return(PMAP_NULL);

		/*
		 * Allocate space and protection IDs for the pmap.
		 * If all are allocated, there is nothing we can do.
		 */
		s = splvm();
		simple_lock(&sid_pid_lock);
		if (pid_counter <= MAX_PID) {
			pid = pid_counter;
			pid_counter += 2;
		} else
			pid = 0;
		simple_unlock(&sid_pid_lock);
		splx(s);
		if (pid == 0)
			return(PMAP_NULL);

		/* 
		 * Initialize the sids and pid. This is a user pmap, so 
		 * sr4, sr5, and sr6 are identical. sr7 is always KERNEL_SID.
		 */
		pmap->pid = pid;
		simple_lock_init(&pmap->lock);
	}
	pmap->space = pmap->pid;
	pmap->ref_count = 1;
	pmap->next = PMAP_NULL;
	pmap->stats.resident_count = 0;
	pmap->stats.wired_count = 0;
	pmap_add_pmap(pmap);
#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("-> %x\n", pmap);
#endif
	return(pmap);
}

#ifdef	SAS
/*
 * pmap_spawn
 *
 * Create and return a physical map that shares the space id of the provided
 * pmap, but has a new protection id.
 *
 */
pmap_t
pmap_spawn(parent_pmap)
	pmap_t parent_pmap;
{
	pmap_t pmap;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_spawn(pmap=0x%x)\n", parent_pmap);
#endif
	pmap = pmap_create(0);
	pmap->space = parent_pmap->space;

	return(pmap);
}
#endif

/* 
 * pmap_destroy
 * 
 * Gives up a reference to the specified pmap.  When the reference count 
 * reaches zero the pmap structure is added to the pmap free list.
 *
 * Should only be called if the map contains no valid mappings.
 */
void
pmap_destroy(pmap)
	pmap_t pmap;
{
	int ref_count;
	int s;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_destroy(pmap=%x)\n", pmap);
#endif

	if (pmap == PMAP_NULL)
		return;

	s = splvm();
	simple_lock(&pmap->lock);
	ref_count = --pmap->ref_count;
	simple_unlock(&pmap->lock);
	splx(s);

	if (ref_count < 0)
		panic("pmap_destroy(): ref_count < 0");
	if (ref_count > 0)
		return;

	pmap_rem_pmap(pmap);
	/* 
	 * Add the pmap to the pmap free list. 
	 */
	simple_lock(&free_pmap_lock);
	pmap->next = free_pmap;
	free_pmap = pmap;
	simple_unlock(&free_pmap_lock);
}

/*
 * pmap_reference(pmap)
 *	gains a reference to the specified pmap.
 */
void
pmap_reference(pmap)
	pmap_t pmap;
{
	int s;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_reference(pmap=%x)\n", pmap);
#endif

	if (pmap != PMAP_NULL) {
		s = splvm();
		simple_lock(&pmap->lock);
		pmap->ref_count++;
		simple_unlock(&pmap->lock);
		splx(s);
	}
}

/*
 * pmap_remove(pmap, s, e)
 *	unmaps all virtual addresses v in the virtual address
 *	range determined by [s, e) and pmap.
 *	s and e must be on machine independent page boundaries and
 *	s must be less than or equal to e.
 */
void
pmap_remove(pmap_t pmap, vm_offset_t sva, vm_offset_t eva)
{
	register vm_offset_t qeva;
	pa_space_t space;
	struct mapping *mp;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_remove(pmap=%x, sva=%x, eva=%x)\n",
		       pmap, sva, eva);
#endif

	if (pmap == PMAP_NULL)
		return;

	while (sva < eva) {
		space = pmap_sid(pmap, sva);
		if (space == PARISC_SID_KERNEL)
			qeva = parisc_trunc_page(~0);
		else
			qeva = parisc_round_quad(sva+PAGE_SIZE);
		if (eva < qeva)
			qeva = eva;
		while (sva < qeva) {
			if (mp = pmap_find_mapping(space, sva)) {
				pmap_free_mapping(mp);
				pmap->stats.resident_count--;
			}
			sva += PAGE_SIZE;
		}
	}
}

/*
 * pmap_remove_all(phys)
 *	removes the machine independent physical page specified by phys from
 *	all the pmaps in which it resides.
 *	Reflects back modify bits to the pager.
 */
static void
pmap_remove_all(pa)
	vm_offset_t pa;
{
	register struct phys_entry *pp;
	register struct mapping *mp;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_remove_all(pa=%x)\n", pa);
#endif

	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return;

	while (!queue_empty(&pp->phys_link)) {
		mp = (struct mapping *) queue_first(&pp->phys_link);
		((pmap_t)pmap_get_pmap(mp->space))->stats.resident_count--;
		pmap_free_mapping(mp);
	}
}

/*
 *	Routine:
 *		pmap_page_protect
 *
 *	Function:
 *		Lower the permission for all mappings to a given page.
 */
void
pmap_page_protect(vm_offset_t pa, vm_prot_t prot)
{
	register struct phys_entry *pp;
	register struct mapping *mp;
	unsigned tlbprot;
#ifdef DOTLBENTER
	boolean_t waswriter;
#endif

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_page_protect(pa=%x, prot=%x)\n", pa, prot);
#endif

	switch (prot) {
	case VM_PROT_READ:
	case VM_PROT_READ|VM_PROT_EXECUTE:
		break;
	case VM_PROT_ALL:
		return;
	default:
		pmap_remove_all(pa);
		return;
	}

	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return;

#ifdef DOTLBENTER
	/*
	 * Write back any dirty cache.
	 * XXX is this strictly necessary?
	 */
#ifdef USEALIGNMENT
	if (pp->tlbprot & TLB_ALIGNED) {
		pmap_flush_writers(pp);
		/*
		 * Still need to change TLB entries since TLB_ALIGNED does
		 * not imply only one valid mapping as the writer field
		 * does and pmap_flush_writers only invalidates mappings
		 * for dirty pages in the dtlb.
		 */
		waswriter = FALSE;
	} else
#endif
	if (mp = pp->writer) {
#ifdef DEBUG
		if ((mp->tlbprot & TLB_DIRTY) == 0)
			panic("protect: writer w/o TLB_DIRTY");
		if ((mp->tlbprot & TLB_DCACHE) == 0)
			panic("protect: writer w/o TLB_DCACHE");
#endif
		fdcache(mp->space, mp->offset, PAGE_SIZE);
		pdtlb(mp->space, mp->offset);
		mp->tlbprot &= ~(TLB_DIRTY|TLB_DCACHE);
#ifdef HPT
		pmap_clear_cache(mp);
#endif
		pp->writer = MAPPING_NULL;
		waswriter = TRUE;
	} else
		waswriter = FALSE;
#endif

	/*
	 * Modify mappings if necessary.  If waswriter is true then there is
	 * no need to modify the actual TLBs since the writer would have the
	 * only valid mapping in the TLB and it has already been flushed above.
	 */
	queue_iterate(&pp->phys_link, mp, struct mapping *, phys_link) {
		/*
		 * Compare new protection with old to see if
		 * anything needs to be changed.
		 */
#ifndef	FOOEY
		tlbprot = ((mp->space == 0) ? kern_prot : user_prot)[prot];
#else		
		tlbprot = pmap_prot(mp->pmap, prot);
#endif
		if ((mp->tlbprot & TLB_AR_MASK) != tlbprot) {
			mp->tlbprot &= ~TLB_AR_MASK;
			mp->tlbprot |= tlbprot;
#ifdef HPT
			pmap_clear_cache(mp);
#endif
#ifdef DOTLBENTER
			if (!waswriter) {
				iitlbp(mp->tlbprot, mp->space, mp->offset);
				idtlbp(mp->tlbprot, mp->space, mp->offset);
			}
#else
			/*
			 * Purge the current TLB entry (if any) to force
			 * a fault and reload with the new protection.
			 */
			pdtlb(mp->space, mp->offset);
			pitlb(mp->space, mp->offset);
#endif
		}
	}
}

static boolean_t
pmap_verify_free(pa)
	vm_offset_t pa;
{
	struct phys_entry *pp;
	boolean_t result;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_verify_free(pa=%x)\n", pa);
#endif

	if (!pmap_initialized)
		return(TRUE);

	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return(FALSE);

	return(queue_empty(&pp->phys_link));
}

/*
 * pmap_protect(pmap, s, e, prot)
 *	changes the protection on all virtual addresses v in the 
 *	virtual address range determined by [s, e) and pmap to prot.
 *	s and e must be on machine independent page boundaries and
 *	s must be less than or equal to e.
 */
void
pmap_protect(pmap_t pmap, vm_offset_t sva, vm_offset_t eva, vm_prot_t prot)
{
	register struct mapping *mp;
	register vm_offset_t qeva;
	unsigned tlbprot;
	pa_space_t space;
#ifdef DOTLBENTER
	struct phys_entry *pp;
#endif

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_protect(pmap=%x, sva=%x, eva=%x, prot=%x)\n",
		       pmap, sva, eva, prot);
#endif

	if (pmap == PMAP_NULL)
		return;

	if (prot == VM_PROT_NONE) {
		pmap_remove(pmap, sva, eva);
		return;
	}
	if (prot & VM_PROT_WRITE)
		return;

	while (sva < eva) {
		space = pmap_sid(pmap, sva);
		if (space == PARISC_SID_KERNEL)
			qeva = parisc_trunc_page(~0);
		else
			qeva = parisc_round_quad(sva+PAGE_SIZE);
		if (eva < qeva)
			qeva = eva;

		for ( ; sva < qeva; sva += PAGE_SIZE) {
			mp = pmap_find_mapping(space, sva);
			if (mp == MAPPING_NULL)
				continue;
#ifdef DEBUGFOO
			if (pmap != mp->pmap)
				panic("protect: pmap mismatch");
#endif
			/*
			 * Determine if mapping is changing.
			 * If not, nothing to do.
			 */
			tlbprot = pmap_prot(pmap, prot);
			if ((mp->tlbprot & TLB_AR_MASK) == tlbprot)
				continue;

			mp->tlbprot &= ~TLB_AR_MASK;
			mp->tlbprot |= tlbprot;
#ifdef HPT
			pmap_clear_cache(mp);
#endif
#ifdef DOTLBENTER
			/*
			 * If we are the writer for a page we write back
			 * the cache and purge the TLB, otherwise modify
			 * the TLB protection.
			 */
			pp = pmap_find_physentry(tlbtopa(mp->tlbpage));
#ifdef USEALIGNMENT
			if (pp->tlbprot & TLB_ALIGNED) {
				pmap_flush_writers(pp);
				iitlbp(mp->tlbprot, mp->space, mp->offset);
				idtlbp(mp->tlbprot, mp->space, mp->offset);
			} else
#endif
			if (mp->tlbprot & TLB_DIRTY) {
#ifdef DEBUG
				if (pp->writer != mp)
					panic("protect: TLB_DIRTY not writer");
				if ((mp->tlbprot & TLB_DCACHE) == 0)
					panic("protect: writer w/o TLB_DCACHE");
#endif
				fdcache(mp->space, mp->offset, PAGE_SIZE);
				pdtlb(mp->space, mp->offset);
				mp->tlbprot &= ~(TLB_DIRTY|TLB_DCACHE);
#ifdef HPT
				pmap_clear_cache(mp);
#endif
				pp->writer = MAPPING_NULL;
			} else {
				iitlbp(mp->tlbprot, mp->space, mp->offset);
				idtlbp(mp->tlbprot, mp->space, mp->offset);
			}
#else
			/*
			 * Purge the current TLB entry (if any) to force
			 * a fault and reload with the new protection.
			 */
			pdtlb(mp->space, mp->offset);
			pitlb(mp->space, mp->offset);
#endif
		}
	}

}

/*
 * pmap_enter
 *
 * Create a translation for the virtual address (virt) to the physical address
 * (phys) in the pmap with the protection requested. If the translation is 
 * wired then we can not allow a page fault to occur for this mapping.
 */
void
pmap_enter(pmap_t pmap, vm_offset_t va, vm_offset_t pa, vm_prot_t prot, boolean_t wired)
{
	register struct mapping *mp;
	struct phys_entry *pp;
	unsigned tlbpage, tlbprot;
	pa_space_t space;
	boolean_t inst, waswired;
#ifdef DOTLBENTER
	boolean_t dotlbenter;
#endif

#ifdef PMAPDEBUG
	if (pmdebug & (PDB_USER|PDB_ENTER))
		printf("pmap_enter(pmap=%x, va=%x, pa=%x, prot=%x, wire=%x)\n",
		       pmap, va, pa, prot, wired);
#endif

        if (pmap == PMAP_NULL)
                return;

#ifdef DEBUG
#ifdef IO_HACK
	/*
	 * We should never see enters on IO pages after pmap_init has been
	 * called.  If we do, then our assumption that such addresses are
	 * never multiply-mapped is bogus.
	 */
	if ((long)pa >= io_end && (long)pa < 0) {
		if (pmap_initialized)
			panic("enter: called with IO page");
		return;
	}
#endif
#ifdef EQUIV_HACK
	/*
	 * We should never see enters on equivalently mapped pages.
	 * If we do, then our assumption that such addresses are never
	 * multiply-mapped is bogus.
	 */
	if ((long)pa >= 0 && (long)pa < equiv_end)
		panic("enter: called with equiv page");
#endif
#endif

	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return;

	/*
	 * Align page # for easy insertion in TLB.
	 */
	tlbpage = patotlb(pa);

#ifdef DOTLBENTER
	/*
	 * We don't really know if this entry is a mapping for data or 
	 * instructions.  We'll assume that if it has execute permission then
	 * it's for instuctions.  If were wrong then we'll take a TLB miss and 
	 * fix it then.
	 */
	inst = prot & VM_PROT_EXECUTE;
	dotlbenter = FALSE;
#endif

	space = pmap_sid(pmap, va);
	mp = pmap_find_mapping(space, va);
	
	/*
	 * Mapping exists but for a different physical page.
	 * Clean up the old mapping first to simplify matters later.
	 */
	if (mp && mp->tlbpage != tlbpage) {
		pmap_free_mapping(mp);
		pmap->stats.resident_count--;
		mp = MAPPING_NULL;
	}
#ifdef USEALIGNMENT
	/*
	 * Determine if the new mapping changes the cache alignment status
	 * of this physical page.
	 */
	if (pmap_alignmask && mp == MAPPING_NULL) {
		if (queue_empty(&pp->phys_link)) {
			pp->tlbprot |= TLB_ALIGNED;
		} else {
			register struct mapping *amp;

			if (pp->tlbprot & TLB_ALIGNED) {
#ifdef DEBUG
				if (pp->writer)
					panic("enter: aligned with writer");
#endif
				amp = (struct mapping *)
					queue_first(&pp->phys_link);
				if (!pmap_aligned(va, amp->offset)) {
					pmap_flush_writers(pp);
					pp->tlbprot &= ~TLB_ALIGNED;
				}
			}
		}
	}
#endif

#ifdef DOTLBENTER
	/*
	 * If we will be entering a new address in a TLB we need to flush
	 * any writer mapping for the physical page to maintain invarient
	 * that writer mapping will be the only one in the TLB.
	 */
	if (mp == MAPPING_NULL && pp->writer) {
		register struct mapping *wmp = pp->writer;

#ifdef DEBUG
		if ((wmp->tlbprot & TLB_DIRTY) == 0)
			panic("enter: writer without TLB_DIRTY");
		if ((wmp->tlbprot & TLB_DCACHE) == 0)
			panic("enter: TLB_DIRTY w/o TLB_DCACHE");
#endif
		fdcache(wmp->space, wmp->offset, PAGE_SIZE);
		pdtlb(wmp->space, wmp->offset);
		wmp->tlbprot &= ~(TLB_DIRTY|TLB_DCACHE);
#ifdef HPT
		pmap_clear_cache(wmp);
#endif
		pp->writer = MAPPING_NULL;
	}
#endif

	if (mp == MAPPING_NULL) {
#ifdef	TLB_ZERO
		/*
		 * See if we need to zero the page, and do it now that
		 * we have a virtual address to use. 
		 */
		if (pp->tlbprot & TLB_ZERO) {
			lpage_zero(tlbpage, space, va);
			pp->tlbprot &= ~TLB_ZERO;
		}
#endif
		/*
		 * Mapping for this virtual address doesn't exist.
		 * Get a mapping structure and fill it in.
		 */
		mp = pmap_enter_mapping(pmap, space, va, tlbpage, 0);
#ifdef DOTLBENTER
		/*
		 * Insert address into the appropriate TLB.
		 */
		if (inst) {
			iitlba(tlbpage, space, va);
			mp->tlbprot |= (TLB_ICACHE);
		} else {
			idtlba(tlbpage, space, va);
			mp->tlbprot |= (TLB_DCACHE);
		}
		dotlbenter = TRUE;
#endif
	}
#ifndef DOTLBENTER
	else {
		/*
		 * We are just changing the protection.
		 * Flush the current TLB entry to force a fault and reload.
		 */
		pdtlb(mp->space, mp->offset);
		pitlb(mp->space, mp->offset);
	}
#endif

	/*
	 * Determine the protection information for this mapping.
	 *
	 * WARNING: It looks like there is a big race condition here.
	 *          We enter the mapping above with no prot field, and
	 *          then set it here. Another TLB miss could come in
	 *          the meantime, and get a bogus mp->tlbprot value.
	 */
	tlbprot = pmap_prot(pmap, prot) | pmap->pid;
	tlbprot |= (mp->tlbprot & ~(TLB_AR_MASK|TLB_PID_MASK));
#ifdef DOTLBENTER
	/*
	 * Change the TLB entry if it exists.
	 */
	if (dotlbenter) {
		if (inst)
			iitlbp(tlbprot, space, va);
		else
			idtlbp(tlbprot, space, va);
		/*
		 * Treat the entry as if it has been referenced under the
		 * assumption that we are entering it with that intent.
		 */
		pp->tlbprot |= TLB_REF;
	}
#endif

	/*
	 * Add in software bits and adjust statistics
	 */
	waswired = tlbprot & TLB_WIRED;
	if (wired && !waswired) {
		tlbprot |= TLB_WIRED;
		pmap->stats.wired_count++;
	} else if (!wired && waswired) {
		tlbprot &= ~TLB_WIRED;
		pmap->stats.wired_count--;
	}
	mp->tlbprot = tlbprot;
#ifdef HPT
	pmap_clear_cache(mp);
#endif
	pmap->stats.resident_count++;
}

/*
 *	Routine:	pmap_change_wiring
 *	Function:	Change the wiring attribute for a map/virtual-address
 *			pair.
 *	In/out conditions:
 *			The mapping must already exist in the pmap.
 *
 * Change the wiring for a given virtual page. This routine currently is 
 * only used to unwire pages and hence the mapping entry will exist.
 */
void
pmap_change_wiring(pmap, va, wired)
	register pmap_t	pmap;
	vm_offset_t	va;
	boolean_t	wired;
{
	register struct mapping *mp;
	boolean_t waswired;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_change_wiring(pmap=%x, va=%x, wire=%x)\n",
		       pmap, va, wired);
#endif

	if (pmap == PMAP_NULL)
		return;

	if ((mp = pmap_find_mapping(pmap_sid(pmap, va), va)) == MAPPING_NULL)
		panic("pmap_change_wiring: can't find mapping entry");

	waswired = mp->tlbprot & TLB_WIRED;
	if (wired && !waswired) {
		mp->tlbprot |= TLB_WIRED;
		pmap->stats.wired_count++;
	} else if (!wired && waswired) {
		mp->tlbprot &= ~TLB_WIRED;
		pmap->stats.wired_count--;
	}
}

/*
 * pmap_extract(pmap, va)
 *	returns the physical address corrsponding to the 
 *	virtual address specified by pmap and va if the
 *	virtual address is mapped and 0 if it is not.
 */
vm_offset_t
pmap_extract(pmap, va)
	pmap_t pmap;
	vm_offset_t va;
{
	struct mapping *mp;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_extract(pmap=%x, va=%x)\n", pmap, va);
#endif

	/*
	 * Handle easy special cases
	 */
	if (pmap == kernel_pmap) {
#if defined(EQUIV_HACK)
		if (va < (vm_offset_t)equiv_end)
			return(va);
#endif
		if ((long)va >= io_end && (long)va < 0)
			return(va);
	}
	mp = pmap_find_mapping(pmap_sid(pmap, va), va);
	if (mp == MAPPING_NULL)
		return(0);
	return(tlbtopa(mp->tlbpage) + (va & page_mask));
}

/*
 *	pmap_attributes:
 *
 *	Set/Get special memory attributes
 *
 */
kern_return_t
pmap_attribute(pmap, address, size, attribute, value)
	pmap_t		pmap;
	vm_offset_t	address;
	vm_size_t	size;
	vm_machine_attribute_t	attribute;
	vm_machine_attribute_val_t* value;		/* IN/OUT */
{
	vm_offset_t 	sva, eva;
	kern_return_t		ret;

	if (attribute != MATTR_CACHE)
		return KERN_INVALID_ARGUMENT;

	if (pmap == PMAP_NULL)
		return KERN_SUCCESS;

	sva = trunc_page(address);
	eva = round_page(address + size);
	ret = KERN_SUCCESS;

	simple_lock(&pmap->lock);

	switch (*value) {

	case MATTR_VAL_CACHE_FLUSH:	/* flush from all caches */
	case MATTR_VAL_DCACHE_FLUSH:	/* flush from data cache(s) */
	case MATTR_VAL_ICACHE_FLUSH:	/* flush from instruction cache(s) */
	{
		int op;

		/* which cache should we flush */
		op = 0;
		if (*value == MATTR_VAL_CACHE_FLUSH ||
		    *value == MATTR_VAL_ICACHE_FLUSH)
			op = 1;
		if (*value == MATTR_VAL_CACHE_FLUSH ||
		    *value == MATTR_VAL_DCACHE_FLUSH)
			op |= 2;

		for (; sva < eva; sva += PAGE_SIZE) {
			if (pmap_extract(pmap, sva) == 0)
				continue;
			if (op & 1)
				ficache(pmap_sid(pmap, sva), sva, PAGE_SIZE);
			if (op & 2)
				fdcache(pmap_sid(pmap, sva), sva, PAGE_SIZE);
		}
		break;
	}

	case MATTR_VAL_OFF:		/* (generic) turn attribute off */
	case MATTR_VAL_ON:		/* (generic) turn attribute on */
	case MATTR_VAL_GET:		/* (generic) return current value */
	default:
		ret = KERN_INVALID_ARGUMENT;
		break;

	}

	simple_unlock(&pmap->lock);

	return ret;
}

/*
 * pmap_collect
 * 
 * Garbage collects the physical map system for pages that are no longer used.
 */
/* ARGSUSED */
void
pmap_collect(pmap)
	pmap_t pmap;
{
#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_collect(pmap=%x)\n", pmap);
#endif
}

/*
 *	Routine:	pmap_activate
 *	Function:
 *		Binds the given physical map to the given
 *		processor, and returns a hardware map description.
 */
/* ARGSUSED */
static void
pmap_activate(pmap, th, which_cpu)
	pmap_t pmap;
	thread_t th;
	int which_cpu;
{
#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_activate(pmap=%x, th=%x, cpu=%x)\n",
		       pmap, th, which_cpu);
#endif
}

/* ARGSUSED */
static void
pmap_deactivate(pmap, th, which_cpu)
	pmap_t pmap;
	thread_t th;
	int which_cpu;
{
#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_activate(pmap=%x, th=%x, cpu=%x)\n",
		       pmap, th, which_cpu);
#endif
}

/*
 * pmap_zero_page(pa)
 *
 * pmap_zero_page zeros the specified page. 
 */
void
pmap_zero_page(pa)
	vm_offset_t pa;
{
	register struct mapping *mp;
	register struct phys_entry *pp;

#ifdef PMAPDEBUG
	if (pmdebug & (PDB_USER|PDB_ZERO))
		printf("pmap_zero_page(pa=%x)\n", pa);
#endif
#ifdef DEBUG
	/*
	 * XXX can these happen?
	 */
	if (!PHYS_INRANGE(pa))
		panic("zero_page: physaddr out of range");
	if ((long)pa < 0)
		panic("zero_page: IO page");
#endif
#ifdef EQUIV_HACK
	if (pa < equiv_end) {
		printf("pmap_zero_page: pa(%x)\n", pa);
		panic("Call for Help!");
	}
#endif

#ifdef	TLB_ZERO
	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL) {
		phys_bzero(pa, PAGE_SIZE);
		return;
	}

	/*
	 * If there are no mappings for this page, then just mark this page
	 * to be zeroed later, when pmap_enter is called, and there is a
	 * virtual address to use. Using virtual addresses means that we
	 * do not have to flush the cache (the physical addresses). This
	 * shortens the time to zero the page, and it results in a warm cache.
	 * Otherwise, do a slower physical page zero.
	 */
	if (queue_empty(&pp->phys_link)) {
		pp->tlbprot |= TLB_ZERO;
	}
	else {
		pmap_clear_mappings(pa, MAPPING_NULL);
		phys_bzero(pa, PAGE_SIZE);
	}
#else
	pmap_clear_mappings(pa, MAPPING_NULL);
	phys_bzero(pa, PAGE_SIZE);
#endif
}

/*
 * pmap_copy_page(src, dst)
 *
 * pmap_copy_page copies the src page to the destination page. If a mapping
 * can be found for the source, we use that virtual address. Otherwise, a
 * slower physical page copy must be done. The destination is always a
 * physical address since there is usually no mapping for it.
 */
void
pmap_copy_page(spa, dpa)
        vm_offset_t spa, dpa;
{
	register struct phys_entry *pp;
	vm_offset_t soffset;
	pa_space_t sspace;
	struct mapping *mp;
	boolean_t realmode = FALSE;

#ifdef PMAPDEBUG
	if (pmdebug & (PDB_USER|PDB_COPY))
		printf("pmap_copy_page(spa=%x, dpa=%x)\n", spa, dpa);

	/*
	 * XXX can these happen?
	 */
	if (!PHYS_INRANGE(spa) || !PHYS_INRANGE(dpa))
		panic("copy_page: physaddr out of range");
	if ((long)spa < 0 || (long)dpa < 0)
		panic("copy_page: IO page");
#endif
#ifdef EQUIV_HACK
	if ((long)spa < equiv_end || (long)dpa < equiv_end) {
		printf("pmap_copy_page: spa(%x), dpa(%x)\n", spa, dpa);
		panic("Call for Help!");
	}
#endif
	/*
	 * Determine the best source address to use:
	 *
	 * 1. If currently mapped for a writer use that
	 * 2. If otherwise mapped use that
	 * 3. Else must do real mode copy
	 */
	pp = pmap_find_physentry(spa);
#ifdef	TLB_ZERO
	if (pp->tlbprot & TLB_ZERO) {
#ifdef PMAPDEBUG
		if (pmdebug & PDB_COPY)
			printf("Copy, but no zero %x\n", pp);
#endif
		pp->tlbprot &= ~TLB_ZERO;
	}
#endif
	if (mp = pp->writer) {
		sspace = mp->space;
		soffset = mp->offset;
	} else if (!queue_empty(&pp->phys_link)) {
		mp = (struct mapping *) queue_first(&pp->phys_link);
		sspace = mp->space;
		soffset = mp->offset;
	} else {
		realmode = TRUE;
	}

	pp = pmap_find_physentry(dpa);
#ifdef	TLB_ZERO
	if (pp->tlbprot & TLB_ZERO) {
#ifdef PMAPDEBUG
		if (pmdebug & PDB_COPY)
			printf("Copy, but no zero %x\n", pp);
#endif
		pp->tlbprot &= ~TLB_ZERO;
	}
#endif
	/*
	 * If no mapping for the source, we do a real mode copy. If we
	 * found a mapping for the source, we can copy from that virtual
	 * address to the destination physical address.
	 */
	if (realmode) {
#ifdef PMAPDEBUG
		if (pmdebug & PDB_COPY)
			printf("pmap_copy_page: phys_copy_page(%x, %x)\n",
			       spa, dpa);
#endif
		/*
		 * pmap_clear_mappings will handle flushing the proper
		 * source virtual address. We flush the destination 
		 * also, even though it is probably not mapped.
		 *
		 * XXX should really purge rather than flush any dpa cache.
		 */
		pmap_clear_mappings(spa, MAPPING_NULL);
		pmap_clear_mappings(dpa, MAPPING_NULL);
		phys_page_copy(spa, dpa);
	} else {
#ifdef PMAPDEBUG
		if (pmdebug & PDB_COPY)
			printf("pmap_copy_page: lpage_copy(%x, %x, %x, %x)\n",
			       mp->tlbpage, sspace, soffset, dpa);
#endif
		lpage_copy(mp->tlbpage, sspace, soffset, dpa);
	}
}

/*
 * pmap_pageable(pmap, s, e, pageable)
 *	Make the specified pages (by pmap, offset)
 *	pageable (or not) as requested.
 *
 *	A page which is not pageable may not take
 *	a fault; therefore, its page table entry
 *	must remain valid for the duration.
 *
 *	This routine is merely advisory; pmap_enter()
 *	will specify that these pages are to be wired
 *	down (or not) as appropriate.
 *
 *	(called from vm/vm_fault.c).
 */
/* ARGSUSED */
pmap_pageable(pmap, sva, eva, pageable)
	pmap_t pmap;
	vm_offset_t sva;
	vm_offset_t eva;
	boolean_t pageable;
{
#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_pageable(pmap=%x, sva=%x, eva=%x, page=%x)\n",
		       pmap, sva, eva, pageable);
#endif
}

/*
 * pmap_clear_modify(phys)
 *	clears the hardware modified ("dirty") bit for one
 *	machine independant page starting at the given
 *	physical address.  phys must be aligned on a machine
 *	independant page boundary.
 */
void
pmap_clear_modify(vm_offset_t pa)
{
	register struct phys_entry *pp;
	register struct mapping *mp;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_clear_modify(pa=%x)\n", pa);
#endif

	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL || (pp->tlbprot & TLB_DIRTY) == 0)
		return;

	/*
	 * Flush any writer cache to ensure memory is up to date.
	 */
#ifdef USEALIGNMENT
	if (pp->tlbprot & TLB_ALIGNED)
		pmap_flush_writers(pp);
	else
#endif
	if (mp = pp->writer) {
		fdcache(mp->space, mp->offset, PAGE_SIZE);
		pdtlb(mp->space, mp->offset);
		mp->tlbprot &= ~(TLB_DIRTY|TLB_DCACHE);
#ifdef HPT
		pmap_clear_cache(mp);
#endif
		pp->writer = MAPPING_NULL;
	}
	pp->tlbprot &= ~TLB_DIRTY;
}

/*
 * pmap_is_modified(phys)
 *	returns TRUE iff the given physical page has been modified 
 *	since the last call to pmap_clear_modify().
 */
boolean_t
pmap_is_modified(vm_offset_t pa)
{
	struct phys_entry *pp;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_is_modified(pa=%x)\n", pa);
#endif

	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL || (pp->tlbprot & TLB_DIRTY) == 0)
		return(FALSE);

	return(TRUE);
}

/*
 * pmap_clear_reference(phys)
 *	clears the hardware referenced bit in the given machine
 *	independant physical page.  
 *
 *	Currently, we treat a TLB miss as a reference; i.e. to clear
 *	the reference bit we flush all mappings for pa from the TLBs.
 */
void
pmap_clear_reference(vm_offset_t pa)
{
	register struct phys_entry *pp;
	register struct mapping *mp;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_clear_reference(pa=%x)\n", pa);
#endif

	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return;

	queue_iterate(&pp->phys_link, mp, struct mapping *, phys_link) {
		pitlb(mp->space, mp->offset);
		pdtlb(mp->space, mp->offset);
		mp->tlbprot &= ~(TLB_REF);
#ifdef HPT
		pmap_clear_cache(mp);
#endif
	}
	pp->tlbprot &= ~TLB_REF;
}

/*
 * pmap_is_referenced(phys)
 *	returns TRUE iff the given physical page has been referenced 
 *	since the last call to pmap_clear_reference().
 */
boolean_t
pmap_is_referenced(vm_offset_t pa)
{
	struct phys_entry *pp;
	register struct mapping *mp;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_is_referenced(pa=%x)\n", pa);
#endif

	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return(FALSE);
	if (pp->tlbprot & TLB_REF)
		return(TRUE);
	queue_iterate(&pp->phys_link, mp, struct mapping *, phys_link) {
		if (mp->tlbprot & TLB_REF) {
			pp->tlbprot |= TLB_REF;
			return(TRUE);
		}
	}
	return(FALSE);
}

#ifndef IO_HACK
/*
 * XXX
 *
 * Wretched hack for user IO mapping
 */
boolean_t
pmap_iomap(va)
	vm_offset_t va;
{
	struct phys_entry *pp;
	vm_offset_t eva;
	vm_size_t len = 0x2000000;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_IO)
		printf("pmap_iomap(va=%x): ", va);
#endif
	if ((long)va < io_end || (long)va >= 0)
		return(FALSE);

	pp = pmap_find_physentry(trunc_page((long)va >> 14));
#ifdef PMAPDEBUG
	if (pmdebug & PDB_IO)
		printf("tlbprot=%x\n", pp->tlbprot);
#endif
	if ((pp->tlbprot & TLB_AR_MASK) == TLB_AR_URWX)
		return(TRUE);
	pp->tlbprot &= ~TLB_AR_MASK;
	pp->tlbprot |= TLB_AR_URWX;
	eva = round_page(va + len);
	for (va = trunc_page(va); va < eva; va += PAGE_SIZE) {
		pitlb(PARISC_SID_KERNEL, va);
		pdtlb(PARISC_SID_KERNEL, va);
	}
	return(TRUE);
}

void
pmap_iounmap(va)
	vm_offset_t va;

{
	struct phys_entry *pp;
	vm_offset_t eva;
	vm_size_t len = 0x2000000;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_IO)
		printf("pmap_iounmap(va=%x): ", va);
#endif
	if ((long)va < io_end || (long)va >= 0)
		panic("pmap_iounmap: bad address");

	pp = pmap_find_physentry(trunc_page((long)va >> 14));
#ifdef PMAPDEBUG
	if (pmdebug & PDB_IO)
		printf("tlbprot=%x\n", pp->tlbprot);
#endif
	if ((pp->tlbprot & TLB_AR_MASK) == TLB_AR_URWX) {
		pp->tlbprot &= ~TLB_AR_MASK;
		pp->tlbprot |= TLB_AR_KRWX;
		eva = round_page(va + len);
		for (va = trunc_page(va); va < eva; va += PAGE_SIZE) {
			pitlb(PARISC_SID_KERNEL, va);
			pdtlb(PARISC_SID_KERNEL, va);
		}
	}
}
#endif

/*
 * Auxilliary routines for manipulating mappings
 */

/* static */
static struct mapping *
pmap_enter_mapping(pmap, space, offset, page, prot)
	pmap_t pmap;
	pa_space_t space;
	vm_offset_t offset;
	unsigned page, prot;
{
	register struct mapping *mp, *pmp;
	struct phys_entry *pp;	
	int x, hash;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_MAPPING)
		printf("pmap_enter_mapping(pmap=%x, sp=%x, off=%x, page=%x, prot=%x)",
		       pmap, space, offset, page, prot);
#endif
	x = splvm();
	simple_lock(&free_mapping_lock);
	if (queue_empty(&free_mapping))
		pmap_reap_mappings();
	queue_remove_first(&free_mapping, mp, struct mapping *, hash_link);
	simple_unlock(&free_mapping_lock);
	splx(x);
	if (mp == MAPPING_NULL)
		panic("pmap_enter: can't find a free mapping entry");

	mp->space = space;
	mp->offset = offset;
	mp->tlbpage = page;
	mp->tlbprot = prot;

	page = tlbtopa(page);
	pp = pmap_find_physentry(page);	
	hash = pmap_hash(space, offset);
	x = splvm();
	queue_enter_first(&pp->phys_link,
			  mp, struct mapping *, phys_link);
	queue_enter_first(&vtop_table[hash].hash_link,
			  mp, struct mapping *, hash_link);
#ifdef	USEALIGNMENT
	/*
	 * Now set the TLB_ALIGNED bit for each mapping in the ptov chain
	 * so that the tlb miss handler can check the mapping directly,
	 * instead of looking in the ptov table.
	 */
	queue_iterate(&pp->phys_link, pmp, struct mapping *, phys_link) {
		if (pp->tlbprot & TLB_ALIGNED)
			pmp->tlbprot |= TLB_ALIGNED;
		else
			pmp->tlbprot &= ~TLB_ALIGNED;
#ifdef HPT
		pmap_clear_cache(pmp);
#endif
	}
#endif
#ifdef	HPT
	hpt_table[hash].valid = 0;
	hpt_table[hash].space = -1;
#endif
	splx(x);
#ifdef PMAPDEBUG
	if (pmdebug & PDB_MAPPING)
		printf(" -> %x\n", mp);
#endif
	return(mp);
}

/*
 * pmap_find_mapping(space, offset)
 *	Lookup the virtual address <space,offset> in the mapping "table".
 *	Returns a pointer to the mapping or NULL if none.
 *
 *	XXX what about MP locking?
 */
/* static */
static struct mapping *
pmap_find_mapping(space, offset)
	register pa_space_t space;
	register vm_offset_t offset;
{
	register queue_t qp;
	register struct mapping *mp;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_MAPPING)
		printf("pmap_find_mapping(sp=%x, off=%x) -> ", space, offset);
#endif
	offset = trunc_page(offset);
	qp = &vtop_table[pmap_hash(space, offset)].hash_link;
	queue_iterate(qp, mp, struct mapping *, hash_link) {
		if (mp->space == space && mp->offset == offset) {
#ifdef PMAPDEBUG
			if (pmdebug & PDB_MAPPING)
				printf("%x\n", mp);
#endif
			return(mp);
		}
	}
#ifdef PMAPDEBUG
	if (pmdebug & PDB_MAPPING)
		printf("MAPPING_NULL\n");
#endif
	return(MAPPING_NULL);
}

/* static */
static void
pmap_free_mapping(mp)
	register struct mapping *mp;
{
	register pa_space_t space;
	register vm_offset_t offset;
	struct phys_entry *pp;
	int x, hash;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_MAPPING)
		printf("pmap_free_mapping(mp=%x): tlbprot %x\n",
		       mp, mp->tlbprot);
#endif
	space = mp->space;
	offset = mp->offset;

	/*
	 * First flush caches and TLBs.
	 * Must be done before removing from virtual/physical lists
	 * since cache flush may cause TLB miss.  Note that we also
	 * must flush the I-cache first since the FIC instruction uses
	 * the DTLB.  We also clear the writer pointer if it references
	 * this mapping.  This prevents an interrupt-time TLB miss from
	 * reentering the mapping in the TLB after we have just removed it.
	 *
	 * XXX I think we have a race here.  D-cache flush needs to be
	 * atomic wrt clearing the writer pointer.
	 */
#ifdef DEBUG
	if ((mp->tlbprot & (TLB_DCACHE|TLB_DIRTY)) == TLB_DIRTY)
		panic("free_mapping: 0x%x TLB_DIRTY w/o TLB_DCACHE", mp);
#endif
	pp = pmap_find_physentry(tlbtopa(mp->tlbpage));
	if (pp->writer == mp)
		pp->writer = MAPPING_NULL;
	if (mp->tlbprot & TLB_ICACHE)
		ficache(space, offset, PAGE_SIZE);
	pitlb(space, offset);
	if (mp->tlbprot & TLB_DCACHE)
		fdcache(space, offset, PAGE_SIZE);
	pdtlb(space, offset);

	/*
	 * Now remove from the PTOV/VTOP lists and return to the free list.
	 * Note that we must block interrupts since they might cause TLB
	 * misses which would search (potentially inconsistant) lists.
	 */
	hash = pmap_hash(space, offset);
	x = splvm();
	queue_remove(&vtop_table[hash].hash_link,
		     mp, struct mapping *, hash_link);
	queue_remove(&pp->phys_link, mp, struct mapping *, phys_link);
	simple_lock(&free_mapping_lock);
	queue_enter(&free_mapping, mp, struct mapping *, hash_link);
	simple_unlock(&free_mapping_lock);
#ifdef	HPT
	hpt_table[hash].valid = 0;
	hpt_table[hash].space = -1;
#endif
	splx(x);
#ifdef USEALIGNMENT
	/*
	 * If we were not aligned, check all remaining entries to see if
	 * they are now aligned (i.e. we removed the unaligned one).
	 */
	if (pp->tlbprot & TLB_ALIGNED) {
		if (queue_empty(&pp->phys_link))
			pp->tlbprot &= ~TLB_ALIGNED;
	} else if (!queue_empty(&pp->phys_link)) {
		boolean_t aligned = TRUE;

		mp = (struct mapping *) queue_first(&pp->phys_link);
		offset = mp->offset;
		while (queue_last(&pp->phys_link) != (queue_entry_t)mp) {
			mp = (struct mapping *) queue_next(&mp->phys_link);
			if (!pmap_aligned(offset, mp->offset)) {
				aligned = FALSE;
				break;
			}
		}
		if (aligned) {
			pp->writer = MAPPING_NULL;
			pp->tlbprot |= TLB_ALIGNED;
		}
	}
	/*
	 * Now set/clear the TLB_ALIGNED bit for each mapping in the ptov 
	 * chain so that the tlb miss handler can check the mapping directly,
	 * instead of looking in the ptov table.
	 */
	queue_iterate(&pp->phys_link, mp, struct mapping *, phys_link) {
		if (pp->tlbprot & TLB_ALIGNED)
			mp->tlbprot |= TLB_ALIGNED;
		else
			mp->tlbprot &= ~TLB_ALIGNED;
#ifdef HPT
		pmap_clear_cache(mp);
#endif
	}
#endif
}

/*
 * Flush caches and TLB entries refering to physical page pa.  If cmp is
 * non-zero, we do not affect the cache or TLB entires for that mapping.
 */
/* static */
static void
pmap_clear_mappings(pa, cmp)
	vm_offset_t pa;
	struct mapping *cmp;
{
	register struct phys_entry *pp;
	register struct mapping *mp;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_MAPPING)
		printf("pmap_clear_mappings(pa=%x, cmp=%x)\n", pa, cmp);
#endif
	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return;

	/*
	 * If writer is non-zero we know that there is only one
	 * valid mapping active (i.e. with TLB and/or cache entries).
	 * Hence, there is no need to search the entire list.
	 *
	 * XXX needs to be atomic?
	 */
	if (mp = pp->writer) {
#ifdef DEBUG
		if ((mp->tlbprot & TLB_DIRTY) == 0)
			panic("clear_mappings: writer w/o TLB_DIRTY");
		if ((mp->tlbprot & TLB_DCACHE) == 0)
			panic("clear_mappings: writer w/o TLB_DCACHE");
#endif
		if (mp != cmp) {
			fdcache(mp->space, mp->offset, PAGE_SIZE);
			pdtlb(mp->space, mp->offset);
			mp->tlbprot &= ~(TLB_DIRTY|TLB_DCACHE);
#ifdef HPT
			pmap_clear_cache(mp);
#endif
			pp->writer = MAPPING_NULL;
		}
		return;
	}
	/*
	 * Otherwise we traverse all mappings associated with this address
	 * purging TLBs and flushing caches.
	 */
	queue_iterate(&pp->phys_link, mp, struct mapping *, phys_link) {
		if (mp == cmp)
			continue;
		/*
		 * have to clear the icache first since fic uses the dtlb.
		 */
		if (mp->tlbprot & TLB_ICACHE)
			ficache(mp->space, mp->offset, PAGE_SIZE);
		pitlb(mp->space, mp->offset);
		if (mp->tlbprot & TLB_DCACHE)
			fdcache(mp->space, mp->offset, PAGE_SIZE);
		pdtlb(mp->space, mp->offset);
		mp->tlbprot &= ~(TLB_ICACHE|TLB_DCACHE);
#ifdef USEALIGNMENT
		mp->tlbprot &= ~TLB_DIRTY;
#endif
#ifdef HPT
		pmap_clear_cache(mp);
#endif
	}
}

#ifdef HPT
/*
 * Clear the HPT table entry for the corresponding space/offset to reflect
 * the fact that we have possibly changed the mapping, and need to pick
 * up new values from the mapping structure on the next access.
 */
static void
pmap_clear_cache(mp)
	register struct mapping *mp;
{
	int	hash = pmap_hash(mp->space, mp->offset);

	hpt_table[hash].valid = 0;
	hpt_table[hash].space = -1;
}
#endif

#ifdef USEALIGNMENT
/*
 * Flush all write mappings for a physical page.
 *
 * If all mappings for a page are aligned in the cache we may have
 * multiple writers.  They will all be identified by having TLB_DIRTY
 * set in their tlbprot field.  We flush the TLB entries for all of
 * these to ensure we detect the first attempt to write the page again.
 * Note that since all mappings are aligned, we only need one cache flush
 * and that can use any mapping.
 */
static void
pmap_flush_writers(pp)
	register struct phys_entry *pp;
{
	register struct mapping *mp;
	boolean_t flushed = FALSE;
	int x;

	if (queue_empty(&pp->phys_link))
		return;

#ifdef DEBUG
	if (pp->writer)
		panic("flush_writers: writer set");
#endif
	mp = (struct mapping *) queue_first(&pp->phys_link);
	/*
	 * XXX ugh!  This needs to be atomic to prevent TLB misses in
	 * interrupt routines from seeing inconsistent cache info.
	 */
	x = splvm();
	while (!queue_end(&pp->phys_link, (queue_entry_t)mp)) {
		if (mp->tlbprot & TLB_DIRTY) {
			if (!flushed) {
				fdcache(mp->space, mp->offset, PAGE_SIZE);
				flushed = TRUE;
			}
			pdtlb(mp->space, mp->offset);
			mp->tlbprot &= ~(TLB_DIRTY|TLB_DCACHE);
#ifdef HPT
			pmap_clear_cache(mp);
#endif
		}
		mp = (struct mapping *) queue_next(&mp->phys_link);
	}
	splx(x);
}
#endif

/*
 * Collect un-wired mappings.  How best to do this?
 */
/* static */
static void
pmap_reap_mappings()
{
#ifdef PMAPDEBUG
	if (pmdebug & PDB_MAPPING)
		printf("pmap_reap_mappings()\n");
#endif

	panic("out of mapping structs");
}


/*
 * ltor(space, offset)
 *	returns the physical address corresponding to the virtual address
 *	at space and offset if the address is mapped and 0 if it is not.
 *
 *	This is used by device drivers that don't want to know too much about 
 *	the VM.
 */
vm_offset_t
ltor(space, offset)
	pa_space_t	    space;
	vm_offset_t offset;
{
	struct mapping *mp;

#ifdef PMAPDEBUG
	if (pmdebug & PDB_USER)
		printf("ltor(space=%x, offset=%x)\n", space, offset);
#endif

	if (space == 0) {
#if defined(EQUIV_HACK)
		if (offset < (vm_offset_t)equiv_end)
			return(offset);
#endif
		if ((long)offset >= io_end && (long)offset < 0)
			return(offset);
	}
	mp = pmap_find_mapping(space, offset);
	if (mp == MAPPING_NULL)
		return(0);
	return(tlbtopa(mp->tlbpage) + (offset & page_mask));
}

/*
 *	Find the logarithm base 2 of a value
 */
int 
log2(value)
	unsigned	value;
{
	unsigned	i;

	for (i = 0; i < 32; i++)
		if ((1 << i) >= value)
			break;

	return i;
}

/*
 * Maintain a table of pmap structures so that we can map from a space
 * id to the associated pmap (for the benefit of pmap_remove_all()).
 *
 * NOTE: Originally, there was a pointer in the mapping structure that
 *       could be used to get to the pmap. However, that pointer made
 *	 the mapping structure 4 bytes more than a cache line (32 bytes).
 *	 This is bad, bad, bad. Hence, this seemingly gratuatious code.
 */
#define	PMAP_STASH_SIZE		32
#define pmap_stash_hash(space)  ((space) % PMAP_STASH_SIZE)

queue_head_t pmap_stash[PMAP_STASH_SIZE];

static void
pmap_stash_init()
{
	int 	i;

	for (i = 0; i < PMAP_STASH_SIZE; i++)
		queue_init(&pmap_stash[i]);
}

static void
pmap_add_pmap(pmap)
	pmap_t	pmap;
{
	int 	hash = pmap_stash_hash(pmap->space);
	
	queue_enter_first(&pmap_stash[hash], pmap, pmap_t, pmap_link);
}

static pmap_t
pmap_get_pmap(space)
	pa_space_t	space;
{
	pmap_t	pmap;
	int 	hash = pmap_stash_hash(space);

	if (space == PARISC_SID_KERNEL)
		return(kernel_pmap);
	
	queue_iterate(&pmap_stash[hash], pmap, pmap_t, pmap_link) {
		if (pmap->space == space)
			return(pmap);
	}
	panic("pmap_get_pmap: could not find pmap for space %d\n", space);
}

static void
pmap_rem_pmap(pmap)
	pmap_t	pmap;
{
	int 	hash = pmap_stash_hash(pmap->space);

	queue_remove(&pmap_stash[hash], pmap, pmap_t, pmap_link);
}

#ifdef PMAPDEBUG
/*
 * debugging stuff
 */


/*
 * fake_mapping
 *
 * fake mappings into the low level TLB data structures for testing
 */
static struct mapping *
fake_mapping(space, va, pa, tlbprot)
	pa_space_t     space;
	vm_offset_t va;
	vm_offset_t pa;
	unsigned    tlbprot;
{
	register struct mapping *mp;
	int x;

	/*
	 * assume that mapping for this virtual address doesn't exist.
	 * Get a mapping structure and fill it in.
	 */
	x = splvm();
	simple_lock(&free_mapping_lock);
	if (queue_empty(&free_mapping))
		pmap_reap_mappings();
	queue_remove_first(&free_mapping, mp, struct mapping *, hash_link);
	simple_unlock(&free_mapping_lock);
	splx(x);
	if (mp == MAPPING_NULL)
		panic("fake_mapping: can't find a free mapping entry");

	mp->space   = space;
	mp->offset  = va;
	mp->tlbpage = patotlb(pa);
	mp->tlbprot = tlbprot;

	queue_enter_first(&pmap_find_physentry(pa)->phys_link,
			  mp, struct mapping *, phys_link);
	queue_enter_first(&vtop_table[pmap_hash(space, va)].hash_link,
			  mp, struct mapping *, hash_link);
	return mp;
}


static
chase_queues(mpe)
	struct mapping *mpe;
{
	register queue_t qp;
	register struct mapping *mp;
	register struct phys_entry *pe;

	printf("chasing queues for mapping address = %x\n", mpe);

	if (mpe == (struct mapping *)0)
		return;

	qp = &vtop_table[pmap_hash(mpe->space, mpe->offset)].hash_link;
	printf("dumping vtop queue... hash = %x, entry = %x\n", 
	       pmap_hash(mpe->space, mpe->offset), qp);

	queue_iterate(qp, mp, struct mapping *, hash_link) {
		printf("\n");
		dump_mapping(mp);
	}

	pe = pmap_find_physentry(tlbtopa(mpe->tlbpage));

	printf("physical table...\n");

	dump_ptov(pe);

	qp = &(pmap_find_physentry(tlbtopa(mpe->tlbpage))->phys_link);
	printf("dumping phys_table queue... entry = %x\n", qp);
	queue_iterate(qp, mp, struct mapping *, phys_link) {
		printf("\n");
		dump_mapping(mp);
	}
}

static
dump_ptov(pp)
	struct phys_entry *pp;
{
	printf("ptov address = %x\n", pp);
	printf("   phys_link.next = %x\n", pp->phys_link.next);
	printf("   phys_link.prev = %x\n", pp->phys_link.prev);

	printf("   writer = %x\n", pp->writer);
	printf("   tlbprot = ");
	dump_tlbprot(pp->tlbprot);
	printf("\n\n");
}

static
dump_vtop(qp)
	struct queue_entry *qp;
{
	printf("   vtop.next = %x\n", qp->next);
	printf("   vtop.prev = %x\n", qp->prev);
}


static
dump_mapping(mp)
	struct mapping *mp;
{
	printf("mapping address = %x\n", mp);
	printf("   hash_link.next = %x\n", mp->hash_link.next);
	printf("   hash_link.prev = %x\n", mp->hash_link.prev);
	printf("   phys_link.next = %x\n", mp->phys_link.next);
	printf("   phys_link.prev = %x\n", mp->phys_link.prev);

	printf("   space = %x\n", mp->space);
	printf("   offset = %x\n", mp->offset);

	printf("   tlbpage = %x\n", mp->tlbpage);
	printf("   tlbprot = ");
	dump_tlbprot(mp->tlbprot);
}


static struct {
	unsigned prot;
	char *name;
} prot_names[] = {
	TLB_AR_NA,	"NONE",
	TLB_AR_KR,	"KR",
	TLB_AR_KRW,	"KRW",
	TLB_AR_KRX,	"KRX",
	TLB_AR_KRWX,	"KRWX",
	TLB_AR_UR,	"UR",
	TLB_AR_URW,	"URW",
	TLB_AR_URX,	"URX",
	TLB_AR_URWX,	"URWX"
	};


static
dump_tlbprot(prot)
	int prot;
{
	int ar = prot & TLB_AR_MASK;
	int found = 0;
	int i;

	printf("%c%c%c%c%c",
	       (prot & TLB_REF) ? 'R' : '-',
	       (prot & TLB_ALIGNED) ? 'A' : '-',
	       (prot & TLB_TRAP) ? 'T' : '-',
	       (prot & TLB_DIRTY) ? 'D' : '-',
	       (prot & TLB_BREAK) ? 'B' : '-');

	for (i = 0; i < 9; i++)
		if (ar == prot_names[i].prot) {
			printf(" %s",prot_names[i].name);
			found = 1;
			break;
		}

	if (!found)
		printf(" %x", prot >> 20 & 0x7f);

	printf("%s%s", 
	       (prot & TLB_ICACHE) ? " ICACHE" : "",
	       (prot & TLB_DCACHE) ? " DCACHE" : "");

	printf(" %x", prot >> 1 & 0x7fff);

	if (prot & 1)
		printf(" WIRED");
}

#endif

#include <mach_vm_debug.h>
#if	MACH_VM_DEBUG
int
pmap_list_resident_pages(pmap, listp, space)
	register pmap_t		pmap;
	register vm_offset_t	*listp;
	register int		space;
{
	return 0;
}
#endif	/* MACH_VM_DEBUG */
