/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * Network Interface for Intel i82596 LAN network coprocessor
 */

#define	DEBUG 0
#if DEBUG
int ecdebug;
int ec4all;
int ec4multi;
int ec4me;
int ec4any;
int ec4other;
int ec4who;

#define EC_DEBUG_PRINT		0x01
#define EC_DEBUG_CHECK		0x02
#define EC_DEBUG_PRINTALL	0x04
#define EC_DEBUG_RX		0x08
#define EC_DEBUG_TX		0x10

#define ECDEBUG			((ecdebug & EC_DEBUG_PRINT)!=0)
#define ECDEBUG_CHECK		((ecdebug & EC_DEBUG_CHECK)!=0)
#define ECDEBUG_PRINTALL	((ecdebug & EC_DEBUG_PRINTALL)!=0)
#define ECDEBUG_RX		((ecdebug & EC_DEBUG_RX)!=0)
#define ECDEBUG_TX		((ecdebug & EC_DEBUG_TX)!=0)
#endif

#include "ec.h"
#if NEC >0

#include <mach_kdb.h>

#include "sys/param.h"
#include "sys/mbuf.h"
#include "sys/buf.h"
#include "sys/protosw.h"
#include "sys/socket.h"
#include "sys/syslog.h"
#include "sys/vmmac.h"
#include "sys/ioctl.h"
#include "sys/errno.h"
#include "sys/time.h"
#include "sys/kernel.h"

#include "kern/lock.h"

#include "net/if.h"
#include "net/if_types.h"
#include "net/netisr.h"
#include "net/route.h"

#include "netinet/in.h"
#include "netinet/in_systm.h"
#include "netinet/in_var.h"
#include "netinet/ip.h"
#include "netinet/if_ether.h"

#include "netns/ns.h"
#include "netns/ns_if.h"

#if DEBUG
#define private
#else
#define private static
#endif

#include "hp/io.h"
#include "hp/interrupt.h"
#include "hp/ioconf.h"
#include "hp/handler.h"

#include "machine/led.h"
#include "machine/memory.h"
#include "machine/iodefs.h"
#include "machine/pmap.h"

/* Ethernet controller hardware register */
struct csr {
	volatile int reset;	/* hardware reset (write) */
	volatile int port;	/* LAN port select (write) */
	volatile int ca;	/* channel attention (write) */
};

/*
 RBDOFFSET: dma offset into receive buffer. we dma receive data at this offset
 into the naturally aligned receive buffer so that the fields in the IP header
 are naturally aligned. thus we do not have to tweek the TCP/IP source code to
 deal with unaligned data. since we do not take advantage of trailer packets
 anyway this is an easy solution at no cost.

 NBUF: maximum number of packets we can hold in our receive area (RFA)

 NRFDBUF: each incoming packet uses one receive frame descriptor (RFD)

 NCLPACKET: number of clusters need to contain maximum length ethernet packet

 NRBDBUF: each incoming packet uses several receive buffer descriptors (RBD)

 NTBDMBUF: number of TBDs that fit in a simple MBUF
*/

#define RBDOFFSET	2
#define NBUF		20
#define NRFDBUF		NBUF			/* number of Receive Frame Descriptors */
#define ETHERMAXPKT	(sizeof(struct ether_header) + ETHERMTU)
#define NCLPACKET	(1+ETHERMAXPKT/(MCLBYTES-RBDOFFSET))
#define NRBDBUF		(NBUF*NCLPACKET)	/* number of Receive Buffer Descriptors */
#define NTCBBUF		8			/* number of Transmit Command Blocks */
#define NTBDBUF		8			/* number of Transmit Buffer Descriptors per TCB (at least 2) */

#define INVALID_POINTER		0xffffffff	/* indicates an invalid pointer to i82596 */
#define DCACHE_LINE_SIZE	32		/* maximum size of cacheline we can handle */
typedef u_char			dcache_line_t[DCACHE_LINE_SIZE];

/*
 * BIG FAT NOTICE :
 *	
 * due to the cache policy of PA RISC, several control blocks for the ethernet controller
 * that reside in main memory must be laid out very carefully to prevent overwriting of
 * status bits when issueing commands. the problem is that a control block may have command
 * bits (written by the cpu) and status bits (written by the controller) in adjacent memory
 * locations. since writing to the command bits will write an the entire cache line we must
 * make sure that the status bits do not fall into that same cache line otherwise the write
 * back will possibly overwrite the status bits that may have been updated by the controller.
 * in the type declarations below the cache line boundary is indicated in the comments.
 */
#define ALIGNED(x,y)		(((int) (x) + (y)) & ~((y) - 1))
#define CACHE_ALIGNED(x)	ALIGNED(x, DCACHE_LINE_SIZE)

/*
 * SECOND BIG FAT NOTICE:
 *
 * due to a subtle but significant bug in the i82596 step A1 chip (supposedly to be fixed in step B)
 * significant changes have to be made to the driver to detect and/or prevent trashing of i82596
 * memory structures. it is essential to read the intel erratum sheet that descries this bug, but
 * here is a summary. there is a bug in the microcode that handles one of the DMA engines that
 * is used to write status information back to the memory resident control structures. Under certain
 * timing conditions, it is possible that two consecutive writes to be performed by this DMA channel
 * get collapsed during a one cycle window in end-of-transmit processing and can result in an erroneous
 * write. if the intention is for the first write to write A to ADDR1 and B to ADDR2 than the result
 * of a write collapse can be a write of B to ADDR1, or a write of B to ADDR2+DIS, where DIS is the
 * size of A, or a missing write, or a combination of the first two. in the intel erratum sheet all
 * consecutive two writes are examined and a workaround is given. in case we detect the occurrence of
 * the bug, we will reinitialize the chip and start over. this is acceptable overhead since the occurrence
 * of the big is not frequent. for our driver we have choosen to implement workarounds as described below.
 * 
 * transmit path:
 *
 * the basic effect of the bug on the transmit path is that the write to set the busy bit in the CB
 * can wind up going to the SCB status word (A) instead, or to the low address bits for the forward link
 * pointer in the CB (B) or that a write to the SCB status winds up in the SCB command word (C).
 * in case (A) we would see a value of 0x4000 in the SCB status word. since this implies that the CU is
 * idle, but there is no CNR interrupt (only a FR interrupt) this is a bogus interrupt and we check for
 * this value and throw it away (even if we did process a FR interrupt it would not do any damage).
 * any interrupt that is hidden by the bogus status field will be reposted by the chip for the next time.
 * in case (B) we would wind up going to a wrong address for the next CB which could be disastrous. since
 * the lower word of the link address is forced to 0x4000, we will make sure to allocate all our CBs starting
 * in a page with the lower 16 bits of the PPN set to 0x4000 and will place a NOP CB at the start of that page
 * with the EL bit set. thus if trashing occurs, the CU will wind up executing the NOP CB and not the expected
 * chain of transmit commands. this can be detected by examining the C bit of the NOP CB.
 * in case (C) we are safe because all the possible values of the SCB status that we can expect will not cause
 * a problem when issued as a SCB command word.
 * because of another chip bug, the transmit engine can hang when a late collision occurs during transmit. we
 * must detect this by keeping a timeout on any pending transmit. the old detection trick of issueing resume
 * commands to the SCB is aggrevating this DMA bug, so we have to use timeouts to transmit hang detection.
 * 
 * receive path:
 *
 * since end-of-receive processing has multiple consecutive DMA writes that can be collapsed, there are many
 * scenarios for failure. while it is possible to predict all these cases and add logic to detect them, i opted
 * instead for trying to detect when things did not go as planned. you really need to read intel's document, but
 * by studying all cases, it seems sufficient to track when the completion bit in the RFD does not get set,
 * or gets set by an erroneous write of the low 16-bits of the RBD ptr field. thus the detection scheme is that
 * when we get a FR interrupt AND there is NO rfd with a valid setting of the completion bit, THEN we will look
 * ahead into the next RFD and if that one has its completion bit set, we know that corruption happened in which
 * case we reinitialize. In case we run out of receive buffers, we also reinitialize the chip, to cover cases where
 * the completion bit was set, but other fields were damaged.
 * 
 */
#define I82596CORRUPTION	0x4000			/* value of trash in xmit corruption */
#define I82596RFDVALID(r)	(((r->rfd.command.cacheline.cb_stat) & 0xf) != 0xe)	/* RFD status will never end in 0xe ... */

/*
 * note that we must place some 32-bit words on halfword boundaries due to the above cache
 * line requirements. since the compiler aligns data naturally, we must use the following
 * data type to write a word on a halfword boundary in two halfword writes. this comes in
 * handy since in most cases the halfwords need to swapped anyway to satisfy the 82596.
 */
#ifndef KERNEL_COFF
#pragma hp_align DOMAIN_WORD
#endif
typedef struct {
	short	low;
	short	high;
} swap_t;

/* transmit buffer descriptor (TBD) */
typedef struct tbd {
	boolean_t	eof:1;		/* end-of-file: last buffer */
	unsigned	:1;
	unsigned	count:14;	/* actual count of bytes in buffer */ 
	unsigned	clear:16;
	swap_t		link;		/* swapped next buffer */
	swap_t		buffer;	 	/* swapped where buffer lives */
	char		oddchar[2];	/* hack so that xmit buffers can always be 16-bit aligned */
#if DEBUG
	struct tbd	*va;		/* va of next tbd */
#endif
} tbd_t;

/* CB command codes */
typedef enum {
	cb_cuc_nop,				/* nop */
	cb_cuc_ia_setup,			/* setup individual address */
	cb_cuc_configure,			/* configure */
	cb_cuc_multicast_setup,			/* setup multicast */
	cb_cuc_transmit,			/* transmit */
	cb_cuc_time_domain_reflectometer,	/* line analysis */
	cb_cuc_dump,				/* dump internal status	*/
	cb_cuc_diagnose				/* diagnosis */
} cb_cuc_t;

/* CB header */
typedef union cb_command {
	struct {
		u_char		pad[DCACHE_LINE_SIZE-2];	/* pad to force cache aligment */
		boolean_t	completion:1;			/* set to true by CU after completion */
		boolean_t	busy:1;				/* true while CU is executing a command */
		boolean_t	ok:1;				/* true if completed successfully */
		boolean_t	abnormal:1;			/* true if abnormally terminated due to 'scb_cuc_abort' command */
		boolean_t	fail:1;				/* true if sef-test procedure failed */
		unsigned	:11;				/* padding */

		/***** PA: the cache line boundary must fall here (see 'big fat notice') *****/

		boolean_t	el:1;				/* end of list */
		boolean_t	suspend:1;			/* suspend after completion */
		boolean_t	interrupt:1;			/* interrupt after completion */
		unsigned	:8;
		boolean_t	no_crc:1;			/* disables crc insertion, overrules crc mode of 'cb_cuc_configure' command */
		boolean_t	sf:1;				/* simplified mode (false) as opposed to flexible mode (true) */
		cb_cuc_t	cuc:3;				/* command block control unit command */
		swap_t		link;				/* unaligned and swapped pointer to next header */
	} cmd;
	struct {
		u_char		pad[DCACHE_LINE_SIZE-2];	/* pad to force cache aligment */
		unsigned	:4;
		boolean_t	late_collision:1;
		boolean_t	no_carrier_sense:1;
		boolean_t	lost_clear_to_send:1;
		boolean_t	dma_underrun:1;
		boolean_t	deferred:1;
		boolean_t	heartbeat:1;
		boolean_t	too_many_collisions:1;
		unsigned	:1;
		unsigned	maxcoll:4;
	} xmit;
	struct {
		u_char		pad[DCACHE_LINE_SIZE-2];	/* pad to force cache aligment */
		unsigned	skip:3;
		boolean_t	length_error:1;
		boolean_t	crc_error:1;
		boolean_t	alignment_error:1;
		boolean_t	no_resources:1;
		boolean_t	dma_overrun:1;
		boolean_t	individual_address_match:1;
		boolean_t	no_eop_flag:1;
		unsigned	zeros:5;
		boolean_t	collision_detected:1;
	} recv;
	struct {
		swap_t		link;				/* unaligned and swapped pointer to next tcb */
	} save;
	struct {
		u_char		pad[DCACHE_LINE_SIZE-2];	/* pad to reach status bits */
		short		cb_stat;			/* handle on bits set by controller */

		/***** PA: the cache line boundary must fall here (see 'big fat notice') *****/

		short		cb_cmd;				/* handle on bits set by cpu */
	} cacheline;
} cb_command_t;

/* nop CB */
typedef union {
	struct {
		cb_command_t	command;
	} nop;
	struct {
		dcache_line_t	stat;				/* first cache line with bits set by controller */

		/***** PA: the cache line boundary must fall here (see 'big fat notice') *****/

		dcache_line_t	cmd;				/* second cache line with bits set by cpu */
	} cacheline;
} cb_cuc_nop_t;
#define nop_t cb_cuc_nop_t

/* ethernet address, defined as array for general use */
typedef u_char eaddr_t[6];

/* individual address setup CB */
typedef struct {
	cb_command_t	command;
	eaddr_t	address;	
	short	pad;
} cb_cuc_ia_setup_t;

/*configure CB */
typedef struct {
	cb_command_t	command;
	union {
		struct {
			/* byte 8 */
			boolean_t	prefetched:1;		/* enable to write the prefetched bit in all prefetch RBDs */
			unsigned	:3;
			unsigned	byte_cnt:4;		/* number of bytes, including this one, that hold parameters to be configured */
	
			/* byte 9 */
			unsigned	monitor0:2;		/* receive monitor options */
			unsigned	:2;
			unsigned	fifo_limit:4;		/* FIFO limit */
	
			/* byte 10 */
			boolean_t	sav_bf:1;		/* true to save bad frames in memory */
			unsigned	:7;
	
			/* byte 11 */
			unsigned	loopback_mode:2;	/* loopback mode */
			unsigned	preamble_lenght:2;	/* preamble length */
			boolean_t	no_src_addr_ins:1;	/* true for no source address insertion */
			unsigned	addr_len:3;		/* address length */
	
			/* byte 12 */
			boolean_t	bof_metd:1;		/* true for exponential backoff method */
			unsigned	exp_prio:3;		/* exponential priority */
			unsigned	:1;
			unsigned	lin_prio:3;		/* linear priority */
	
			/* byte 13 */
			u_char		interframe_spacing;	/* interframe spacing */
	
			/* byte 14 */
			u_char		slot_time_lo;		/* slot time, low byte */
	
			/* byte 15 */
			unsigned	max_retry_num:4;	/* maximum number of retries */
			unsigned	:1;
			unsigned	slot_time_hi:3;		/* slot time, high part */
	
			/* byte 16 */
			boolean_t	padding:1;		/* padding requested */
			boolean_t	bit_stuffing:1;		/* bit stuffing */
			boolean_t	crc_16:1;		/* 16 (true) or 32 (false) bit crc */
			boolean_t	no_crc_insertion:1;	/* no crc insertion */
			boolean_t	xmit_on_no_crs:1;	/* transmit on no carrier sense */
			boolean_t	manchester_nrz:1;	/* manchester or NRZ encoding */
			boolean_t	broadcast_disable:1;	/* broadcast disable */
			boolean_t	promiscious_mode:1;	/* promiscuous mode */
	
			/* byte 17 */
			boolean_t	cdt_src:1;		/* collision detect source */
			unsigned	cdt_filter:3;		/* collision detect filter length */
			boolean_t	crs_src:1;		/* carrier sense source */
			unsigned	crs_filter:3;		/* carrier sense filter length */
	
			/* byte 18 */
			u_char		min_frame_len;		/* minimum frame length */
	
			/* byte 19 */
			unsigned	monitor1:2;		/* receive monitor options */
			boolean_t	mc_all:1;		/* enable to receive all MC frames */
			boolean_t	coll_detect_bsac:1;	/* collision detect by source address recognition */
			boolean_t	auto_retransmit:1;	/* auto-retransmit */
			boolean_t	crc_in_memory:1;	/* receive crc appended to the frame in memory */
			boolean_t	length_field:1;		/* length field - enables padding at the end-of-carrier framing */
			boolean_t	precrs:1;		/* preamble until carrier sense */
	
			/* byte 20 */
			boolean_t	dcr:1;			/* deterministic collision resolution */
			boolean_t	fdx:1;			/* full duplex resolution */
			unsigned	dcr_slot_address:6;	/* station index in dcr mode */
	
			/* byte 21 */
			boolean_t	disable_backoff:1;	/* disable the backoff algorithm */
			boolean_t	multiple_ia:1;		/* multiple individual address */
			unsigned	dcr_no_of_stations:6;	/* number of stations in dcr mode */
	
			/* byte 22 and 23 must be zero */
			short		clear;
		} field;
		u_char bytes[24];
	} cfg;
} cb_cuc_configure_t;

/* multicast setup CB */
#define NMULTICAST	20				/* chip accepts many more (14 bit number), driver keeps this limit */
typedef struct {
	cb_command_t	command;
	unsigned	:2;
	unsigned	cnt:14;				/* number of bytes in multicast list */
	eaddr_t		mc[NMULTICAST];			/* multicast list */
} cb_cuc_multicast_setup_t;

/* transmit CB (also known as TCB) */
typedef union tcb {
	struct {
		cb_command_t	command;
		swap_t		tbd_ptr;		/* unaligned and swapped pointer to transmit buffer descriptor */
		boolean_t	eof:1;			/* true if whole frame is kept in transmit CB */
		unsigned	:1;
		unsigned	tcb_count:14;		/* TCB byte count to be transmitted */
		unsigned	:16;

		/* defined by us for s/w use */
                unsigned        :16;
		struct mbuf	*m;			/* pointer to MBUF chain we are transmitting */
		union tcb	*next;			/* va of next TCB */
	} tcb;
	struct {
		dcache_line_t	stat;			/* first cache line with bits set by controller */

		/***** PA: the cache line boundary must fall here (see 'big fat notice') *****/

		dcache_line_t	cmd;			/* second cache line with bits set by cpu */
	} cacheline;
} cb_cuc_transmit_t;
#define tcb_t cb_cuc_transmit_t

/* time domain reflectometer CB */
typedef struct {
	cb_command_t	command;
	boolean_t	link_ok:1;
	boolean_t	transceiver_cable:1;
	boolean_t	ethernet_open:1;
	boolean_t	ethernet_short:1;
	boolean_t	:1;
	unsigned	time:11;
	unsigned	clear:16;
} cb_cuc_time_domain_reflectometer_t;

/* dump CB */
typedef struct {
	cb_command_t	command;
	/* you need it, you code it */
} cb_cuc_dump_t;

/* diagnose CB */
typedef struct {
	cb_command_t	command;
	/* you need it, you code it */
} cb_cuc_diagnose_t;

/* command block (CB) */
typedef union {
	cb_cuc_nop_t				nop;
	cb_cuc_ia_setup_t			ia_setup;
	cb_cuc_configure_t			configure;
	cb_cuc_multicast_setup_t		multicast_setup;
	cb_cuc_transmit_t			transmit;
	cb_cuc_time_domain_reflectometer_t	time_domain_reflectometer;
	cb_cuc_dump_t				dump;
	cb_cuc_diagnose_t			diagnose;
} cb_t;

/* receive buffer descriptor (RBD), padded to take up two cachelines */
typedef union rbd {
	struct {
		u_char		pad[DCACHE_LINE_SIZE-2];
		boolean_t	eof:1;			/* end-of-file: last buffer related to frame */
		boolean_t	filled:1;		/* filled by adapter */
		unsigned	count:14;		/* actual count of bytes in buffer */

		/***** PA: the cache line boundary must fall here (see 'big fat notice') *****/

		unsigned	:16;			/* s/b zeros */
		swap_t		link;			/* unaligned and swapped pointer to next RBD */
		swap_t		buffer;			/* unaligned and swapped pointer to buffer space */
		boolean_t	el:1;			/* end-of-list: last buffer in FBL */
		boolean_t	prefetched:1;		/* data in this rbd already prefetched */
		unsigned	size:14;		/* actual size of buffer (s/b even) */
		unsigned	:16;			/* s/b zeros */

		/* defined by us for s/w use, not defined by h/w */
		unsigned	:16;			/* s/b zeros */                
		union rbd	*next;		 	/* va of next RFD */
		struct mbuf	*m;			/* mbuf holding our cluster */
#if DEBUG
		swap_t		pa;			/* swapped PA of this rbd */
#endif
	} rbd;
	struct {
		u_char		pad1[DCACHE_LINE_SIZE-2];
		short		stat;			/* flags and count */
	} cmd;
	struct {
		swap_t		link;			/* unaligned and swapped pointer to next tcb */
	} save;
	struct {
		dcache_line_t	stat;		 	/* first cache line with bits set by controller */

		/***** PA: the cache line boundary must fall here (see 'big fat notice') *****/

		dcache_line_t	cmd;			/* second cache line with bits set by cpu */
	} cacheline;
} rbd_t;

/* receive frame descriptor (RFD), padded to take up two cachelines */
typedef union rfd {
	struct {
		cb_command_t	command;
		swap_t		rbd_ptr;	/* unaligned and swapped pointer to receive buffer descriptor */
		boolean_t	eof:1;		 /* end-of-file: last buffer */
		boolean_t	filled:1;	/* filled by adapter */
		unsigned	count:14;	/* actual count of bytes in buffer */
		unsigned	:2;
		unsigned	size:14;	/* size of buffer (must be even) */
		eaddr_t		da;		/* destination address */
		eaddr_t		sa;		/* source address */
		short		length;		/* user-defined - length in 802.3 */

		/* defined by us for s/w use, not defined by h/w */
		union rfd	*next;		/* va of next RFD */
	} rfd;
	struct {
		dcache_line_t	stat;		/* first cache line with bits set by controller */

		/***** PA: the cache line boundary must fall here (see 'big fat notice') *****/

		dcache_line_t	cmd;		/* second cache line with bits set by cpu */
	} cacheline;
} rfd_t;

/* SCB interrupt bits */
typedef int scb_int_t;
#define SCB_CX	0x8 
#define SCB_FR	0x4
#define SCB_CNA 0x2
#define SCB_RNR 0x1

/* CU command codes */
typedef enum {
	scb_cuc_nop,
	scb_cuc_start,
	scb_cuc_resume,
	scb_cuc_suspend,
	scb_cuc_abort,
	scb_cuc_throttle,
	scb_cuc_throttle_immediate,
	scb_cuc_reserved
} scb_cuc_t;

/* RU command codes */
typedef enum {
	scb_ruc_nop,
	scb_ruc_start,
	scb_ruc_resume,
	scb_ruc_suspend,
	scb_ruc_abort,
	scb_ruc_reserved0,
	scb_ruc_reserved1,
	scb_ruc_reserved2
} scb_ruc_t;

/* SCB command word */
typedef union {
	struct {
		scb_int_t	ack:4;		/* misc command bits */
		unsigned	:1;
		scb_cuc_t	cuc:3;		/* command unit commands */
		boolean_t	reset:1;	/* reset */
		scb_ruc_t	ruc:3;		/* receive unit commands */
		unsigned	:4;
	} cmd;
	short all;
} scb_command_t;

/* CU status codes */
typedef enum {
	scb_cus_idle,
	scb_cus_suspended,
	scb_cus_active,
	scb_cus_not_used0,
	scb_cus_not_used1,
	scb_cus_not_used2,
	scb_cus_not_used3,
	scb_cus_not_used4
} scb_cus_t;

/* RU status codes */
typedef enum {
	scb_rus_idle,			/* 0 */
	scb_rus_suspended,		/* 1 */
	scb_rus_no_resources,		/* 2 */
	scb_rus_not_used0,
	scb_rus_ready,			/* 4 */
	scb_rus_not_used1,
	scb_rus_not_used2,
	scb_rus_not_used3,
	scb_rus_no_rbd,			/* 8 */
	scb_rus_not_used4,
	scb_rus_no_rbd_resources,	/* A */
	scb_rus_not_used5,
	scb_rus_not_used6,
	scb_rus_not_used7,
	scb_rus_not_used8,
	scb_rus_not_used9
} scb_rus_t;

/* SCB status word */
typedef struct {
	scb_int_t	stat:4;		/* misc status bits */
	unsigned	:1;
	scb_cus_t	cus:3;		/* command unit status */
	scb_rus_t	rus:4;		/* receive unit status */
	boolean_t	throad:1;	/* bus throttle timers loaded */
	unsigned	:3;
} scb_status_t;

/* system control block (SCB), padded to take up 3 cachelines */
typedef union {
	struct {
		u_char		pad[DCACHE_LINE_SIZE-2];
		scb_status_t	status;	 		/* status word */
	

                /***** PA: the cache line boundary must fall here (see 'big fat notice') *****/
	
		scb_command_t	command;		/* command word */
		swap_t		cbl_ptr;		/* unaligned and swapped pointer to the command block list */
		swap_t		rfa_ptr;		/* unaligned and swapped pointer to the receive frame area */
		swap_t		crcerrs;		/* swapped number of frames discarded due to crc errors */
		swap_t		alnerrs;		/* swapped number of frames that are both misaligned AND contain crc errors */
		swap_t		rscerrs;		/* swapped number of frames that are discarded solely for lack of resources */
		swap_t		ovrnerrs;		/* swapped number of frames that are lost because local bus was not available */
		swap_t		rcvcdt;			/* swapped number of collisions detected during frame reception */
		swap_t		shrtfrm;		/* swapped number of frames received that are shorter that minimum frame length */
		unsigned short	toff;			/* min time to stay off the bus */
		unsigned short	ton;			/* max time to stay on	the bus */
	} scb;
	struct {
		dcache_line_t	stat;			/* first cache line with bits set by controller */

		/***** PA: the cache line boundary must fall here (see 'big fat notice') *****/

		dcache_line_t	cmd;			/* second cache line with bits set by cpu/controller */
		dcache_line_t	pad;			/* third cache line with more bits */
	} cacheline;
} scb_t;

/* ISCP (intermediate system configuration pointer) */
typedef struct {
	short		busy;
	unsigned	:16;
	swap_t		scb_ptr;
} iscp_t;

typedef enum {
	mode_82586,
	mode_32_bit_segmented,
	mode_linear,
	mode_reserved
} sysbus_mode_t;
 
typedef struct {
/*	unsigned	:9; */
        unsigned        :25;
	boolean_t	csw:1;
	boolean_t	interrupt:1;
	boolean_t	lock:1;
	boolean_t	trg:1;
	sysbus_mode_t	mode:2;
	boolean_t	x:1;
} sysbus_t;

/* SCP (system configuration pointer), aligned on 16-byte boundary */
typedef struct {
/*	unsigned	:16; */
	sysbus_t	sysbus;
	unsigned	:32;
	swap_t		iscp_ptr;	/* pointer to the intermediate system configuration pointer */
} scp_t;

/* Selftest format, aligned on a 16-byte boundary, rounded to take up a whole # of cachelines	*/
typedef struct {
	unsigned	signature:32;				/* ROM signature */
	union {
		struct {
			unsigned	selftest:1;		/* selftest result */
			unsigned	:9;
			unsigned	diagnosetest:1;		/* diagnose result */
			unsigned	bustimertest:1;		/* bus timer result */
			unsigned	registertest:1;		/* register result */
			unsigned	romcontenttest:1;	/* rom content result */
			unsigned	zeros:18;		/* set to zero when done */
		} of;
		unsigned all:32;				/* aggregate pass/fail */
	} result;
	u_char	pad[DCACHE_LINE_SIZE-8];			/* pad to take a whole cacheline */
} selftest_t;

/* Dump format (look in the handbook...), aligned on a 16-byte boundary, rounded up to take up whole # of cachelines */
typedef u_char dump_t[CACHE_ALIGNED(306)];

/* port operations */
typedef enum {
	port_reset,
	port_selftest,	
	port_scp,	
	port_dump	
} iport_t;

/*
 * Cache aligned data for Ethernet controller. 
 */
typedef struct {
	nop_t	nop;					/* space for special NOP command block */
	scb_t	scb;					/* System Control Block */
	rfd_t	rfd_list[NRFDBUF];			/* list of Receive Frame Descriptors */
	rbd_t	rbd_list[NRBDBUF];			/* list of Receive Buffer Descriptors */
	tcb_t	tcb_list[NTCBBUF];			/* list of Transmit Command Blocks */
	dump_t	dump;					/* Dump of i82596 internal guts */
} ecsc_t;

typedef enum {
	warn_link_ok,
	warn_transceiver_cable,
	warn_ethernet_open,
	warn_ethernet_short,
	warn_cb_timeout,
} warning_t;

/*
 * Ethernet Controller Software Interface.
 *
 * Each interface is referenced by a network interface structure,
 * ec_if, which the routing code uses to locate the interface.
 * This structure contains the output queue for the interface, its address, ...
 */
typedef struct {
	struct arpcom	en_ac;				/* common Ethernet structures */
#define ec_if		en_ac.ac_if			/* network-visible interface */
#define ec_addr	 en_ac.ac_enaddr			/* hardware Ethernet address */
#define	ec_bcastaddr	en_ac.ac_bcastaddr		/* link broadcast address */
#define ECQ		ec->ec_if.if_snd		/* interface queue */
	volatile struct	csr *ec_csr;			/* hardware registers */
	boolean_t	initialized;			/* finally, a flag to say we are up to it */
	boolean_t	rxtimeout;			/* receiver restart timer running */
	boolean_t	txactive;			/* transmitter active */
	boolean_t	q_op;				/* queued operation mode */
	boolean_t	txsetup;			/* some process is busy setting up transmits */
	boolean_t	cbtimeout;			/* timeout during eccbstart() */
	boolean_t	cbabort;			/* abort in eccbstart() */
	boolean_t	thin_lan;			/* true if we are using THinLAN, AUI otherwise */
	boolean_t	recover;			/* request watchdog to recover ec interface */
	boolean_t	scberrs;			/* true when we observe the error counters */
	boolean_t	txq;				/* true if want to use full command chaining */
	rfd_t		*rfd_head;			/* pointer to head of list of RFDs */
	rfd_t		*rfd_tail;			/* pointer to tail of list of RFDs */
	rbd_t		*rbd_head;			/* pointer to head of list of RBDs */
	rbd_t		*rbd_tail;			/* pointer to tail of list of RBDs */
	tcb_t		*tcb_head;			/* pointer to head of list of TCBs */
	tcb_t		*tcb_tail;			/* pointer to tail of list of TCBs */
	lock_data_t	init_lock;			/* lock to initialize controller */
	lock_data_t	q_lock;				/* lock to non-queued controller access */
	lock_data_t	ioctl_lock;			/* ioctl access lock to controller */
	ecsc_t		*ecsc;				/* pointer to cache aligned data (allocated space follows) */
	ihandler_id_t	*int_id;
	unsigned long	signature;			/* i82596 ROM signature to determine chip version */
	unsigned long	dogwalks;			/* # time watchdog ran (equals seconds of driver operation) */
	unsigned long	ru_noresources;			/* # times receive unit ran out of buffers (benign) */
	struct {
		unsigned long	late_collision;
		unsigned long	no_carrier_sense;
		unsigned long	lost_clear_to_send;
		unsigned long	dma_underrun;
		unsigned long	deferred;
		unsigned long	too_many_collisions;
	} xmit_status;					/* transmit completion status */
	warning_t	last_warning;			/* remember last warning printed */
	int		request_tdr;			/* accumulate requests for TDR */

#define IFF_CHIPCONFIG	(IFF_LOOPBACK|IFF_PROMISC|IFF_ALLMULTI)
	swap_t		rscerrs;			/* last resource error count observed by watchdog */
	swap_t		ovrnerrs;			/* last overrun	error count observed by watchdog */
	int		retransmit;			/* force retransmission of this packet */
	int		nretransmit;			/* # times we did above */
	int		ufotransmit;			/* # times transmit did not have OK nor error bits set */
	int		txtimestamp;			/* internal clock value of last tx activity */
	int		txtimelap;			/* how many internal clock ticks we will wait */
	struct {					/* reasons to reinitialize the i82596 */
		int		 scb_not_ready;		/* # times SCB was not ready */
		int		 scb_corrupted;		/* # times SCB status word was corrupted */
		int		 tcb_corrupted;		/* # times TCB forward link was corrupted */
		int		 nop_completed;		/* # times NOP CB was completed */
		int		 xmit_timeout;		/* # times transmit was hung */
		int		 rfd_not_completed;	/* # times RFD was not marked completed */
		int		 ru_not_ready;		/* # times RU was not ready */
	} reason;
	eaddr_t		mc[NMULTICAST];			/* packed multicast address list */
	int		nmc;				/* number of valid entries in above list */
	tbd_t		tbd_list[NTCBBUF][NTBDBUF];	/* list of Transmit Buffer Descriptors for each TCB */
	int		ntbd_shortages;			/* # of times we had to copy an output packet */
	char		pkt[CACHE_ALIGNED(ETHERMAXPKT)];/* space to compose packets that are short of TBDs */
	tcb_t		*pkt_owner;			/* TCB that owns above pkt buffer */
	boolean_t	force_selftest;			/* true if we are want to execute selftest */
	short		if_flags;			/* current chip config for following flags only: */        
} ecsi_t;
							
ecsi_t ecsi[NEC];
#ifndef KERNEL_COFF
#pragma hp_align POP
#endif
        
private rbd_t	rbd_template;		/* template to initialize new RBDs */
private cb_t	cb_template;		/* template to initialize new CBs */

/* cache handling macros */
#define PURGE(x,y)	pmap_purge_range(kernel_pmap, (x), (y))
#define FLUSH(x,y)	pmap_flush_range(kernel_pmap, (x), (y))

/* convert va to pa, swap long halves and store in halfword aligned location */
#if DEBUG
#define POINT(unaligned_ptr, va) { \
	register unsigned long _pa; \
	extern vm_offset_t avail_start, avail_end; \
	_pa = pmap_extract(kernel_pmap, va); \
	if (_pa == 0) panic("ec: vtop failed"); \
	if (_pa > avail_end) panic("ec: illegal physical address"); \
	(unaligned_ptr)->low  = _pa & 0xffff; \
	(unaligned_ptr)->high = _pa >> 16; } 
#else
#define POINT(unaligned_ptr, va) { \
	register unsigned long _pa; \
	_pa = pmap_extract(kernel_pmap, va); \
	if (_pa == 0) panic("ec: vtop failed"); \
	(unaligned_ptr)->low  = _pa & 0xffff; \
	(unaligned_ptr)->high = _pa >> 16; } 
#endif

/* mark unaligned pointer as invalid */
#define INVALID(unaligned_ptr) { \
	(unaligned_ptr)->low  = INVALID_POINTER; \
	(unaligned_ptr)->high = INVALID_POINTER; }

/* macros to compare ethernet addresses */
#define ECADDRMATCH(a,b) \
	((((vm_offset_t)(a)+0) == ((vm_offset_t)(b)+0)) && \
	 (((vm_offset_t)(a)+2) == ((vm_offset_t)(b)+2)) && \
	 (((vm_offset_t)(a)+4) == ((vm_offset_t)(b)+4)))

/* macro to compare link addresses */
#define LINKMISMATCH(a,b) (((a).low != (b).low) || ((a).high != (b).high))

/*
 * macro to set for valid individual address. note that the least significant
 * bit of a byte is transmitted first, so the group/individual address bit,
 * which is the first bit on the wire, is the lsb bit of the first address byte.
 */
#define INVALIDADDR(ia)	(ia[0] & 0x1)


/* flag to panic instead of silently recover for debugging */
int eccrash = 0;

/* special page with low 16 bits of PPN set to 0x4000, see ec_special_page() */
vm_offset_t ecsc_area;


/*
 * Access i82596 through port interface.
 */
private void
ecportaccess(ec, pa, cmd)
ecsi_t *ec;
unsigned long pa;
iport_t cmd;
{
	volatile struct csr *csr = ec->ec_csr;

	csr->port = (pa >> 16);
	DELAY(1);		 /* keep at least one cycle between writes */
	csr->port = (pa & 0xfff0) | cmd;
	DELAY(100);		/* give chip time (20 clocks + 0.5 microsecs) to get its acts together */
}


/*
 * Wait until the SCB is ready to accept the next command. The chip may	
 * delay acknowledging a command written to the SCB, because it is busy
 * handling higher priority tasks (e.g. prefetching buffers, finishing
 * frame reception). This should not take long and it will soon clear the
 * SCB command word. We must wait until this clearing has taken place so
 * that we do not interfere with a previously issued command. Iff the SCB
 * does not clear within 100 microseconds, the chip is hung and we return
 * FALSE.
 */
private boolean_t
ecready(ec)
ecsi_t *ec;
{
	scb_t		*scb = &ec->ecsc->scb;
	unsigned int	microsec = 100;

	if (ec->recover)
		return FALSE;

	while (microsec-- != 0) {
		PURGE(&scb->scb.command, sizeof(scb->scb.command));
		if (!scb->scb.command.all)
			return TRUE;
		/* wait one microsecond */
		DELAY(1);
	}
#if DEBUG
	if (ECDEBUG) printf("ecready: cb never cleared\n");
#endif
	ec->reason.scb_not_ready++;
	if (eccrash) {
		printf("ec: SCB never cleared\n");
		kdb_break();
	}
	ec->recover = TRUE;
	return FALSE;
}


/*
 * Setup receive area, returns TRUE if all receive buffers could be allocated.
 */
private boolean_t
ecrxsetup(ec, canwait)
ecsi_t *ec;
int canwait;
{
	rfd_t	*rfd;
	rbd_t	*rbd;

	/* recycle RFDs */
	while ((rfd = ec->rfd_tail->rfd.next)->rfd.next != ec->rfd_head) {
		/* reinitialize RFD */
		rfd->rfd.command.cacheline.cb_stat = 0;
		rfd->rfd.command.cmd.el = TRUE;
#if DEBUG
		/* these fields are never supposed to be set */
		rfd->rfd.eof = FALSE;
		rfd->rfd.count = 0;
#endif
		rfd->rfd.filled = FALSE;	/* this IS set by controller... */
		INVALID(&rfd->rfd.rbd_ptr);
		FLUSH(rfd, sizeof(rfd_t));
		/* and remove end of list marker of previous RFD */
		PURGE(&ec->rfd_tail->rfd.command.cacheline.cb_cmd, DCACHE_LINE_SIZE);
		ec->rfd_tail->rfd.command.cmd.el = FALSE;
		FLUSH(&ec->rfd_tail->rfd.command.cacheline.cb_cmd, DCACHE_LINE_SIZE);
		ec->rfd_tail = rfd;
	}

	/* recycle RBDs and allocate receive buffers */
	while ((rbd = ec->rbd_tail->rbd.next)->rbd.next != ec->rbd_head) {

		struct mbuf *m;

		/* obtain an MBUF HDR */
		MGETHDR(m, canwait, MT_DATA);
		if (!m) {
#if DEBUG
			if (ECDEBUG) printf("ecrxsetup: no mbuf...\n");
#endif
			return FALSE;
		}

		/* attach a cluster */
		MCLGET(m, canwait);
		if (!m->m_ext.ext_buf) {
#if DEBUG
			if (ECDEBUG_RX) printf("ecrxsetup: no cluster...\n");
#endif
			m_free(m);
			return FALSE;
		}

		/* reinitialize RBD */
		rbd->rbd.m = m;
		rbd->rbd.el = TRUE;
		rbd->rbd.eof = FALSE;
		rbd->rbd.filled = FALSE;
		rbd->rbd.count = 0;
		/* dma further into buffer to align IP header */
		rbd->rbd.m->m_data += RBDOFFSET;
		POINT(&rbd->rbd.buffer, rbd->rbd.m->m_data);
		/* purge the receive buffer (and do not touch it until data has been received) */
		PURGE(mtod(rbd->rbd.m, char *), MIN(rbd->rbd.size, ETHERMAXPKT));
		/* now push rbd to memory */
		FLUSH(rbd, sizeof(rbd_t));
		/* and remove end of list marker of previous RBD */
		PURGE(ec->rbd_tail->cacheline.cmd, DCACHE_LINE_SIZE);
		ec->rbd_tail->rbd.el = FALSE;
		FLUSH(ec->rbd_tail->cacheline.cmd, DCACHE_LINE_SIZE);
		ec->rbd_tail = rbd;
	}
	return TRUE;
}


/*
 * Check incoming packet for errors and handle trailer packets
 */
private struct mbuf *
ecread(ec, eh)
ecsi_t *ec;
struct ether_header *eh;
{
	rfd_t		*rfd;
	rbd_t		*rbd;
	struct mbuf	*t, *m;
	unsigned int	datlen, traildatlen;
	int		i, ncl;
	struct trailhdr_t {
		u_short ether_type;
		u_short len;
	} *trailhdr;

	rfd = ec->rfd_head;
	rbd = ec->rbd_head;
	PURGE(&rfd->rfd.command.cacheline.cb_cmd, DCACHE_LINE_SIZE);
	PURGE(rbd, sizeof(rbd_t));

#if DEBUG
	if (ECDEBUG_CHECK) {
		if (!rfd->rfd.command.cmd.ok)
			panic("ecread: !OK in RFD");

		/* these fields are never activated because we do not place receive data in RFD */
		/* the 'filled' flag does get set by i82596... */
		if ((rfd->rfd.eof) || (rfd->rfd.count != 0))
			panic("ecread: EOF or FILLED or COUNT!=0 in RFD");

		/* check on validity of some fields */
		if (!rbd->rbd.filled || (rbd->rbd.count < (ETHERMIN + sizeof (struct ether_header)))
		|| (rbd->rbd.count > (4 /* crc? xxx */ + ETHERMAXPKT))) {
			printf("ecread: bad rbd count is %d\n", rbd->rbd.count);
			panic("ecread: !FILLED or COUNT in first RBD");
		}
	
		/*
		 * note: the RFD should point to the RBD but we cannot garantuee that
		 * the rbd_ptr field did not get overwritten by us which could have
		 * happened: if the controller was writing the rbd_ptr field at the
		 * same moment that we were resetting the EL bit in that RFD while 
		 * linking in a new RFD, we may have bashed it. Intel assures us that
		 * RBDs get taken from the RBD list in strict order, so we can do without
		 * that pointer. as a sanity check, we see if the rbd_ptr field is odd
		 * which means that it has not been set, or otherwise we compare the
		 * pointer to what we expect. its ugly but only for debug...
		 */
		/* check the pointer in RFD to first RBD */	
		if ((rfd->rfd.rbd_ptr.low & 1) == 0)
/*			if (rbd->rbd.pa != rfd->rfd.rbd_ptr) {
				printf("ecread: mismatch rbd->rbd.pa=%x:%x rfd->rbd_ptr=%x:%x\n",
				(int) rbd->rbd.pa.high, (int) rbd->rbd.pa.low,
				(int) rfd->rfd.rbd_ptr.high, (int) rfd->rfd.rbd_ptr.low);
				panic("ecread: out of sync");
			}
*/
		/* sanity check on mbuf pointer */
		if (!rbd->rbd.m)
			panic("ecread: m is nill");
	}
#endif
	/*
	 * note: we do not purge the data cache to look at the received data 
	 * because we already did when we allocated this buffer and posted
	 * it to the RU. we assume that nobody touched it since than so that
	 * no old data has been cached at this time
	 */

	/* copy out ethernet header */
	m = rbd->rbd.m;
	bcopy(mtod(m, char *), eh, sizeof(struct ether_header));
	m->m_data += sizeof(struct ether_header);

	/* when multicasting, verify the destination address (filter is not exact) */
	if (ec->nmc && (eh->ether_dhost[0] & 1) && !(ec->if_flags & (IFF_ALLMULTI|IFF_PROMISC))) {

		/* see if this is the broadcast address */
		if (ECADDRMATCH(eh->ether_dhost, etherbroadcastaddr)) {
#if DEBUG
			ec4all++;
#endif
			goto forus;
		}
		/* see if this is an accepted multicast address */
		for (i = 0; i < ec->nmc; i++)
			if (ECADDRMATCH(ec->mc[i], eh->ether_dhost)) {
#if DEBUG
				ec4multi++;
#endif
				goto forus;
		}
		/* not for us */
#if DEBUG
		ec4other++;
#endif
		goto ignore; 
	}

#if DEBUG
	if (ECADDRMATCH(eh->ether_dhost, ec->ec_addr))
		ec4me++;
	else if (ECADDRMATCH(eh->ether_dhost, etherbroadcastaddr))
		ec4all++;
	else if ((eh->ether_dhost[0] & 1) && (ec->if_flags & IFF_ALLMULTI))
		ec4multi++;
	else if (ec->if_flags & IFF_PROMISC) 
		ec4any++;
	else
		ec4who++;
#endif

	/* set packet header */
forus:	m->m_pkthdr.rcvif = &ec->ec_if;
#if DEBUG
	if (ECDEBUG_PRINTALL)
		printf("ecread #1: %d bytes\n", rbd->rbd.count);
#endif
	/* It seems we get 4 extra bytes on every read.	Apparently the frame CRC.	Ignore them. */
	m->m_pkthdr.len = m->m_len = rbd->rbd.count - sizeof(struct ether_header) - 4 /*frame CSM : XXX*/;

	/* check for more buffers in current packet */
	ncl = 1;
	while (!ec->rbd_head->rbd.eof) {
		rbd_t *curr, *next;

#if DEBUG
		/* check to see that we do not have too many buffers */
		if (++ncl > NCLPACKET)
			panic("ecread: too many buffers per packet");
#endif

		next = (curr = ec->rbd_head)->rbd.next;
		PURGE(next, sizeof(rbd_t));
#if DEBUG
		if (ECDEBUG_CHECK) {
			if (!next->rbd.filled)
				panic("ecread: !FILLED or COUNT in RBD");

			/* check the pointer in RBD to next RBD */
			if ((curr->rbd.link.low & 1) == 0)
				if (((curr->rbd.link.low & 0xffffL) | (curr->rbd.link.high << 16)) != (int) &next->cmd.stat)
					panic("ecread: rbd out of sync");
		}
		if (ECDEBUG_PRINTALL)
			printf("ecread #2: %d bytes\n", next->rbd.count);
#endif
		/* set mbuf data */
		next->rbd.m->m_len = next->rbd.count;
		next->rbd.m->m_flags &= ~M_PKTHDR;
		m->m_pkthdr.len += next->rbd.count;
		/* link to previous packet */
		curr->rbd.m->m_next = next->rbd.m;
		/* next packet to process */
		ec->rbd_head = next;
	}

#if DEBUG
	/* verify total packet length */
	if (ECDEBUG_CHECK	&&	m->m_pkthdr.len > ETHERMTU) {
		printf("ecread: m_pkthdrlen is %d\n", m->m_pkthdr.len);
		panic("ecread: received packet too large");
	}
#endif

	/* and check for trailer packets */
	if ((eh->ether_type >= ETHERTYPE_TRAIL) && (eh->ether_type < ETHERTYPE_TRAIL+ETHERTYPE_NTRAILER)) {
		datlen = (eh->ether_type - ETHERTYPE_TRAIL) << 9;
		traildatlen = m->m_pkthdr.len - datlen;
		if (traildatlen >= sizeof(struct trailhdr_t)) {
			/* find start of trailer */

			/*
			 * note that we assume that the trailer is entirely contained in
			 * contiguous memory. this is always true. the packet data is a
			 * multiple of 512 bytes and comes first, filling up the cluster.
			 * the trailer data will never be split between clusters: if the
			 * clustersize (MCLSIZE) is larger than 512 bytes, the trailer will
			 * start at least 512-RBDOFFSET bytes before the end of the cluster.
			 * if the clustersize is smaller or equal to 512 bytes, the trailer
			 * will always start at RBDOFFSET bytes from the start of the cluster.
			 * we (heuristically) assume that the trailer fits in MHLEN bytes and
			 * clusters will always be larger than MBUFs themselves. so there...
			*/
			trailhdr = NULL;
			for (t = m; t; t = t->m_next, datlen -= t->m_len)
				if (t->m_len > datlen) {
					trailhdr = (struct trailhdr_t *) (mtod(t, char *) + datlen);
					break;
				}
#if DEBUG
			if (ECDEBUG_CHECK) {
				if (!trailhdr)
					panic("ecread: can't find trailer header");
				if (t->m_len < (datlen + traildatlen))
					panic("ecread: not enough data for trailer");
			}
#endif
			/* truncate trailer data from this mbuf */
			t->m_len = datlen;
			if (trailhdr->len <= traildatlen) {
				eh->ether_type = trailhdr->ether_type;
				/* place trailer data in separate hdr mbuf */
				MGETHDR(t, M_DONTWAIT, MT_DATA);
				if (t != NULL) {
					/* copy trailer data */
					t->m_len = trailhdr->len - sizeof(struct trailhdr_t);
					if (t->m_len > MHLEN)
						panic("ecread: insufficient space for trailer");
					bcopy((char *) (trailhdr+1), mtod(t, char *), t->m_len);
					/* copy pkthdr data */
					t->m_pkthdr.rcvif = m->m_pkthdr.rcvif;
					t->m_pkthdr.len = m->m_pkthdr.len - sizeof(struct trailhdr_t);
					m->m_flags &= ~M_PKTHDR;
					/* link trailer data to head of chain */
					t->m_next = m;
					m = t;
					goto success;
				}
			}
		}
		/* error in trailer packet if we get here */
#if DEBUG
		if (ECDEBUG_RX) printf("ecread: ipacket discarded\n");
#endif
		ec->ec_if.if_ierrors++;
		if (ec->ec_if.if_flags & IFF_DEBUG)
			printf("ecread: if_ierrors++\n");
ignore:	 m_freem(m);
		m = NULL;
		/* fall through */
	} 

success:
	/* signal ecreset that these mbufs do not need to be deallocated */
	rbd->rbd.m = NULL;
	FLUSH(&rbd->rbd.m, sizeof(rbd->rbd.m));
	while (rbd != ec->rbd_head) {
		rbd = rbd->rbd.next;
		rbd->rbd.m = NULL;
		FLUSH(&rbd->rbd.m, sizeof(rbd->rbd.m));
	}

#if DEBUG
	if (m && (ECDEBUG))
		printf(" R%d", rbd->rbd.count);
#endif
	/* return pointer to mbuf chain */
	return m;
}


#if DEBUG
/*
 * keep list of mbuf chains we send and receive
 */
#define ECNMBUF	20
struct mbuf *ecsend[ECNMBUF], *ecreceive[ECNMBUF];
int ecsindex, ecrindex;
#endif


/*
 * Frame Received interrupt routine. In order to protect against a hung receive side as a result of the
 * I82596 corruption (see 'second big fat notice' above), we must make sure that we do not get stuck
 * with a missing completion bit in a RFD. Thus if we get a FR interrupt and do not have at least one
 * RFD to process, we will look ahead to the next RFD (if there is any) and see if its completion bit
 * is set. if that is the case, the bug happened and we must reinitialize.
 */
private void
ecfrint(ec)
ecsi_t *ec;
{
	struct ifnet		*ifp = &ec->ec_if;
	struct ether_header	eh;
	struct mbuf		*m;
	rfd_t			*rfd;
	int			n = 0;

	/* receive and route incoming packets */
	while (!ec->recover) {
		/* see if we received a packet */
		rfd = ec->rfd_head;
		PURGE(&rfd->rfd.command.cacheline.cb_stat, sizeof(rfd->rfd.command.cacheline.cb_stat));
		if (rfd->rfd.command.cmd.completion && I82596RFDVALID(rfd)) {
			n++;
			if (m = ecread(ec, &eh)) {
#if DEBUG
				/* keep mbuf pointer for debugging */
				ecreceive[ecrindex] = m;
				if (++ecrindex == ECNMBUF)
					ecrindex = 0;
#endif
				/* route the packet */
				ether_input(ifp, &eh , m);
				ifp->if_ipackets++;
				panel_led(PANEL_LED_RECV, PANEL_LED_TOGGLE);
			}
			ec->rfd_head = ec->rfd_head->rfd.next;
			ec->rbd_head = ec->rbd_head->rbd.next;
 
			/* recycle receive buffers */
			ecrxsetup(ec, M_DONTWAIT);
		} else {
			if (n == 0) {
				/* check to see if next RFD has completion bit set */
				PURGE(&rfd->rfd.command.cacheline.cb_cmd, sizeof(rfd->rfd.command.cacheline.cb_cmd));
				if (!rfd->rfd.command.cmd.el) {
					/* there is a next RFD */
					rfd = rfd->rfd.next;
					PURGE(&rfd->rfd.command.cacheline.cb_stat, sizeof(rfd->rfd.command.cacheline.cb_stat));
					if (rfd->rfd.command.cmd.completion && I82596RFDVALID(rfd)) {
						ec->reason.rfd_not_completed++;
						if (eccrash) {
							printf("ec: RFD completion bit not set\n");
							kdb_break();
						}
						ec->recover = TRUE;
					}
				}
			}				
			return;
		}
	}
}


/*
 * Startup the Receive Unit.	Can be called from softclock scan.
 */
void
ecrxstart(ec)
ecsi_t *ec;
{
	scb_t	*scb;
	int	s = splimp();

	/* if receiver timer running, cancel it */
	if (ec->rxtimeout) {
#if DEBUG
		if (ECDEBUG_RX) printf("ecrxstart: timeout\n");
#endif
		untimeout(ecrxstart,ec);
		ec->rxtimeout = FALSE;
	}

	/* allocate receive buffers */
	if (!ecrxsetup(ec, M_DONTWAIT)) {
		/* no resources, try later in one sec */
#if DEBUG
		if (ECDEBUG_RX) printf("ecrxstart: no buffers\n");
#endif
		ec->rxtimeout = TRUE;
		timeout(ecrxstart, ec, hz);
		/* ask watchdog for mercy on RU not ready */
		goto out;
	}

	/* wait until the chip is ready to accept command */
	scb = &ec->ecsc->scb;
	if (!ecready(ec)) {
		/* let watchdog restart chip */
		goto out;
	}

	/* point first RFD to first RBD */
	POINT(&ec->rfd_head->rfd.rbd_ptr, &ec->rbd_head->cmd.stat);
	FLUSH(&ec->rfd_head->rfd.rbd_ptr, sizeof(ec->rfd_head->rfd.rbd_ptr));

	/* tell SCB to accept a CB (note that the remaining SCB command bits are already zero, see ecready) */
	scb->scb.command.cmd.ruc = scb_ruc_start;
	POINT(&scb->scb.rfa_ptr, &ec->rfd_head->rfd.command.cacheline.cb_stat);

	/* push these out to memory */
	FLUSH(scb->cacheline.cmd, sizeof(scb->cacheline.cmd));

	/* assert channel attention */
	ec->ec_csr->ca = 0;
out:
	splx(s);
}


/*
 * Startup the Command Unit. Must be called with interrupts disabled.
 */
private void
eccustart(ec, cb)
ecsi_t *ec;
cb_t *cb;
{
	scb_t *scb;

	/* wait until the chip is ready to accept commands */
	scb = &ec->ecsc->scb;
	if (!ecready(ec))
		/* let watchdog restart chip */
		return;

	/* tell SCB to accept a CB (note that remaining SCB command bits are already zero, see ecready) */
	scb->scb.command.cmd.cuc = scb_cuc_start;
	POINT(&scb->scb.cbl_ptr, &cb->nop.nop.command.cacheline.cb_stat);

	/* push these out to memory (all on one cacheline) */
	FLUSH(scb->cacheline.cmd, sizeof(scb->cacheline.cmd));

	/* assert channel attention */
	ec->ec_csr->ca = 0;
}


/*
 * Drop input queue
 */
private void
ecdropq(ec)
ecsi_t *ec;
{
	struct mbuf	*m;
	int		pl = splimp();

	while (TRUE) {
		IF_DEQUEUE(&ECQ, m);
		if (m == NULL)
			break;
#if DEBUG
		if (ECDEBUG_PRINTALL) printf("ecdropq: ...\n");
#endif
		IF_DROP(&ECQ);
		m_freem(m);
	}
	splx(pl);
}


#if DEBUG
/* fake transmit timeouts */
int ecto = 0;
#endif


/*
 * Transmit complete processing
 */
private void
ectxint(ec)
ecsi_t *ec;
{
	tcb_t		*tcb;

#if DEBUG
	if (ecto)
		return;
#endif
	while ((tcb = ec->tcb_head) != ec->tcb_tail->tcb.next) {

		/* see if we transmitted this packet */
		PURGE(&tcb->tcb.command.cacheline.cb_stat, sizeof(tcb->tcb.command.cacheline.cb_stat));
		if (tcb->tcb.command.cmd.completion) {

			boolean_t	fatal = FALSE;

			/* refresh timestamp to keep watchdog at bay */
			ec->txtimestamp = it_read();

			/* inspect the link address to check for address corruption */
			PURGE(&tcb->tcb.command.cacheline.cb_cmd, sizeof(tcb->tcb.command.cacheline.cb_cmd));
			if (LINKMISMATCH(tcb->tcb.command.save.link, tcb->tcb.command.cmd.link)) {
				ec->reason.tcb_corrupted++;
				if (eccrash) {
					printf("ec: TCB forward pointer corrupted\n");
					kdb_break();
				}
				ec->recover = TRUE;
				return;
			}
#if DEBUG
			/* keep mbuf pointer for debugging */
			ecsend[ecsindex] = (ec->retransmit) ? (struct mbuf *) 0x22222222 : tcb->tcb.m;
			if (++ecsindex == ECNMBUF)
				ecsindex = 0;
#endif

			/* inspect error bits */
			if (tcb->tcb.command.xmit.late_collision) {
				ec->xmit_status.late_collision++;
				fatal = TRUE;
				ec->ec_if.if_collisions++;
			}
			if (tcb->tcb.command.xmit.no_carrier_sense) {
				ec->xmit_status.no_carrier_sense++;
				fatal = TRUE;
				ec->request_tdr++;	/* more reason to check the cable */
			}
			if (tcb->tcb.command.xmit.lost_clear_to_send) {
				ec->xmit_status.lost_clear_to_send++;
				fatal = TRUE;
				ec->request_tdr++;	/* more reason to check the cable */
			}
			if (tcb->tcb.command.xmit.dma_underrun) {
				ec->xmit_status.dma_underrun++;
				fatal = TRUE;
			}
			if (tcb->tcb.command.xmit.deferred) {
				ec->xmit_status.deferred++;
			}
			if (tcb->tcb.command.xmit.maxcoll) {
				/* count number of collisions */
				ec->ec_if.if_collisions += tcb->tcb.command.xmit.maxcoll;
				/* show collisions on front panel */
				panel_led(PANEL_LED_COLLISION, PANEL_LED_TOGGLE);
			} else if (tcb->tcb.command.xmit.too_many_collisions) {
				ec->xmit_status.too_many_collisions++;
				fatal = TRUE;
				ec->ec_if.if_collisions += 16;
			}

			if (tcb->tcb.command.cmd.ok) {
#if DEBUG
				if (ECDEBUG) printf(" T");
#endif
				/* all is fine */
				ec->ec_if.if_opackets++;
				ec->request_tdr = 0;
				panel_led(PANEL_LED_XMIT, PANEL_LED_TOGGLE);
			} else {
				/*
				 * record number of funny cases where we do not have the OK bit
				 * (and we dont) but neither have any fatal error bits set
				 */
				if (!fatal) {
					ec->ufotransmit++;
					if (eccrash) {
						printf("ec: no error status on transmit\n");
						kdb_break();
					}
				}
							
				if (ec->retransmit++ == 0) {
					/* we will try to retransmit this packet once */
					ec->nretransmit++;
					tcb->tcb.command.cmd.el = TRUE;	 /* this block only ! */
					tcb->tcb.command.cacheline.cb_stat = 0; /* clear status bits */
					/* push TCB out to memory */
					FLUSH(tcb, sizeof(*tcb));
					/* this packet will be retransmitted by interrupt handler */
					return;
				} else {
#if DEBUG
					if (ECDEBUG) printf(" X");
#endif
					if (eccrash) {
						printf("ec: retransmit failed\n");
						kdb_break();
					}
					ec->ec_if.if_oerrors++;
					if (ec->ec_if.if_flags & IFF_DEBUG)
						printf("ectxint: if_oerrors++\n");
#if DEBUG
					panel_led(PANEL_LED_DIAG3, PANEL_LED_TOGGLE);
#endif
					/*
					 * consecutive errors (and troublesome errors) will cause us to
					 * ask the watchdog to perform time domain reflectometry once
					 * the following value has been incremented past a certain limit.
					 */
					ec->request_tdr++;
				}
			}
			/* no retries */
			ec->retransmit = 0;

			/* check if pkt buffer is now free */
			if (tcb == ec->pkt_owner)
				ec->pkt_owner = NULL;

			/* deallocate xmit chain (and record the fact for ecreset) */
			m_freem(tcb->tcb.m);
			tcb->tcb.m = NULL;
			FLUSH(&tcb->tcb.m, sizeof(tcb->tcb.m));

			/* we have at least one free slot */
			ec->tcb_head = ec->tcb_head->tcb.next;
			ec->ec_if.if_flags &= ~IFF_OACTIVE;
		} else
			break;
	}
}


/*
 * Setup transmit structures. Take next packet from the send queue and hand it to the controller
 * until we run out of Transmit Control Blocks.
 */
private int
ectxsetup(ec)
ecsi_t *ec;
{
	struct mbuf	*m, *mp, *ml, *tm, *mnext;
	int		len, pl, i, ntbd;
	tbd_t		*tbd, *tbd_head;
	tcb_t		*tcb;
	boolean_t	last;

	if (!ec->initialized)
		/* not yet */
		return;

	/* we must have the controller up and running */
	if ((ec->ec_if.if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)) {
		ecdropq(ec);
		return;
	}

	/* dont bother if chip is stuck */
again:	if (ec->recover)
		return;

	/* exclude interrupts */
	pl = splimp();

	/* some process can already be in this routine */
	if (ec->txsetup) {
		splx(pl);
		return;
	}

	/* start transmit of any pending packets while we setup the next */
	if (!ec->txactive && ec->q_op && (ec->tcb_head != ec->tcb_tail->tcb.next)) {
		/* start transmitting */
		ec->txactive = TRUE;
		ec->txtimestamp = it_read();
#if DEBUG
		if (ECDEBUG) printf("C");
		if (!ecto) 
#endif
		eccustart(ec, (cb_t *) ec->tcb_head);
	}

	/* if we are using the pkt buffer we must wait until it is free */
	if (ec->pkt_owner) {
wait:		ec->ec_if.if_flags |= IFF_OACTIVE;
		splx(pl);
		return;
	}

	/* we must have a TCB available to setup for transmission */
	if ((tcb = ec->tcb_tail->tcb.next)->tcb.next == ec->tcb_head) {
		/* see if any TCBs have already been completed */
		ectxint(ec);
		if ((tcb = ec->tcb_tail->tcb.next)->tcb.next == ec->tcb_head)
			goto wait;
	}

	/* we must have another packet to setup */
	IF_DEQUEUE(&ECQ, m);
	if (!m) {
		splx(pl);
		return;
	}

	/* indicate we are busy formatting this TCB */
	ec->txsetup = TRUE;
	splx(pl);

#if DEBUG
	if (ECDEBUG_PRINTALL) {
		printf("ectxsetup: dequeued mbuf:\n");
		mbuf_print_list(m);
	}
#endif

	/* fill in the list of Transmit Buffer Descriptors */
	len = ntbd = 0;
	tbd_head = tbd = ec->tbd_list[tcb - ec->ecsc->tcb_list];
	for (mp = m; mp; mp = mnext) {

		char	*mptr = mtod(mp,char*);
		int	mlen = mp->m_len;

		mnext = mp->m_next;

		/* no data, no TBD */
		if (!mlen)
			continue;

		/* count number of TBDs used */
		if (++ntbd >= (NTBDBUF - 1)) {
			/*
			 * we are using the one but last TBD we have. we want to keep the last TBD for use
			 * in padding the data, but will use this TBD to point to a buffer in which we gather
			 * the remaining transmit data from the mbuf chain. we could try to allocate an mbuf
			 * with a cluster to gather this data, but this could fail and we may need multiple...
			 * since this is a rare case (only telnet seems to generate occasional packets with
			 * more than 18 mbufs) we will take a simple solution. we have space left in the 
			 * 'ecsc_t' page, we will use this as a buffer. ofcourse since we only have one such
			 * buffer, we must keep track of ownership.
			 */
			char *cp;

			/* copy all remaining transmit data of the current mbuf chain to pkt buffer */
			for (mptr = cp = ec->pkt, mlen = 0; mp; mp = mp->m_next) {
				mlen += mp->m_len;
				if (mlen > sizeof(ec->pkt)) {
					/* too big */
					if (eccrash) {
						printf("ec: cannot copy packet\n");
						kdb_break();
					}
					goto error;
				}
				bcopy(mtod(mp, caddr_t), cp, mp->m_len);
				cp += mp->m_len;
			}

			/* indicate this buffer is in use */
			ec->pkt_owner = tcb;
			ec->ntbd_shortages++;

			/* note that we will drop out of the loop after setting up the current TBD */
			mnext = NULL;
		}

		/*
		 * the A step chip has trouble with transmit buffers which start on an odd address.
			 * so move one byte into the TBD to transmit it from there and transmit the remaining
			 * mbuf data (now 16-bit aligned) from the next TBD in the next loop iteration.
		 */
		if ((int)mptr & 1) {
			mptr = tbd->oddchar;
/*			*mptr = *mtod(mp,char*)++; */
                 	*mptr = *(mp->m_data)++;
			mp->m_len--;
			mlen = 1;
			mnext = mp;
		} else
			/* flush the mbuf transmit data out to memory */
			FLUSH(mptr, mlen);

		/* fill in the TBD */
		len += mlen;
		tbd->count = mlen;
		tbd->eof = FALSE;
		POINT(&tbd->buffer, mptr);

		/* next tbd */
		tbd++;
	}

	/* see if the packet has an acceptable size */
	if ((len < sizeof(struct ether_header)) || (len > (ETHERMAXPKT))) {
		if (eccrash) {
			printf("ec: illegal output packet size\n");
			kdb_break();
		}
		goto error;
	}

	/* pad the packet out to minimum size if necessary */
	if (len < (ETHERMIN + sizeof(struct ether_header))) {
		/* TBD data for padding packet to minimum length */
		static u_char pbuffer[ETHERMIN];

		if (++ntbd > NTBDBUF)
			panic("ec: not enough TBDs for padding");
		tbd->count = ETHERMIN + sizeof(struct ether_header) - len;
		tbd->eof = TRUE;
		POINT(&tbd->buffer, pbuffer);
	} else
		/* set end of list */
		(tbd - 1)->eof = TRUE;

	/* flush TBDs */
	FLUSH(tbd_head, ntbd * sizeof(tbd_t));

	/* fill the Transmit Command Block */
	tcb->tcb.command.cmd.el = TRUE;	 /* last block on the list */
	tcb->tcb.command.cacheline.cb_stat = 0; /* clear status bits */
	tcb->tcb.m = m;			 /* keep chain pointer so we can free it after xmit */

	/* push TCB out to memory */
	FLUSH(tcb, sizeof(*tcb));

	if (ec->txactive) {
		/* try adding it to previous xmit: remove end of list marker of previous TCB */
			PURGE(&ec->tcb_tail->tcb.command.cacheline.cb_cmd, DCACHE_LINE_SIZE);
		ec->tcb_tail->tcb.command.cmd.el = FALSE;
			FLUSH(&ec->tcb_tail->tcb.command.cacheline.cb_cmd, DCACHE_LINE_SIZE);
	}

#if DEBUG
	if (ECDEBUG) printf(" t%d", len);
#endif

	ec->tcb_tail = tcb;
	ec->txsetup = FALSE;
	goto again;

error:
#if DEBUG
	if (ECDEBUG) printf("x", len);
#endif
	pl = splimp(); IF_DROP(&ECQ); splx(pl);
	m_freem(m);
	ec->txsetup = FALSE;
	goto again;

}


/* 
 * Warn user about cable condition (but only once)
 */
private void
ecwarn(ec, warning, tdr)
ecsi_t *ec;
warning_t warning;
int tdr;
{
	int unit = ec - &ecsi[0];

#if DEBUG
	if (ECDEBUG) printf("ecwarn: returns %d\n", warning);
#endif

	if (tdr) {
		/* The TDR value ticks at 10MHz.
		 * Convert to distance using velocity of propagation.
		 * Distance = 1/2 * v * (ticks/freq).
		 * Speed of light is 3 * 10E8 m/s.
		 * So thick @ 78% propagation => 11.7m/tick
		 *    thin  @ 66% propagation =>  9.9m/tick
		 */
		tdr = (tdr * ((ec->thin_lan) ? 99 : 117)) / 10;
	}

	if (warning != ec->last_warning) switch (warning) {
		case warn_link_ok:
			log(LOG_ALERT, "ec%d: Transmission line access is restored.\n", unit);
			break;
		case warn_ethernet_open:
			log(LOG_ALERT, "ec%d: Transmission line is not properly terminated (around %d meters).\n", unit, tdr);
			break;
		case warn_ethernet_short:
			log(LOG_ALERT, "ec%d: Transmission line has a short circuit (around %d meters).\n", unit, tdr);
			break;
		case warn_transceiver_cable:
			log(LOG_ALERT, "ec%d: Transceiver cable is missing or inoperative.\n", unit);
			break;
		case warn_cb_timeout:
			log(LOG_ALERT, "ec%d: Command timeout, check transmission line access.\n", unit);
			break;
		default:
			log(LOG_ALERT, "ec%d: Unknown TDR status\n", unit);
			break;
	}
	ec->last_warning = warning;
}


/*
 * Timeout during an outstanding operation to i82596DX. Apart
 * from hardware failure this is usually caused by a loose cable.
 */
private void
eccbtimeout(ec)
ecsi_t *ec;
{
	if (ec->ec_if.if_flags & IFF_RUNNING)
		ecwarn(ec, warn_cb_timeout, 0);
	ec->cbabort = TRUE;
	wakeup(ec);
}


/*
 * Make sure this interface is worth waiting for
 */
private boolean_t
eccbabort(ec)
ecsi_t *ec;
{
	if (ec->cbtimeout) {
		/* cancel timeout */
		untimeout(eccbtimeout,ec);
		ec->cbtimeout = FALSE;
	}
	if (ec->cbabort || (ec->ec_if.if_flags & IFF_RUNNING) == 0) {
		/* interface has been taken down or timed out */
		return (ec->cbabort = TRUE);
	} else {
		/* set new timeout for two seconds */
		ec->cbtimeout = TRUE;
		timeout(eccbtimeout, ec, 2*hz);
		return (FALSE);
	}
}


/*
 * Suspend normal queued operation for transmission and issue a command to the controller
 * Note that the hardware design can cause us to hang indefinitely if the cable has fallen
 * out of the connector. In that case we cannot even perform time domain reflectometry because
 * the transmit will fail. thus we must implement a timeout.
 */
private boolean_t
eccbstart(ec, cb)
ecsi_t *ec;
cb_t *cb;
{
	boolean_t	rval;
	int		pl;

	/* flush CB out to memory */
	FLUSH(cb, sizeof(*cb));

	/* take a mutual exclusion lock */
	lock_write(&ec->q_lock);
	pl = splimp();

	ec->cbabort = FALSE;

	/* wait for queued operation mode to complete */
	ec->q_op = FALSE;
	while (!eccbabort(ec) && ec->txactive)
		sleep(ec, PRIBIO);

	if (!ec->cbabort) {
		/* issue control block to controller to execute */
		eccustart(ec, cb);
	
		/* wait for completion of this control block */
		while (!eccbabort(ec)) {
			PURGE(&cb->nop.nop.command.cacheline.cb_stat, sizeof(cb->nop.nop.command.cacheline.cb_stat));
			if (!cb->nop.nop.command.cmd.completion) {
				sleep(ec, PRIBIO);
			} else
				break;
		}
	}

	if (ec->cbtimeout) {
		/* cancel timeout */
		untimeout(eccbtimeout,ec);
		ec->cbtimeout = FALSE;
	}

	/* restore queued operation mode */
	ec->q_op = TRUE;

	/* determine return value before releasing lock */
	rval = !ec->cbabort && cb->nop.nop.command.cmd.ok;

	splx(pl);
	lock_done(&ec->q_lock);

	/* resume transmission */
	ectxsetup(ec);
	return (rval);
}


/*
 * Perform Time Domain Reflectometry to get cable status
 */
private void
ectdr(ec)
ecsi_t *ec;
{
	u_char	cb_area[sizeof(cb_t) + DCACHE_LINE_SIZE];
	cb_t	*cb = (cb_t *) CACHE_ALIGNED(cb_area);

#if DEBUG
	if (ECDEBUG) printf("ectdr: ...\n");
#endif
	*cb = cb_template;
	cb->time_domain_reflectometer.command.cmd.cuc = cb_cuc_time_domain_reflectometer;
	cb->time_domain_reflectometer.link_ok = FALSE;
	cb->time_domain_reflectometer.transceiver_cable = FALSE;
	cb->time_domain_reflectometer.ethernet_open = FALSE;
	cb->time_domain_reflectometer.ethernet_short = FALSE;
	cb->time_domain_reflectometer.time = 0;
	cb->time_domain_reflectometer.clear = 0;
	if (eccbstart(ec, cb)) {
		if (cb->time_domain_reflectometer.link_ok)
			ecwarn(ec, warn_link_ok, 0);
		else if (cb->time_domain_reflectometer.ethernet_open)
			ecwarn(ec, warn_ethernet_open, cb->time_domain_reflectometer.time);
		else if (cb->time_domain_reflectometer.ethernet_short)
			ecwarn(ec, warn_ethernet_short, cb->time_domain_reflectometer.time);
		else if (cb->time_domain_reflectometer.transceiver_cable)
			ecwarn(ec, warn_transceiver_cable, cb->time_domain_reflectometer.time);
	}
}


/*
 * Interrupt Routine. Return 0 if no interrupt pending for indicated unit
 * or return 1 if interrupt was pending and handled here (see comments below...).
 */
private int
ecint(unit)
int unit;
{
	ecsi_t		*ec = &ecsi[unit];
	scb_t		*scb = &ec->ecsc->scb;
	scb_int_t	scb_int;

	/* wait for scb ready */
	if (!ecready(ec))
		/* let watchdog restart chip */
		return;

	/* use one snapshot of the interrupt status bits to work with */
	PURGE(&scb->scb.status, sizeof(scb->scb.status));
	scb_int = scb->scb.status.stat;

	/* see if we have any interrupts to handle or if we have a bogus status value */
	if (!scb_int) {
		/*
		 * this may seem weird. we have an interrupt but the chip indicates that there
		 * is no cause for the interrupt. what happens is that the interrupt request
		 * register in ASP gets read to determine the interrupt source and reading that
		 * register will clear it, allowing a consecutive interrupt to latch. the i82596
		 * will drop and re-assert the interrupt line for each internal interrupt request
		 * to trigger edge-triggered interrupt controllers. thus the following scenario
		 * is possible:
		 *
		 * i82596 posts interrupt
		 * system interrupt handler reads/clears the ASP interrupt request register
		 * i82596 posts another interrupt and retriggers ASP 
		 * ec interrupt handler reads i82596 interrupt register and acknowledges/handles both
		 * system interrupt handler reads/clears the ASP interrupt request register
		 * ec interrupt handler reads i82596 interrupt register and find nothing to do
		 *
		 * so we do not have anything to do, but we still return 1 (saying that we did). the
		 * return value is only used for chaining multiple devices. we do not want to upset
		 * the system interrupt handler. we know we are and will always be the only device on
		 * this particular ASP request line anyway.
		 */
		return (1);
	}

	if (*(short *) &scb->scb.status == I82596CORRUPTION) {
		/* we hit an intel chip bug, reset the chip */
		ec->reason.scb_corrupted++;
		if (eccrash) {
			printf("ec: SCB status corrupted\n");
			kdb_break();
		}
		ec->recover = TRUE;
		return (1);
	}

	/* acknowledge these interrupts (note that remaining fields are zero already, see ecready) */
	scb->scb.command.cmd.ack = scb_int;
	FLUSH(&scb->scb.command, sizeof(scb->scb.command));

	/* assert Channel Attention to acknowledge the interrupts */
	ec->ec_csr->ca = 0;

	if ((ec->ec_if.if_flags & IFF_RUNNING) && !ec->recover) {

		/* now handle the interrupts */
		if (scb_int & (SCB_CNA | SCB_CX)) {
			/* command unit not active and command executed interrupt */
			ec->txactive = FALSE;
			ectxint(ec);
			if (ec->q_op)
				/* restart the output, if any */
				ectxsetup(ec);
			else
				/* wake up process that is waiting for non queued operation mode */
				wakeup(ec);
		}

		if (scb_int & SCB_FR) {
			/* receive frame interrupt */
			ecfrint(ec);
		}

		if (scb_int & SCB_RNR) {
			/* receive unit not ready */
#if DEBUG
			panel_led(PANEL_LED_DIAG4, PANEL_LED_TOGGLE);
#endif
			ec->ru_noresources++;
			if (ec->initialized) {
				/* see 'second big fat notice' above: reinitialize */
				ec->reason.ru_not_ready++;
				if (eccrash) {
					printf("ec: RU not ready\n");
					kdb_break();
				}
				ec->recover = TRUE;
			} else 
				ecrxstart(ec);
		}
	}
	return (1);	/* interrupt handled */
}


/*
 * Setup individual address
 */
private boolean_t
ecsetupaddr(ec)
ecsi_t *ec;
{
	u_char	cb_area[sizeof(cb_t) + DCACHE_LINE_SIZE];
	cb_t	*cb = (cb_t *) CACHE_ALIGNED(cb_area);

#if DEBUG
	if (ECDEBUG) printf("ecsetupaddr: %s\n", ether_sprintf(ec->ec_addr));
#endif
	if (INVALIDADDR(ec->ec_addr))
		return FALSE;
	*cb = cb_template;
	cb->ia_setup.command.cmd.cuc = cb_cuc_ia_setup;
	bcopy(ec->ec_addr, cb->ia_setup.address, sizeof(eaddr_t));
	return eccbstart(ec, cb);
}


/*
 * Load multicast list
 */
private boolean_t
ecloadmulticast(ec)
ecsi_t *ec;
{
	u_char	cb_area[sizeof(cb_t) + DCACHE_LINE_SIZE];
	cb_t	*cb = (cb_t *) CACHE_ALIGNED(cb_area);

#if DEBUG
	if (ECDEBUG) printf("ecloadmultcast: %d entries\n", ec->nmc);
#endif
	*cb = cb_template;
	cb->multicast_setup.command.cmd.cuc = cb_cuc_multicast_setup;
	cb->multicast_setup.cnt = ec->nmc * sizeof(eaddr_t);
	bcopy(ec->mc, cb->multicast_setup.mc, cb->multicast_setup.cnt);
	return eccbstart(ec, cb);
}


/*
 * Configure the Intel 82596DX
 */
private boolean_t
eccfg(ec)
ecsi_t *ec;
{
	volatile struct	csr *csr = ec->ec_csr;
	scb_t		*scb = &ec->ecsc->scb;
	unsigned long	pa;
	struct mbuf	*m;
	char		scp_area[sizeof(scp_t) + DCACHE_LINE_SIZE];
	scp_t		*scp = (scp_t *) CACHE_ALIGNED(scp_area);
	iscp_t		iscp;
	int		i, j;
	u_char		cb_area[sizeof(cb_t) + DCACHE_LINE_SIZE];
	cb_t		*cb = (cb_t *) CACHE_ALIGNED(cb_area);

	/* setup SCP (must be on 16-byte boundary) */
	bzero(scp_area, sizeof(scp_area));
	scp->sysbus.csw = scp->sysbus.interrupt = scp->sysbus.trg = scp->sysbus.lock = TRUE;
	scp->sysbus.mode = mode_linear;
	POINT(&scp->iscp_ptr, &iscp);

	/* setup ISCP */
	bzero(&iscp, sizeof(iscp));
	POINT(&iscp.scb_ptr, &scb->scb.status); /* skip padding */
	iscp.busy = 1;

	/* clear SCB */
	bzero(scb, sizeof(*scb));

	/* flush those things out to memory */
	FLUSH(scp, sizeof(*scp));
	FLUSH(&iscp, sizeof(iscp));
	FLUSH(scb, sizeof(*scb));

	/* hand alternate SCP address to the chip */
	ecportaccess(ec, pmap_extract(kernel_pmap, scp), port_scp);

	/* assert channel attention to start initialization process */
	csr->ca = 0;

	/* wait for busy flag to go low (ignore the resulting interrupt) */
	i = 2000;	/* wait 2 seconds max */
	do {
		if (i-- == 0) {
			log(LOG_EMERG, "ec%d: i82596DX failed to respond.\n", ec - &ecsi[0]);
			return FALSE;
		}
		DELAY(1000);	/* wait a msec */
		PURGE(&iscp, sizeof(iscp.busy));
	} while (iscp.busy);

	/* initialialize Control Block template to be a Transmit Control Block */
	bzero(&cb_template, sizeof(cb_template));
	cb_template.transmit.tcb.command.cmd.el = TRUE;				/* only one block on the list */
	cb_template.transmit.tcb.command.cmd.sf = TRUE;				/* use flexible mode */
	cb_template.transmit.tcb.command.cmd.cuc = cb_cuc_transmit;		/* default to transmit command */
	INVALID(&cb_template.transmit.tcb.command.cmd.link);			/* no associated pointer */

	/* initalialize TCB lists */
	for (i = 0; i < NTCBBUF; i++) {

		ec->ecsc->tcb_list[i] = cb_template.transmit;

		if (i == NTCBBUF-1) {
			ec->ecsc->tcb_list[i].tcb.next = &ec->ecsc->tcb_list[0];
			POINT(&ec->ecsc->tcb_list[i].tcb.command.cmd.link, &ec->ecsc->tcb_list[0].tcb.command.cacheline.cb_stat);
		} else {
			ec->ecsc->tcb_list[i].tcb.next = &ec->ecsc->tcb_list[i+1];
			POINT(&ec->ecsc->tcb_list[i].tcb.command.cmd.link, &ec->ecsc->tcb_list[i+1].tcb.command.cacheline.cb_stat);
		}
		/* save the link address so we can dynamically check for trashing */
		ec->ecsc->tcb_list[i].tcb.command.save.link = ec->ecsc->tcb_list[i].tcb.command.cmd.link;

		/* setup TBD list for this TCB */
		POINT(&ec->ecsc->tcb_list[i].tcb.tbd_ptr, ec->tbd_list[i]);
		for (j = 0; j < NTBDBUF - 1; j++) {
			POINT(&ec->tbd_list[i][j].link, &ec->tbd_list[i][j + 1]);
		INVALID(&ec->tbd_list[i][NTBDBUF - 1].link);
		}
	}
	ec->tcb_head = &ec->ecsc->tcb_list[0];
	ec->tcb_tail = &ec->ecsc->tcb_list[NTCBBUF-1];
	FLUSH(ec->ecsc->tcb_list, sizeof(ec->ecsc->tcb_list));
	FLUSH(ec->tbd_list, sizeof(ec->tbd_list));

	/* initialize the RFD lists */
	bzero(ec->ecsc->rfd_list, sizeof(ec->ecsc->rfd_list));
	for (i = 0; i < NRFDBUF; i++) {
		/* Receive Frame Descriptors */
		ec->ecsc->rfd_list[i].rfd.command.cmd.sf = TRUE;				/* use flexible mode */
		if (i == NRFDBUF-1) {
			ec->ecsc->rfd_list[i].rfd.next = &ec->ecsc->rfd_list[0];
			POINT(&ec->ecsc->rfd_list[i].rfd.command.cmd.link, &ec->ecsc->rfd_list[0].rfd.command.cacheline.cb_stat);
		} else {
			ec->ecsc->rfd_list[i].rfd.next = &ec->ecsc->rfd_list[i+1];
			POINT(&ec->ecsc->rfd_list[i].rfd.command.cmd.link, &ec->ecsc->rfd_list[i+1].rfd.command.cacheline.cb_stat);
		}
		/* save the link address so we can dynamically check for trashing */
		ec->ecsc->rfd_list[i].rfd.command.save.link = ec->ecsc->rfd_list[i].rfd.command.cmd.link;
	}
	ec->rfd_head = &ec->ecsc->rfd_list[0];
	ec->rfd_tail = &ec->ecsc->rfd_list[NRFDBUF-1];
	FLUSH(ec->ecsc->rfd_list, sizeof(ec->ecsc->rfd_list));

	/* initialize the RBD lists */
	bzero(ec->ecsc->rbd_list, sizeof(ec->ecsc->rbd_list));
	for (i = 0; i < NRBDBUF; i++) {
		/* Receive Buffer Descriptors */
		ec->ecsc->rbd_list[i].rbd.size = MCLBYTES - RBDOFFSET;	/* max buffer size */
#if DEBUG
		/* keep my swapped pa for debug */
		POINT(&ec->ecsc->rbd_list[i].rbd.pa, &ec->ecsc->rbd_list[i].cmd.stat);
#endif
		if (i == NRBDBUF-1) {
			ec->ecsc->rbd_list[i].rbd.next = &ec->ecsc->rbd_list[0];
			POINT(&ec->ecsc->rbd_list[i].rbd.link, &ec->ecsc->rbd_list[0].cmd.stat);
		} else {
			ec->ecsc->rbd_list[i].rbd.next = &ec->ecsc->rbd_list[i+1];
			POINT(&ec->ecsc->rbd_list[i].rbd.link, &ec->ecsc->rbd_list[i+1].cmd.stat);
		}
		/* save the link address so we can dynamically check for trashing */
		ec->ecsc->rbd_list[i].save.link = ec->ecsc->rbd_list[i].rbd.link;
	}
	ec->rbd_head = &ec->ecsc->rbd_list[0];
	ec->rbd_tail = &ec->ecsc->rbd_list[NRBDBUF-1];
	FLUSH(ec->ecsc->rbd_list, sizeof(ec->ecsc->rbd_list));

	/* enable interrupt handling */
	ec->ec_if.if_flags |= IFF_RUNNING;

	/* try a simple NOP to see if the unit interrupts */
	*cb = cb_template;
	cb->nop.nop.command.cmd.cuc = cb_cuc_nop;

	/* also initialize our special NOP CB (see 'second big fat notice') */
	bcopy(&cb->nop.nop.command.cacheline.cb_stat, &ec->ecsc->nop, sizeof(cb_command_t));
	FLUSH(&ec->ecsc->nop, sizeof(cb_command_t));

	if (!eccbstart(ec, cb)) {
		log(LOG_ALERT, "ec%d: i82596DX failed to interrupt.\n", ec - &ecsi[0]);
		return FALSE;
	}

	/* configure chip to our needs */
	*cb = cb_template;
	cb->nop.nop.command.cmd.cuc = cb_cuc_configure;

	/* first the default settings (from the i82596 user manual) */
	cb->configure.cfg.bytes[ 0] = 0x0e;
	cb->configure.cfg.bytes[ 1] = 0xc8;
	cb->configure.cfg.bytes[ 2] = 0x40;
	cb->configure.cfg.bytes[ 3] = 0x26;
	cb->configure.cfg.bytes[ 4] = 0x00;
	cb->configure.cfg.bytes[ 5] = 0x60;
	cb->configure.cfg.bytes[ 6] = 0x00;
	cb->configure.cfg.bytes[ 7] = 0xf2;
	cb->configure.cfg.bytes[ 8] = 0x00;
	cb->configure.cfg.bytes[ 9] = 0x00;
	cb->configure.cfg.bytes[10] = 0x40;
	cb->configure.cfg.bytes[11] = 0xff;			/* (yes this was wrong previously) */
	cb->configure.cfg.bytes[12] = 0x00;
	cb->configure.cfg.bytes[13] = 0x3f;
	cb->configure.cfg.bytes[14] = 0x00;
	cb->configure.cfg.bytes[15] = 0x00;

	/* now our customizations */
	cb->configure.cfg.field.crc_in_memory = FALSE;		/* we dont need the frame crc in memory */
	cb->configure.cfg.field.no_src_addr_ins = TRUE;	 /* source address is in buffer, do not autoinsert */
	cb->configure.cfg.field.fifo_limit = 0xc;		/* request bus more quickly @ 1/4 space available */

	/* special flags requested by IFP */
	cb->configure.cfg.field.loopback_mode	= (ec->ec_if.if_flags & IFF_LOOPBACK)	? 0x03	: 0x00;	/* external loopback */
	cb->configure.cfg.field.promiscious_mode	= (ec->ec_if.if_flags & IFF_PROMISC)	? TRUE	: FALSE;	/* promiscious mode */
	/*** DONT let intel fool you: mc_all == 0 <==> all multicast addresses are accepted ***/
	cb->configure.cfg.field.mc_all		= (ec->ec_if.if_flags & IFF_ALLMULTI)	? FALSE : TRUE;	/* receive all multicasts */

#if DEBUG
	if (ECDEBUG) printf("eccfg: loopback/promisc/mc_all is %d/%d/%d\n", cb->configure.cfg.field.loopback_mode,
	cb->configure.cfg.field.promiscious_mode, cb->configure.cfg.field.mc_all);
#endif

	/* remember this configuration */
	ec->if_flags = ec->ec_if.if_flags & IFF_CHIPCONFIG;

	return eccbstart(ec, cb);
}


/*
 * Reset of Controller.
 */
private int
ecreset(unit)
int unit;
{
	ecsi_t		*ec;
	volatile struct	csr *csr;
	struct mbuf	*m;
	int		pl, i;

	extern int	iticks_per_usec;

	/* no bogus units please */
	if (unit >= NEC) {
		panic("ecreset: bad unit");
		return EINVAL;
	}

#if DEBUG
	if (ECDEBUG) printf("ecreset: ...\n");
	ecto = 0;
#endif
	ec = &ecsi[unit];
	csr = ec->ec_csr;

	/* initial conditions */
	ec->initialized = FALSE;
	ec->ec_if.if_flags &= ~IFF_OACTIVE;
	ec->ec_if.if_flags &= ~IFF_RUNNING;
	ec->ec_if.if_timer = 0;
	ec->scberrs = FALSE;

	/* if receive timer running, cancel it */
	if (ec->rxtimeout) {
		untimeout(ecrxstart,ec);
		ec->rxtimeout = FALSE;
	}

	/* reset chip */
	ecportaccess(ec, 0, port_reset);

	/* get ROM signature/execute selftest */
	if (ec->signature == 0 || ec->force_selftest) {
		selftest_t selftest_area[sizeof(selftest_t) + DCACHE_LINE_SIZE];
		selftest_t *selftest = (selftest_t *) CACHE_ALIGNED(selftest_area);

		selftest->signature = 0;
		selftest->result.all = ~0;
		FLUSH(selftest, sizeof(*selftest));

		/* execute selftest */
		ecportaccess(ec, pmap_extract(kernel_pmap, selftest), port_selftest);

		i = 2000;	/* wait 2 seconds max */
		do {
			if (i-- == 0) {
				/*
				 * selftest does not complete. this is a side effect if no cable
				 * has been plugged in. we will not say anything cause if the user
				 * has no network, this is just fine and we do not wish to alert him
				 * to a selftest problem in that case. note that the TDR check in
				 * ecinit() will print out a cable diagnostic later anyway. we only
				 * complain because the watchdog could not re-init the chip or if
				 * we know we are not a standalone machine.
				 */
				if (ec->force_selftest) {
					log(LOG_EMERG, "ec%d: i82596DX timeout during selftest.\n", ec - &ecsi[0]);
					if (eccrash) {
						printf("ec: timeout during selftest\n");
						kdb_break();
					}
				}
				goto skip;
				break;
			}
			DELAY(1000);	/* wait a msec */
			PURGE(selftest, sizeof(*selftest));
		} while (selftest->result.of.zeros != 0);

		/* only check selftest results if execution was forced */
		if (ec->force_selftest)
			if (selftest->result.all) {
				/* we did not pass */
				log(LOG_EMERG, "ec%d: i82596DX failed selftest (result is %x).\n", ec - &ecsi[0], selftest->result.all);
				if (eccrash) {
					printf("ec: failed selftest\n");
					kdb_break();
				}
			} else
				/* we did pass selftest, dont do it again */
				ec->force_selftest = FALSE;

		/* get a nonzero checksum so we dont do selftest again just to get the signature */
skip:		ec->signature = (selftest->signature) ? selftest->signature : 0xbad0bad;

#if DEBUG
		if (ECDEBUG) printf("i82596DX signature is %x\n", ec->signature);
#endif

		/* reset chip */
		ecportaccess(ec, 0, port_reset);
	}

	/* no transmit is active */
	ec->txactive = FALSE;
	ec->retransmit = 0;
	ec->pkt_owner = NULL;
	ec->q_op = FALSE;
	ec->recover = FALSE;
	ec->txtimelap = iticks_per_usec * 900000;	/* wait 0.9 seconds before xmit times out */

	/* deallocate transmit buffers */
	for (i = 0; i < NTCBBUF; i++)
		if (ec->ecsc->tcb_list[i].tcb.m) {
			m_freem(ec->ecsc->tcb_list[i].tcb.m);
			ec->ecsc->tcb_list[i].tcb.m = NULL;
		}
	/* make sure these NULL pointers get to memory...*/
	FLUSH(ec->ecsc->tcb_list, sizeof(ec->ecsc->tcb_list));

	/* deallocate receive buffers */
	for (i = 0; i < NRBDBUF; i++)
		if (ec->ecsc->rbd_list[i].rbd.m) {
			m_free(ec->ecsc->rbd_list[i].rbd.m);
			ec->ecsc->rbd_list[i].rbd.m = NULL;
		}
	/* make sure these NULL pointers get to memory...*/
	FLUSH(ec->ecsc->rbd_list, sizeof(ec->ecsc->rbd_list));

	/* drop input queue */
	ecdropq(ec);

	/* release any process in eccbstart() */
	wakeup(ec);
}


/*
 * Initialization of interface.
 */
private int
ecinit(unit)
int unit;
{
	ecsi_t	*ec;
	int	thin_lan, fuse_blown, rval = EIO;
	int	pl;

	/* no bogus units please */
	if (unit >= NEC) {
		panic("ecinit: bad unit");
		return EINVAL;
	}

	ec = &ecsi[unit];

	/* not yet, if address still unknown */
	if (ec->ec_if.if_addrlist == (struct ifaddr *) 0)
		return rval;

#if DEBUG
	if (ECDEBUG) printf("ecinit: unit %d\n", unit);
#endif

	/* take a mutual exclusion lock */
	lock_write(&ec->init_lock);

	/* check fuse */
	if (ec_asp_cfg(&thin_lan, &fuse_blown) && fuse_blown) {
		log(LOG_ALERT, "ec%d: Fuse is blown.\n", unit);
		return rval;
	}

	/* reset the controller */
	ecreset(unit);

	/* initialize queue lock */
	lock_init(&ec->q_lock, TRUE);

	/* configure the 82596 chip */
	if (eccfg(ec) && ecsetupaddr(ec) && ecloadmulticast(ec)) {

		/* wait to allocate receive buffers */
		ecrxsetup(ec, M_WAIT);

		/* post receive */
		ecrxstart(ec);

		/* start any output */
		ectxsetup(ec);

		/* allow watchdog to run */
		ec->ec_if.if_timer = 1;

		/* perform cable diagnosis */
		ectdr(ec);

		/* interface is available */
		ec->ec_if.if_flags |= IFF_UP;

		/* succcess */
		ec->initialized = TRUE;
		rval = ESUCCESS;
	} else
		/* failure to initialize */
		ec->ec_if.if_flags &= ~IFF_RUNNING;

	/* release lock */
	lock_done(&ec->init_lock);

#if DEBUG
	if (ECDEBUG) printf("ecinit: returns %d\n", rval);
#endif

	return rval;
}


/*
 * Process an ioctl request.
 */
private int
ecioctl(ifp, cmd, data)
struct ifnet *ifp;
int cmd;
caddr_t data;
{
	ecsi_t	*ec = &ecsi[ifp->if_unit];
	int	i, error = ESUCCESS;

	/* take a mutual exclusion lock */
	lock_write(&ec->ioctl_lock);

	switch (cmd) {

		case SIOCSIFADDR: {
			struct ifaddr *ifa = (struct ifaddr *) data;

			if (ifp->if_flags & IFF_DEBUG)
				printf("ecioctl: SIOCSIFADDR\n");

			switch (ifa->ifa_addr->sa_family) {

				case AF_INET:
					error = ecinit(ifp->if_unit);	/* before if_arpreq */
					((struct arpcom *)ifp)->ac_ipaddr =
						IA_SIN(ifa)->sin_addr;
#if	NATR                                        
					if (ifp->if_arpreq)
						(*ifp->if_arpreq)((struct arpcom *)ifp, &IA_SIN(ifa)->sin_addr, ((struct arpcom *)ifp)->ac_bcastaddr);
#endif                                        
					break;

				case AF_NS: {
					struct ns_addr *ina = &(IA_SNS(ifa)->sns_addr);
	
					if (ns_nullhost(*ina))
						ina->x_host = *(union ns_host *)(ec->ec_addr);
					else {
						if (INVALIDADDR(ina->x_host.c_host)) {
							error = EINVAL;
							break;
						} else
							bcopy(ina->x_host.c_host, ec->ec_addr, sizeof(eaddr_t));
					}
					if (!(ifp->if_flags & IFF_RUNNING))
						/* bring this interface up */
						error = ecinit(ifp->if_unit);
					else if (!ecsetupaddr(ec))
						error = EIO;
					break;
				}

				default: {
					error = ecinit(ifp->if_unit);
					break;
				}
			}
			break;
		}

		case SIOCSIFFLAGS: {
			struct ifreq *ifr = (struct ifreq *)data;

			if (ifr->ifr_flags & IFF_DEBUG)
				printf("ecioctl: SIOSIFFLAGS to 0x%x\n", ifr->ifr_flags);

			if ((ifr->ifr_flags & IFF_UP) == 0) {
				/* take interface down (so what if we do this many times) */
				ecreset(ifp->if_unit);
				/* do not complain about selftest when we are switched off */
				ec->force_selftest = FALSE;
			} else if ((((ifp->if_flags & IFF_RUNNING) == 0) && (ifr->ifr_flags & IFF_UP)) ||	/* bring this interface up */
			((ifr->ifr_flags & IFF_CHIPCONFIG) != ec->if_flags)) {			/* change configuration of this interface */
				ifp->if_flags = (ifp->if_flags & IFF_CHIPCONFIG) | (ifr->ifr_flags & ~IFF_CHIPCONFIG);
				error = ecinit(ifp->if_unit);
			}
			break;
		}

		case SIOCADDMULTI:
		case SIOCDELMULTI: {
			struct ifreq *ifr = (struct ifreq *)data;

			if (ifp->if_flags & IFF_DEBUG)
				printf("ecioctl: SIO%sMULTI %s\n", (cmd == SIOCDELMULTI) ? "DEL" : "ADD", ether_sprintf(ifr->ifr_addr.sa_data));

			if (ifr->ifr_addr.sa_len != sizeof(eaddr_t)) {
				error = EINVAL;
				goto done;
			}
			if (cmd == SIOCDELMULTI) {
				/* find entry in table and delete it */
				for (i = 0; i < ec->nmc; i++)
					if ((bcmp(ec->mc[i], ifr->ifr_addr.sa_data, sizeof(eaddr_t)) == 0) &&
					(--(ec->nmc) > 0) && (ec->nmc != i)) 
						/* copy last mc address to vacanted slot */
						bcopy(ec->mc[ec->nmc], ec->mc[i], sizeof(eaddr_t));
			} else {
				if (ec->nmc == NMULTICAST) {
					/* no space in our table */
					error = ENOBUFS;
					goto done;
				}
				for (i = 0; i < ec->nmc; i++)
					if (bcmp(ec->mc[i], ifr->ifr_addr.sa_data, sizeof(eaddr_t)) == 0)
						/* we already have this multicast address */
						goto done;
				/* add to our multicast list */
				bcopy(ifr->ifr_addr.sa_data, ec->mc[ec->nmc], sizeof(eaddr_t));
				ec->nmc++;
			}

			if (!(ifp->if_flags & IFF_RUNNING))
				/* bring this interface up */
				error = ecinit(ifp->if_unit);
			else if (!ecloadmulticast(ec))
				error = EIO;
			break;
		}

		default: {
			error = EINVAL;
			break;
		}
	}
done:
	if (error && (ifp->if_flags & IFF_DEBUG))
		printf("ecioctl: returns %d\n", error);

	/* release lock */
	lock_done(&ec->ioctl_lock);

	return error;
}


/*
 * Dump and reinit the chip (separate routine for debugging convenience)
 */
private void
ecrecover(unit)
int unit;
{
	ecsi_t	*ec = &ecsi[unit];
	int	i;

	ec->ec_if.if_flags &= ~IFF_RUNNING;
	ecportaccess(ec, pmap_extract(kernel_pmap, ec->ecsc->dump), port_dump);
	PURGE(ec->ecsc->dump, sizeof(ec->ecsc->dump)); /* so we can see it with kdb */
#if DEBUG
	if (ECDEBUG) printf("ecrecover: ec%d was stuck, i82596 dump @ %x\n", unit, ec->ecsc->dump);
#endif
	/* try to reinit twice if interface has not been shut down */
	for (i = 0; i < 2; i++) {
		if (((ec->ec_if.if_flags & IFF_UP) == 0) || (ecinit(unit) == ESUCCESS))
			return;
		/* force selftest on second pass */
		ec->force_selftest = TRUE;
	}
	if (ec->force_selftest) {
		/*
		 * by now we have reset the chip and tried to execute the selftest which
		 * did not pass. we could panic now, but lets instead warn the user that
		 * he has lost his network and allow him to shutdown gracefully.
		 */
		log(LOG_EMERG, "ec%d: interface has malfunctioned and is unavailable.\n", ec - &ecsi[0]);
		log(LOG_EMERG, "ec%d: system shutdown and diagnosis recommended.\n", ec - &ecsi[0]);
	}
}


/*
 * Watchdog to keep eye on i82596 operation.
 */
private int
ecwatchdog(unit)
int unit;
{
	ecsi_t		*ec = &ecsi[unit];
	scb_t		*scb = &ec->ecsc->scb;
	volatile struct	csr *csr = ec->ec_csr;
	int		pl;

	/* we only have work to do if the interface is up */
	if ((ec->ec_if.if_flags & IFF_UP) == 0)
		return;

	/* count # of seconds we are doing this */		
	ec->dogwalks++;

	pl = splimp();

	/* check completion bit in our special nop CB (see 'second big fat notice') */
	PURGE(&ec->ecsc->nop, sizeof(short));
	if (*(short *) &ec->ecsc->nop & 0x8000) {
		ec->reason.nop_completed++;
		if (eccrash) {
			printf("ec: NOP cb completed\n");
			kdb_break();
		}
		ec->recover = TRUE;
	}

	/* check if a transmit interrupt is expected */
	if (ec->tcb_head != ec->tcb_tail->tcb.next || ec->txactive) {
		/* see if any more transmits have completed */
		ectxint(ec);
		/* now check timestamp of last completion */
		if ((it_read() - ec->txtimestamp) > ec->txtimelap) {
			ec->reason.xmit_timeout++;
			if (eccrash) {
				printf("ec: transmit timeout\n");
				kdb_break();
			}
			ec->recover = TRUE;
		}
	}

	/* see if SCB has cleared */
	ecready(ec);

	if (ec->recover) {
		/* we have to reinitialize the chip */
		ecrecover(unit);
#define MINTXERR 4
	} else if ((ec->request_tdr >= MINTXERR) || ec->last_warning) {
		/*
		 * transmits are failing or we have outstanding network trouble.
		 * perform TDR, but refrain from other checks. if the user has
		 * a standalone node, there is no need to try to reset the chip
		 * every couple of seconds. note that the smaller the above value
		 * of MINTXERR, the sooner we will look for even the smallest
		 * disturbences on the cable.
		 */
		ectdr(ec);
		ec->request_tdr = 0;
	}

	/*
	 * record the number of frames lost due to 'no resources' or 'overrun'. we
	 * let the watchdog do this instead of configuring the chip to 'save bad
	 * frames'.
	 */
#define ADDCNTR(new,old)	{ \
	swap_t val = new; \
	if (val.low != old.low || val.high != old.high) { \
		int add = ((val.high << 16)|val.low) - ((old.high<<16)|old.low); \
		if (add > 0 && add < 1000 && ec->scberrs) { \
			ec->ec_if.if_ierrors += add; \
			if (ec->ec_if.if_flags & IFF_DEBUG) \
				printf("watchdog: %d if_ierrors\n", add); \
		} \
		old = val; \
	} \
}

	if (ec->initialized) {
		ADDCNTR(ec->ecsc->scb.scb.ovrnerrs, ec->ovrnerrs);
		ADDCNTR(ec->ecsc->scb.scb.rscerrs,	ec->rscerrs );
		/* we have now synched the error counters */
		ec->scberrs = TRUE;
	}

	/* retrigger in one second */
	ec->ec_if.if_timer = 1;
	splx(pl);
}


/*
 * Controller exists: make available by filling in network interface
 * record.	System will initialize the interface when it is ready
 * to accept packets.
 */

ecattach(hp_devinfo_t *devinfop)
{
	ecsi_t			*ec;
	struct			ifnet *ifp;
	int			i;
	int			unit;
	int			thin_lan, fuse_blown;
	char			*port;
	volatile struct		csr *csr;

	static ihandler_t	me;

	extern unsigned long	dcache_line_size;
	extern eaddr_t		ec_station_address;

	unit = devinfop->devnum;
	ec = &ecsi[unit];
	ifp = &ec->ec_if;

	/*
	 * the code for the on-board unit relies on knowing the data cache line size
	 * to position type definitions across cache line boundaries. with some extra
	 * work it should be possible to adapt the code to deal with varying line
	 * sizes and assume the largest one for DCACHE_LINE_SIZE.
	 */
	if ((unit == 0) && (dcache_line_size != DCACHE_LINE_SIZE))
		panic("ecattach: ec driver not setup for this cache line size");

	/* tie hardware registers to software structure */
	ec->ec_csr = (struct csr *) devinfop->vcsrs[0];	/* devinfo.vcsrs filled in by probe() */

	/* pickup lan address */
	bcopy(ec_station_address, ec->ec_addr, sizeof(eaddr_t));

	ec->en_ac.ac_arphrd = ARPHRD_ETHER;
	ec->ec_bcastaddr	= etherbroadcastaddr;

	ifp->if_unit	= unit;
	ifp->if_name	= "ec";
	ifp->if_mtu	= ETHERMTU;
	ifp->if_type	= IFT_ETHER;
#ifdef	OSC1_0        
	ifp->if_hwaddr	= ec->ec_addr;
#endif        
	ifp->if_addrlen	= 6;
	ifp->if_hdrlen	= sizeof (struct ether_header) + 8;	/* 8 for SNAP */

	ifp->if_init	= ecinit;
	ifp->if_ioctl	= ecioctl;
	ifp->if_start	= ectxsetup;
	ifp->if_output	= ether_output;	/* from net/if_ethersubr.c */
	ifp->if_reset	= ecreset;

#if	NATR
	ifp->if_arpreq	= ether_arprequest;	/* from netinet/if_ether.c */
	ifp->if_arpin	= ether_arpinput;	/* from netinet/if_ether.c */
	ifp->if_arpout	= ether_arpoutput;	/* from netinet/if_ether.c */
	ifp->if_arpres	= ether_arpresolve;	/* from netinet/if_ether.c */
	ifp->if_arpctl	= ether_arpioctl;	/* from netinet/if_ether.c */
#endif
	ifp->if_timer	= 0;
	ifp->if_done	= 0;
	ifp->if_watchdog = ecwatchdog;

	/* trailers only set us back, so avoid them if we can */
	ifp->if_flags = IFF_BROADCAST | IFF_NOTRAILERS | IFF_SIMPLEX;

	if_attach(ifp);

	/* initialize locks */
	lock_init(&ec->init_lock, TRUE);
	lock_init(&ec->ioctl_lock, TRUE);

	/* setup pointer to a page with the lower 16 bits set to 0x4000 */
	ec->ecsc = (ecsc_t *) (ecsc_area + unit * sizeof(ecsc_t));

	/* indicate no receive buffers have been allocated */
	for (i = 0; i < NRBDBUF; i++)
		ec->ecsc->rbd_list[i].rbd.m = NULL;

	/* plug in interrupt handler */
	me.ih_resolver = devinfop->parent_ctlrinfop->ctlr_businfop->resolver;
	me.ih_flags = IH_VEC_MULTIPLE_OK;
	me.ih_state = 0;
	me.ih_busnum = devinfop->parent_ctlrinfop->ctlr_businfop->busnum;
	GETSPL(me.ih_spl, splnet); /* XXX use spl specified in devinfo? */

	me.ih_priority = 0;
	me.ih_stats.intr_type = INTR_DEVICE;
	me.ih_stats.intr_cnt = 0;
	me.ih_rparam[0].intparam = 0;	/* param to resolver */
	me.ih_hparam[0].intparam = unit; /* param to handler */
	me.ih_handler = ecint;
	me.ih_level = devinfop->irq;
	
	if ((ec->int_id = handler_add(&me)) == NULL) {
		panic("ec: cannot register interrupt handler");
	}

	/* enable interrupt handling */
	handler_enable(ec->int_id);

	/* reset the controller */
	ecreset(unit);

	/* inform user of ThinLAN/AUI selection */
	if (!ec_asp_cfg(&ec->thin_lan, &fuse_blown))
		port = "UNKNOWN";
	else
		port = ec->thin_lan ? "ThinLAN" : "AUI";

	printf("ec%d: %s port, MAC address %s.\n", unit, port, ether_sprintf(ec->ec_addr));

	if (INVALIDADDR(ec->ec_addr))
		panic("ecattach: invalid individual address");
}


/*
 * Probe/Attach entries called by autoconfigure.
 */

ecprobe(hp_devinfo_t *devinfop)
{
	/* at present, only support single ether chip on motherboard */

	if (devinfop->devnum != 0)
		return (0);	/* don't try attaching... */

	if (devinfop->parent_ctlrinfop->ctlr_businfop->bustype != BUSTYPE_SYS)
		return (0);	/* only know how to deal with sysbus */

	/* check on the CSRs */
	if (devinfop->numcsrs != 1)
		panic("ec: unexpected number of CSRs");

	if (!devinfop->vcsrs[0])
		panic("ec: CSRs not properly mapped");

	return(1);
}

/* this structure lets probeio() know how to find us */
hp_driverinfo_t ecdriverinfo = {ecprobe, 0, ecattach, "ec"};


#if MACH_KDB
/*
 * for debugging, only if KDB is included, executed by $he command
 */
ectrace()
{
	short	*s;
	ecsi_t	*ec = &ecsi[0];
	int	i;

	PURGE(ec->ecsc, sizeof(ecsc_t));
	printf("ecsi@%x ecsc@%x rfd_list@%x(%d) rbd_list@%x(%d) tcb_list@%x(%d)\n",
		ecsi, ec->ecsc, ec->ecsc->rfd_list, NRFDBUF, ec->ecsc->rbd_list, NRBDBUF, ec->ecsc->tcb_list, NTCBBUF);
	printf("rfd=%x/%x rbd=%x/%x tcb=%x/%x\n", ec->rfd_head, ec->rfd_tail,
		ec->rbd_head, ec->rbd_tail, ec->tcb_head, ec->tcb_tail);
	printf("txactive=%d q_op=%d txsetup=%d\n", ec->txactive, ec->q_op, ec->txsetup);
	if (ec->ru_noresources)
		printf("ru_noresources=%d\n", ec->ru_noresources);
	if (ec->nretransmit)
		printf("nretransmit=%d\n", ec->nretransmit);
	if (ec->ntbd_shortages)
		printf("ntbd_shortages=%d\n", ec->ntbd_shortages);
	if (ec->ufotransmit)
		printf("ufotransmit=%d\n", ec->ufotransmit);
	printf("recover: %d %d %d %d %d %d %d (scbn/r, tcb, scb, nop, xmit, rfd, ru)\n",
		ec->reason.scb_not_ready, ec->reason.scb_corrupted, ec->reason.tcb_corrupted,
		ec->reason.nop_completed, ec->reason.xmit_timeout, ec->reason.rfd_not_completed,
		ec->reason.ru_not_ready);
	printf("xmit: %d %d %d %d %d %d (lcol, nocr, ncts, dmau, defr, tcol)\n",
		ec->xmit_status.late_collision, ec->xmit_status.no_carrier_sense,
		ec->xmit_status.lost_clear_to_send, ec->xmit_status.dma_underrun,
		ec->xmit_status.deferred, ec->xmit_status.too_many_collisions);
#define UNSWAP(x) ((long) ((0xffffL & (x.low)) | (x.high<<16)))
	printf("scb: %d %d (rsc, ovn)\n", UNSWAP(ec->rscerrs), UNSWAP(ec->ovrnerrs));
	printf("ifq: %d %d %d (oerrors, qdrops, ierrors)\n", ec->ec_if.if_oerrors, ec->ec_if.if_snd.ifq_drops, ec->ec_if.if_ierrors);
	/* forward link checks */
	for (i = 0; i < NRFDBUF; i++)
		if (LINKMISMATCH(ec->ecsc->rfd_list[i].rfd.command.save.link, ec->ecsc->rfd_list[i].rfd.command.cmd.link))
			printf("bad link address in RFD @ %x\n", &ec->ecsc->rfd_list[i]);
	for (i = 0; i < NRBDBUF; i++)
		if (LINKMISMATCH(ec->ecsc->rbd_list[i].save.link, ec->ecsc->rbd_list[i].rbd.link))
			printf("bad link address in RBD @ %x\n", &ec->ecsc->rbd_list[i]);
	for (i = 0; i < NTCBBUF; i++)
		if (LINKMISMATCH(ec->ecsc->tcb_list[i].tcb.command.save.link, ec->ecsc->tcb_list[i].tcb.command.cmd.link))
			printf("bad link address in TCB @ %x\n", &ec->ecsc->tcb_list[i]);
}
#endif MACH_KDB

#endif NEC>0

/*
 * we need a special page with the lower 16 bits of the PPN set to 0x4000 to workaround
 * the intel corruption bug (see 'second big fat notice' above). we grab this one early.
 */
ec_special_page(start, end, va)
vm_offset_t start;		/* First available physical address */
vm_offset_t end;		/* Last available physical address + 1 */
vm_offset_t *va;		/* First available virtual address */
{
#if NEC>0
	/*
	 * we grab a special page with the lower 16 bits of the PPN set to 0x4000
	 * and whatever we need beyond it (for as many ecsc_t areas as we support
	 * ethernet controllers), as long as higher order bits do not increment.
	 */
	vm_offset_t	pa;

#define	SPECIAL_SIZE	(NEC * sizeof(ecsc_t))
	
	/* setup pointer to a page with the lower 16 bits set to 0x4000 */
	pa = ((start + 0x10000) & ~0xffff) | 0x4000;
	if (pa < start)
		panic("ec: ecsc address too low");
	if ((pa + SPECIAL_SIZE) >= end)
		panic("ec: ecsc address too high");
	if ((pa & ~0xffff) != ((pa + SPECIAL_SIZE) & ~0xffff))
		/*
		 * see 'second big fat notice' above. since the forward link addresses
		 * of the transmit command blocks can be truncated (low 16 bits sets to
		 * 0x4000) we keep all transmit command blocks in memory following the
		 * magic 'nop' in ecsc_t with the high 16 bits unchnged.
		 */
		panic("ec: ecsc area spans 64Kb boundary");
		
	/* use these physical pages are for us */
	pmap_define_hole(pa, pa + SPECIAL_SIZE);
        ecsc_area = *va;
        *va = pmap_map(ecsc_area, pa, pa + SPECIAL_SIZE, (VM_PROT_READ|VM_PROT_WRITE));
#endif
}
