/*
 *  (c) Copyright 1992 HEWLETT-PACKARD COMPANY
 *
 *  To anyone who acknowledges that this file is provided "AS IS"
 *  without any express or implied warranty:
 *      permission to use, copy, modify, and distribute this file
 *  for any purpose is hereby granted without fee, provided that
 *  the above copyright notice and this notice appears in all
 *  copies, and that the name of Hewlett-Packard Company not be
 *  used in advertising or publicity pertaining to distribution
 *  of the software without specific, written prior permission.
 *  Hewlett-Packard Company makes no representations about the
 *  suitability of this software for any purpose.
 */

/*
 * INTEL 82596DX chip LAN driver
 */
#include <ec.h>
#if NEC > 0

#include <mach_kdb.h>

#include <sys/types.h>
#include <chips/busses.h>
#include <device/device_types.h>
#include <device/io_req.h>
#include <device/if_ether.h>
#include <device/if_hdr.h>
#include <device/net_io.h>
#include <kern/lock.h>
#include <mach/mach_types.h>
#include <vm/pmap.h>
#include <machine/endian.h>

#include <hpdev/device.h>
#include <hpsgc/mach_glue.h>
#include <machine/spl.h>
#include <machine/Jasp.h>

/*
 * Driver Parameters:
 *
 * RBDOFFSET: Received data is DMA'ed at this offset into a
 * naturally-aligned buffer so that fields in the IP header are
 * naturally aligned.  Thus, we do not have to tweak TCP/IP source
 * code in order to handle unaligned data.  Since we do not take
 * advantage of trailer packets, this is a simple, no-cost solution.
 *
 * NBUF: Max. no. of packets that can be held in our Receive Frame
 * Area (RFA) buffer.
 *
 * NRFDBUF: No. of Receive Frame Descriptor (RFD) buffers.  Each
 * received packet uses 1 such descriptor.
 *
 * NRBDBUF: No. of Receive Buffer Descriptor (RBD) buffers.  A
 * received packet may use several such descriptors.
 *
 * NTBDBUF: No. of Transmit Buffer Descriptor (TBD) buffers per TCB.
 * There must be at least 2 of these, and at most 3 are necessary.
 *
 * NTCBBUF:  No. of Transmit Command Block (TCB) buffers.
 *
 * NMULTICAST: Max. no. of addresses in a multicast list.  This is a
 * driver limit, since the chip actually accepts up to 16384-Byte
 * lists.
 */
#define RBDOFFSET	2
#define NBUF		160
#define NRFDBUF		NBUF
#define NRBDBUF		(2*NRFDBUF)
#define NTCBBUF		8
#define NTBDBUF		3
#define NMULTICAST	20

#define DCACHE_LINE_SIZE 32	/* max. dcache line size that we can handle */

#include "ec596reg.h"
#include "ec596var.h"

#ifdef KGDB
    extern int kgdb_enabled;
#   define KDB_BREAK(m)	{if (kgdb_enabled && eccrash) kgdb_break();}
#else
# ifdef MACH_KDB
    extern int kdbenabled;
#   define KDB_BREAK(m)	{if (kdbenabled && eccrash) kdb_break(m);}
# else
#   define KDB_BREAK(m)
# endif
#endif

#ifdef DEBUG
# define ECNMBUF NBUF

# define EC_DEBUG_PRINT 1
# define EC_DEBUG_CHECK 2
# define EC_DEBUG_PRINTALL 4
# define EC_DEBUG_RX 8
# define EC_DEBUG_TX 16

  int ecdebug = ~0;
  io_req_t ecsend[ECNMBUF];
  ipc_kmsg_t ecreceive[ECNMBUF];		/* I/O requests */
  int ec4all, ec4any, ec4me, ec4multi, ec4other, ec4who, ecsindex;
  int ecrindex, ecto;
#endif /* DEBUG */

#define CHANNEL_ATTN(u)		{((struct core_lan *)ecaddr[u])->attn = 0;}

#define PURGE(x,y)	pmap_purge_range(kernel_pmap,(x),(y))
#define FLUSH(x,y)	pmap_flush_range(kernel_pmap,(x),(y))

/*
 * it_read() provides access to a high-resolution timer in the PA-RISC
 * processor, with minimal overhead.  The timer is continually counting
 * up by 1 at a rate which is implementation-dependent and between twice
 * the "peak instruction rate" and half the "peak instruction rate",
 * according to the Architecture Reference Manual.
 * For the 720's it's incremented once every 1/50th usec and for the
 * 730's and 750's it's incremented once every 1/66th usec.
 */
#ifdef __GNUC__
# define it_read()	({u_int r; asm volatile("mfctl 16,%0" : "=r"(r)); r;})
#else
# define it_read()	mfctl(CR_ITMR)
#endif /* __GNUC__ */

/*
 * Convert virtual- to physical address, swap halfwords, and store in
 * a halfword-aligned variable.
 */
#ifdef DEBUG
# define POINT(p,va) \
{ \
   unsigned long _pa; \
   extern vm_offset_t avail_end; \
   _pa = pmap_extract(kernel_pmap, va); \
   if (!_pa) { \
       panic("ec: vtop failed"); \
   } /* if */ \
   if (_pa > avail_end) { \
       panic("ec: illegal physical address"); \
   } /* if */ \
   (p)->low = _pa; \
   (p)->high = _pa >> 16; \
}
#else
# define POINT(p,va) \
{  \
    unsigned long _pa; \
    _pa = pmap_extract(kernel_pmap, va); \
    if (!_pa) { \
	panic("ec: vtop failed"); \
    } /* if */ \
    (p)->low = _pa; \
    (p)->high = _pa>>16; \
} 
#endif
#define SET_EC596_NIL(p)		(((p)->low = (long)EC596_NIL_PTR) & 0xffff, \
				  (p)->high = (long)EC596_NIL_PTR >> 16)
#define LINKMISMATCH(a,b)	((a).low != (b).low || (a).high != (b).high)
#define ECADDRMATCH(a,b)	(!strncmp((a), (b), sizeof(eaddr_t)))

/*
 * This macro tests for a valid individual address.  Note that the
 * least-significant bit of a byte is transmitted first, so the
 * group/individual address bit, which is the first bit on the wire,
 * is the lsb of the 1st address byte.
 */
#define INVALIDADDR(a)(*(a)&1)

/*
 * SECOND BIG FAT NOTICE:
 *
 * Due to a subtle but critical hardware bug in the i82596 Step A1
 * chip revision (supposedly to be fixed in Step B), this driver
 * contains "work-around" code to detect and/or prevent trashing of
 * main-memory data structures used by the chip.  It is important to
 * read the INTEL erratum sheet describing this bug, but here is a
 * summary: The problem concerns chip microcode for one of the DMA
 * engines, which writes Status information to main memory. Under
 * certain timing conditions, 2 consecutive "write"s to the same DMA
 * channel may be collapsed during a 1-cycle window in end-of-transmit
 * processing, producing an erroneous "write".  If the first operation
 * should write A to ADDR1 and B to ADDR2, then the error is either a:
 *
 * 1. "write" of B to ADDR1, or a
 * 2. "write" of B to ADDR2 + <size of A>, or a
 * 3. missing "write", or a
 * 4. combination of 1 and 2.
 *
 * In the INTEL erratum sheet, all consecutive double "write"s are
 * examined and a work-around is given.  If a sequence that would
 * trigger the bug appears, we re-initialize the chip and start over.
 * This overhead seems acceptable, since the triggering conditions are
 * not expected to appear often.  In this driver, the work-around is
 * implemented as follows.
 *
 * [Transmit path]
 *
 * The bug's basic effect in this case is that a "write" to set the
 * Command Block's "busy" bit can end up instead going to:
 *
 * 1. the SCB Status word, or
 * 2. the least-significant address bits of the Command Block's forward-link
 *    pointer, or
 * 3. a "write" to the SCB Status word ends up in the SCB command word.
 *
 * In case 1, we would see 0x4000 in the SCB Status word.  Since this
 * implies that the chip is idle but there is no CNR interrupt (only a
 * FR interrupt), this is a bogus interrupt, so we just check for this
 * value and throw it away.  (Even if we did process a FR interrupt,
 * it would do no damage.)  Any interrupt masked by the bogus status
 * field will be reposted by the chip next time.  In case 2, we would
 * end up going to the wrong address for the next Command Block, which
 * could be disastrous.  Since the least-significant bits of the link
 * address are forced to 0x4000, we simply make a virtue of necessity
 * by allocating all of our Command Blocks in memory pages whose
 * least-significant address bits are 0x4000 and by placing a "no
 * operation" Command Block at the beginning of the page with its "EL"
 * bit set.  Thus, if trashing occurs, the chip will harmlessly
 * execute a "no operation" command instead of the planned chain of
 * "transmit" commands.  This situation can be detected by examining
 * the "C" bit of the "no operation" Command Block.  In case 3, we are
 * safe because none of the possible, expected values of the SCB
 * Status word cause a problem if issued as an SCB command word.
 *
 * Due to another chip bug, the transmit engine may hang if a late
 * collision occurs during transmission.  We must detect this by
 * "timing out" on any pending transmission.  The old detection trick
 * of issuing "resume" commands to the SCB is aggravating this DMA
 * bug, so we use time-outs instead.
 *
 * [Receive path]
 *
 * Since end-of-receive processing has many consecutive DMA writes
 * that could be collapsed, there are many failure scenarios.  Since
 * it is impractical to ex- plicitly test for each possible case, this
 * driver instead tries to check only when things don't go as planned.
 * It seems sufficient to only check whether the RFD "completion" bit
 * has not been set or gets set by an erroneous "write" of the
 * least-significant 16 bits of the RBD pointer field.  Thus, our
 * tactic is that when we get an FR interrupt and there is *no* RFD
 * whose "completion" bit has a valid setting, then we look ahead to
 * the next RFD.  If that one has its "completion" bit set, then we
 * assume it to be corrupted and reinitialize.  If we run out of
 * "receive" buffers, we also reinitialize the chip in order to handle
 * cases in which the "completion" bit was set but other fields were
 * also damaged.  The following test macro assumes that the RFD Status
 * word should never end in 0xe.
 */
#define I82596CORRUPTION	0x4000
#define I82596RFDVALID(r)	(((r)->rfd.command.cacheline.cb_stat&15) != 14)

extern unsigned long	dcache_line_size;
extern int		net_queue_free_min;

extern void		delay(int usec);
extern void		net_kmsg_more(void);

static char	*ecname = "ec";	/* driver name */
static ecsi_t	ecsi[NEC];
static caddr_t	ecaddr[NEC];
static boolean_t ecalive[NEC];
static cb_t	cb_template;	/* initial CB value */
static ecsc_t	*ecsc_area;	/* ->spec. page with PPN{16..31} == 0x4000 */
#if defined(MACH_KDB) || defined(KGDB)
  int	eccrash;	/* <=> panic instead of silently recover for debug. */
#endif


static boolean_t ec596rxsetup (ecsi_t *ec);
static int	ec596intr (int unit);


/*
 * Access chip through port interface.
 */
static void
ec596portaccess(int unit, unsigned long pa, iport_t cmd)
{
    volatile struct core_lan *csr = (struct core_lan *)ecaddr[unit];

    csr->port = pa >> 16;
    /* keep at least 1 cycle between "write"s: */
    delay(1);
    csr->port = pa & 0xfff0 | cmd;
    /* give chip 20 clocks + 0.5 us to get its act together: */
    delay(100);
} /* ec596portaccess */


/*
 * ec596ready == true <=> SCB ready to accept a command.
 */
static boolean_t
ec596ready(ecsi_t *ec)
{
    if (!ec->recover) {
        int us;
        scb_command_t *cmd = &ec->ecsc->scb.scb.command;
	/*
	 * Wait until the SCB is ready to accept the next command. The
	 * chip may delay acknowledging a command written to the SCB
	 * because it is busy handling higher-priority tasks (e.g.,
	 * prefetching buffers, or finishing frame reception), which
	 * should not take long, and it will soon clear the SCB
	 * Command word.  We must wait until then in order to avoid
	 * interfering with a prior command.  If the SCB does not
	 * clear within 100 us, the chip is hung, so we return FALSE.
	 */
        PURGE(cmd, sizeof(*cmd));
        if (!cmd->all) {
	    return TRUE;
	} /* if */
        us = 100;
        do {
            delay(1);
            PURGE(cmd, sizeof(*cmd));
            if (!cmd->all) {
		return TRUE;
	    } /* if */
	} while (--us > 0);
#ifdef DEBUG
        if (ecdebug & EC_DEBUG_PRINT) {
	    printf("ec596ready: SCB never cleared\n");
	} /* if */
#endif
        ++ec->reason.scb_not_ready;
	KDB_BREAK("ec596ready: SCB never cleared");
        ec->recover = TRUE;
    } /* if */
    return FALSE;
} /* ec596ready */


/*
 * Start the Receive Unit.
 */
static boolean_t
ec596rxstart(ecsi_t *ec)
{
    int s = splimp();

    if (ec->rxtimeout) {
	/* cancel restart-timer */
#ifdef DEBUG
        if (ecdebug & EC_DEBUG_RX) {
	    printf("ec596rxstart: time-out\n");
	} /* if */
#endif
        untimeout(ec596rxstart, ec);
        ec->rxtimeout = FALSE;
    } /* if */
    if (!ec596rxsetup(ec)) {
	/* try again in 1 second */
#ifdef DEBUG
        if (ecdebug & EC_DEBUG_RX) {
	    printf("ec596rxstart: no buffers\n");
	} /* if */
#endif
        ec->rxtimeout = TRUE;
        timeout(ec596rxstart, ec, hz);
        splx(s);
        return FALSE;
    } /* if */
    if (ec596ready(ec)) {
	scb_t *scb;
    
	/*
	 * Tell SCB to accept a CB.  Note that the remaining SCB
	 * command bits are already zero (see "ec596ready()").
	 */
	scb = &ec->ecsc->scb;
        scb->scb.command.cmd.ruc = scb_ruc_start;
        POINT(&scb->scb.rfa_ptr, &ec->rfd_head->rfd.command.cacheline.cb_stat);
        FLUSH(scb->cacheline.cmd, sizeof(scb->cacheline.cmd));
	/* assert chan attention: */
        CHANNEL_ATTN(ec - ecsi);
    } /* if */
    splx(s);
    thread_wakeup((int)ec+1);	/* wake up any thread in "ec596init()" */
    return TRUE;
} /* ec596rxstart */


/*
 * Start the Command Unit (call at "splimp()"!).
 */
static void
ec596custart(ecsi_t *ec, cb_t *cb)
{
    if (ec596ready(ec)) {
	scb_t *scb;
	/*
	 * Tell SCB to accept a CB.  Note that the remaining SCB
	 * command bits are already zero (see "ec596ready()").
	 */
	scb = &ec->ecsc->scb;
        scb->scb.command.cmd.cuc = scb_cuc_start;
        POINT(&scb->scb.cbl_ptr, &cb->nop.nop.command.cacheline.cb_stat);
        FLUSH(scb->cacheline.cmd, sizeof(scb->cacheline.cmd));
        CHANNEL_ATTN(ec - ecsi);
    } /* if */
} /* ec596custart */


/*
 * Warn user about cable condition (once).
 */
static void
ec596warn(ecsi_t *ec, warning_t warning, int tdr)
{
    int unit = ec - ecsi;

#ifdef DEBUG
    if (ecdebug & EC_DEBUG_PRINT) {
	printf("ec596warn: warning %d\n", warning);
    } /* if */
#endif
    /*
     * The TDR value increments at a 10MHz rate.  Convert it to
     * distance using an assumed cable velocity factor of 0.78 for
     * thick LAN and 0.66 for thin LAN, according to the formula:
     *
     * <distance> = 1/2*<velocity factor>*<speed of light>*<ticks>/<freq>
     *
     * This yields about 11.7 m/tick in thick LAN and 9.9 m/tick in
     * thin LAN.
     */
    if (tdr) {
	tdr = tdr * (ec->thin_lan ? 99 : 117) / 10;
    } /* if */
    if (warning != ec->last_warning) {
	ec->last_warning = warning;
	switch (warning) {
          case warn_link_ok:
            printf("%s%d: Transmission line access is restored\n",
		   ecname, unit);
            break;
          case warn_ethernet_open:
            printf(
	       "%s%d: Transmission line is improperly terminated %dm away\n",
		   ecname, unit, tdr);
            break;
          case warn_ethernet_short:
            printf("%s%d: Transmission line has short circuit %dm away\n",
		   ecname, unit, tdr);
            break;
          case warn_transceiver_cable:
            printf("%s%d: Transceiver cable is missing or inoperative\n",
		   ecname, unit);
            break;
          case warn_cb_timeout:
            printf("%s%d: Command timeout, check transmission line access\n",
		   ecname, unit);
            break;
          default:
	    printf("%s%d: Unknown TDR status\n", ecname, unit);
        } /* switch */
    } /* if */
} /* ec596warn */


static void
ec596cbtimeout(ecsi_t *ec)
{
    /*
     * Time-out during a pending chip operation.  Apart from hardware
     * failure, this is usually caused by a loose cable.
     */
    if (ec->ec_if.if_flags & IFF_RUNNING) {
	ec596warn(ec, warn_cb_timeout, 0);
    } /* if */
    ec->cbabort = TRUE;
    thread_wakeup((void *)ec);
} /* ec596cbtimeout */


/*
 * Check if interface is worth waiting for(!)
 */
static boolean_t
ec596cbabort(ecsi_t *ec)
{
    if (ec->cbtimeout) {
	/* cancel time-out */
        untimeout(ec596cbtimeout, ec);
        ec->cbtimeout = FALSE;
    } /* if */
    if (!(ec->ec_if.if_flags & IFF_RUNNING)) {
	ec->cbabort = TRUE;
    } /* if */
    if (ec->cbabort) {
	return TRUE;		/* interface down or timed out */
    } /* if */
    ec->cbtimeout = TRUE;	/* set new 2-second time-out */
    timeout(ec596cbtimeout, ec, 2*hz);
    return FALSE;
} /* ec596cbabort */


/*
 * Start a Command Block NOW!
 */
static boolean_t
ec596cbstart(ecsi_t *ec, cb_t *cb)
{
    int s;
    boolean_t status;

    /*
     * Suspend normal transmission queueing and issue a command to the
     * chip.  Note that the hardware design can cause us to hang here
     * indefinitely if the cable has fallen out of the connector and,
     * in that case, we cannot even do time-domain reflectometry
     * because the "Transmit" command will fail. Thus, we must use a
     * time-out.
     */
    FLUSH(cb, sizeof(*cb)); 	/* flush Command Block to main memory */
    lock_write(&ec->q_lock);
    s = splimp();
    ec->cbabort = FALSE;
    ec->q_op = FALSE;		/* wait for queued mode to finish */
    while (!ec596cbabort(ec) && ec->txactive) {
        assert_wait((int)ec, FALSE);
        thread_block((void (*)())0);
    } /* while */
    if (!ec->cbabort) {
        ec596custart(ec, cb); /* issue Command Block to chip, to execute */
	/* await completion of this CB: */
        while (!ec596cbabort(ec)) {
            PURGE(&cb->nop.nop.command.cacheline.cb_stat,
		  sizeof(cb->nop.nop.command.cacheline.cb_stat));
            if (cb->nop.nop.command.cmd.completion) {
		break;
	    } /* if */
            assert_wait((int)ec, FALSE);
            thread_block((void (*)())0);
	} /* while */
    } /* if */
    if (ec->cbtimeout) {
	/* cancel time-out */
        untimeout(ec596cbtimeout, ec);
        ec->cbtimeout = FALSE;
    } /* if */
    ec->q_op = TRUE;		/* restore queued-operation mode */
    status = !ec->cbabort && cb->nop.nop.command.cmd.ok;
    splx(s);
    lock_done(&ec->q_lock);
    ec596txsetup(ec - ecsi);	/* resume transmission */
    return status;
} /* ec596cbstart */


/*
 * Do Time-Domain Reflectometry to get cable status.
 */
static void
ec596tdr(ecsi_t *ec)
{
    unsigned char cb_area[sizeof(cb_t) + DCACHE_LINE_SIZE];
    cb_t *cb = (cb_t *)CACHE_ALIGNED(cb_area);

#ifdef DEBUG
    if (ecdebug & EC_DEBUG_PRINT) {
	printf("ec596tdr: ...\n");
    } /* if */
#endif
    *cb = cb_template;
    cb->time_domain_reflectometer.command.cmd.cuc =
      cb_cuc_time_domain_reflectometer;
    cb->time_domain_reflectometer.link_ok = FALSE;
    cb->time_domain_reflectometer.transceiver_cable = FALSE;
    cb->time_domain_reflectometer.ethernet_open = FALSE;
    cb->time_domain_reflectometer.ethernet_short = FALSE;
    cb->time_domain_reflectometer.time = 0;
    cb->time_domain_reflectometer.clear = 0;
    if (ec596cbstart(ec, cb)) {
	if (cb->time_domain_reflectometer.link_ok) {
	    ec596warn(ec, warn_link_ok, 0);
	} else if (cb->time_domain_reflectometer.ethernet_open) {
	    ec596warn(ec, warn_ethernet_open,
		      cb->time_domain_reflectometer.time);
	} else if (cb->time_domain_reflectometer.ethernet_short) {
	    ec596warn(ec, warn_ethernet_short,
		      cb->time_domain_reflectometer.time);
	} else if (cb->time_domain_reflectometer.transceiver_cable) {
	    ec596warn(ec, warn_transceiver_cable,
		      cb->time_domain_reflectometer.time);
	} /* if */
    } /* if */
} /* ec596tdr */


/*
 * Set-up Individual Address.
 */
static boolean_t
ec596setupaddr(ecsi_t *ec)
{
    unsigned char cb_area[sizeof(cb_t) + DCACHE_LINE_SIZE];
    cb_t *cb = (cb_t *)CACHE_ALIGNED(cb_area);

#ifdef DEBUG
    if (ecdebug & EC_DEBUG_PRINT) {
	printf("ec596setupaddr: [%s]\n", ether_sprintf(ec->ec_addr));
    } /* if */
#endif
    *cb = cb_template;
    cb->ia_setup.command.cmd.cuc = cb_cuc_ia_setup;
    bcopy(ec->ec_addr, cb->ia_setup.address, sizeof(ec->ec_addr));
    return ec596cbstart(ec, cb);
} /* ec596setupaddr */


/*
 * Load multi-cast list.
 */
static boolean_t
ec596loadmulticast(ecsi_t *ec)
{
    unsigned char cb_area[sizeof(cb_t) + DCACHE_LINE_SIZE];
    cb_t *cb = (cb_t *)CACHE_ALIGNED(cb_area);

#ifdef DEBUG
    if (ecdebug & EC_DEBUG_PRINT) {
	printf("ecloadmultcast: %d entries\n", ec->nmc);
    } /* if */
#endif
    *cb = cb_template;
    cb->multicast_setup.command.cmd.cuc = cb_cuc_multicast_setup;
    cb->multicast_setup.cnt = ec->nmc * sizeof(*ec->mc);
    bcopy(ec->mc, cb->multicast_setup.mc, cb->multicast_setup.cnt);
    return ec596cbstart(ec, cb);
} /* ec596loadmulticast */


/*
 * Configure chip.
 */
static boolean_t
ec596cfg(ecsi_t *ec)
{
    iscp_t iscp;
    ipc_kmsg_t kmsg;
    rfd_t *last_rfd;
    tcb_t *last_tcb;
    char scp_area[sizeof(scp_t) + DCACHE_LINE_SIZE];
    unsigned char cb_area[sizeof(cb_t) + DCACHE_LINE_SIZE];
    cb_t *cb = (cb_t *)CACHE_ALIGNED(cb_area);
    scp_t *scp = (scp_t *)CACHE_ALIGNED(scp_area);
    scb_t *scb = &ec->ecsc->scb;
    rbd_t *rbd;
    rfd_t *rfd;
    tcb_t *tcb;
    tbd_t *tbd;
    int i;
    int unit = ec - ecsi;

    bzero(scp, sizeof(*scp));
    scp->sysbus.csw = 1;
    scp->sysbus.int_polarity = ACTIVE_LOW;
    scp->sysbus.trg = 1;
    scp->sysbus.lock = 1;	/* don't request locked cycles */
    scp->sysbus.mode = mode_linear;
    POINT(&scp->iscp_ptr, &iscp);
    bzero(&iscp, sizeof(iscp));
    POINT(&iscp.scb_ptr, &scb->scb.status);	/* skip padding */
    iscp.busy = 1;
    bzero(scb, sizeof(*scb));
    FLUSH(scp, sizeof(*scp));
    FLUSH(&iscp, sizeof(iscp));
    FLUSH(scb, sizeof(*scb));
    /*
     * Hand the alternate SCP address to chip and assert channel
     * attention to start initialization.
     */
    ec596portaccess(unit, pmap_extract(kernel_pmap, scp), port_scp);
    CHANNEL_ATTN(unit);

    /* wait <= 2 seconds for "busy" flag to lower, ignoring interrupt: */
    i = 2000;
    do {
        delay(1000);		/* wait 1 ms */
        PURGE(&iscp, sizeof(iscp.busy));
    } while (iscp.busy && --i > 0);
    if (i <= 0) {
        printf("%s%d: i82596 failed to respond\n", ecname, unit);
        return FALSE;
    } /* if */
#ifdef DEBUG
    if (ecdebug & EC_DEBUG_TX) {
	printf("ec596cfg: TCBs @ %.8x[+%.8x], TBDs @ %.8x[+%.8x]\n",
	       ec->ecsc->tcb_list, sizeof(*tcb),
	       &ec->tbd_list[0][0], sizeof(*tbd));
    } /* if */
#endif
    ec->tcb_head = ec->tcb_tail = tcb = ec->ecsc->tcb_list;
    last_tcb = &tcb[NTCBBUF - 1];
    tbd = &ec->tbd_list[0][0];

    /* initialize "Transmit" Command Block lists: */
    do {
	tbd_t *last_tbd = &tbd[NTBDBUF - 1];

	*tcb = cb_template.transmit;
	tcb->tcb.next = tcb + 1;
	tcb->tcb.prev = tcb - 1;
	POINT(&tcb->tcb.command.cmd.link,
	      &tcb[1].tcb.command.cacheline.cb_stat);
	/*
	 * Save the link address so we can dynamically check for
	 * trashing (see the BIG FAT NOTICEs above).
	 */
	tcb->tcb.command.save.link = tcb->tcb.command.cmd.link;

	/* initialized link field of first NTBDBUF-1 tbd's: */
	POINT(&tcb->tcb.tbd_ptr, tbd);
	do {
	    POINT(&tbd->pa, tbd);
#ifdef DEBUG
	    tbd->next = &tbd[1];
#endif
	    POINT(&tbd->link, &tbd[1]);
	} while (++tbd < last_tbd);
	POINT(&tbd->pa, tbd);
#ifdef DEBUG
	tbd->next = EC596_NIL_PTR;
#endif
	SET_EC596_NIL(&tbd->link);
    } while (++tbd, ++tcb <= last_tcb);
    /* make tcb linked list circular: */
    ec->tcb_head->tcb.prev = last_tcb;
    last_tcb->tcb.next = ec->tcb_head;
    POINT(&last_tcb->tcb.command.cmd.link,
	  &ec->tcb_head->tcb.command.cacheline.cb_stat);
    last_tcb->tcb.command.save.link = last_tcb->tcb.command.cmd.link;
    FLUSH(ec->ecsc->tcb_list, sizeof(ec->ecsc->tcb_list));
    FLUSH(ec->tbd_list, sizeof(ec->tbd_list));
    /*
     * Initialize the RFD and RBD lists synchronously, attaching two
     * adjacent RBDs to each RFD for an Ethernet packet's header and
     * body.
     */
#ifdef DEBUG
    if (ecdebug & EC_DEBUG_RX) {
	printf("ec596cfg: RFDs @ %.8x[+%.8x], RBDs @ %.8x[+%.8x]\n",
	       ec->ecsc->rfd_list,
	       sizeof(*rfd), ec->ecsc->rbd_list, sizeof(*rbd));
    } /* if */
#endif
    ec->rfd_head = ec->rfd_tail = rfd = ec->ecsc->rfd_list;
    last_rfd = &rfd[NRFDBUF - 1];
    ec->rbd_head = ec->rbd_tail = rbd = ec->ecsc->rbd_list;
    bzero(rfd, sizeof(ec->ecsc->rfd_list));

    /* initialize RBD list in pairs, along with RFD list */
    do {
        POINT(&rfd->rfd.rbd_ptr, &rbd->cmd.stat);
        if (rbd->rbd.kmsg) {
	    net_kmsg_put(rbd->rbd.kmsg);
	} /* if */
        bzero(rbd, 2*sizeof(*rbd));
        rbd->rbd.size = sizeof(struct ether_header);
        rbd->rbd.next = &rbd[1];
        POINT(&rbd->rbd.link, &rbd[1].cmd.stat);
        rbd->save.link = rbd->rbd.link;
        SET_EC596_NIL(&rbd->rbd.buffer);
#ifdef DEBUG
        POINT(&rbd->rbd.pa, &rbd->cmd.stat);
#endif
        (++rbd)->rbd.size = NET_RCV_MAX - sizeof(struct packet_header);
        rbd->rbd.next = &rbd[1];
        POINT(&rbd->rbd.link, &rbd[1].cmd.stat);
        rbd->save.link = rbd->rbd.link;
        SET_EC596_NIL(&rbd->rbd.buffer);
#ifdef DEBUG
        POINT(&rbd->rbd.pa, &rbd->cmd.stat);
#endif
        rfd->rfd.command.cmd.sf = TRUE;		/* use flexible mode */
        rfd->rfd.next = &rfd[1];
        POINT(&rfd->rfd.command.cmd.link,
	      &rfd[1].rfd.command.cacheline.cb_stat);

	/*
	 * Save the link address so we can dynamically check for
	 * trashing (see the BIG FAT NOTICEs above).
	 */
        rfd->rfd.command.save.link = rfd->rfd.command.cmd.link;
    } while (++rbd, ++rfd <= last_rfd);
    /* make rfd linked list circular: */
    last_rfd->rfd.next = ec->rfd_head;
    POINT(&last_rfd->rfd.command.cmd.link,
	  &ec->rfd_head->rfd.command.cacheline.cb_stat);
    last_rfd->rfd.command.save.link = last_rfd->rfd.command.cmd.link;
    /* make rbd linked list circular: */
    (--rbd)->rbd.next = ec->rbd_head;
    POINT(&rbd->rbd.link, &ec->rbd_head->cmd.stat);
    rbd->save.link = rbd->rbd.link;
    /*
     * The following is nasty: In order to initialize "ec596rxsetup()"
     * properly, we must create 1-element lists, consisting of an RFD
     * and an RBD-pair with a receive buffer attached to it, all set
     * up and ready for use by the chip.  Above, we just finished
     * making the "head" and "tail" pointers to these lists refer to
     * the same list element, so it only remains to allocate and
     * attach a receive buffer, then flush everything out to main
     * memory.  After this, "ec596rxsetup()" will perform future
     * buffer management without help.
     */
    rbd = ec->rbd_head;
    rfd = ec->rfd_head;
    net_kmsg_more();
    kmsg = net_kmsg_get();
    if (!kmsg) {
	panic("ec596cfg: no buffer");
    } /* if */
    PURGE(kmsg, sizeof(struct net_rcv_msg));
    rbd->rbd.kmsg = kmsg;
    POINT(&rbd->rbd.buffer, net_kmsg(kmsg)->header);
    (++rbd)->rbd.el = TRUE;
    POINT(&rbd->rbd.buffer,
	  net_kmsg(kmsg)->packet + sizeof(struct packet_header));

    /* initialize the corresponding RFD: */
    rfd->rfd.command.cmd.el = TRUE;

    /* now flush everything to main memory: */
    FLUSH(ec->ecsc->rfd_list, sizeof(ec->ecsc->rfd_list));
    FLUSH(ec->ecsc->rbd_list, sizeof(ec->ecsc->rbd_list));
#ifdef DEBUG
    if (ecdebug & EC_DEBUG_RX) {
	printf("ec596cfg: rfd=%.8x, rbd=(%.8x,%.8x), kmsg=%.8x\n",
	       rfd, rbd - 1, rbd, (rbd-1)->rbd.kmsg);
    } /* if */
#endif
    ec->ec_if.if_flags |= IFF_RUNNING;		/* enable interrupt handling */
    *cb = cb_template;	/* try a simple NOP to see if the unit interrupts */
    cb->nop.nop.command.cmd.cuc = cb_cuc_nop;

    /* initialize our special NOP CB (see second BIG FAT NOTICE above): */

    bcopy(&cb->nop.nop.command.cacheline.cb_stat, &ec->ecsc->nop,
	  sizeof(ec->ecsc->nop));
    FLUSH(&ec->ecsc->nop, sizeof(ec->ecsc->nop));
    if (!ec596cbstart(ec, cb)) {
        printf("%s%d: 82596 failed to interrupt\n", ecname, unit);
        return FALSE;
    } /* if */

    /*
     * Configure the chip to our needs, beginning with the default
     * settings from the INTEL manual.
     */
    *cb = cb_template;
    cb->nop.nop.command.cmd.cuc = cb_cuc_configure;
    cb->configure.cfg.bytes[ 0] = 0x0e;
    cb->configure.cfg.bytes[ 1] = 0xc8;
    cb->configure.cfg.bytes[ 2] = 0x40;
    cb->configure.cfg.bytes[ 3] = 0x26;
    cb->configure.cfg.bytes[ 4] = 0x00;
    cb->configure.cfg.bytes[ 5] = 0x60;
    cb->configure.cfg.bytes[ 6] = 0x00;
    cb->configure.cfg.bytes[ 7] = 0xf2;
    cb->configure.cfg.bytes[ 8] = 0x00;
    cb->configure.cfg.bytes[ 9] = 0x00;
    cb->configure.cfg.bytes[10] = 0x40;
    cb->configure.cfg.bytes[11] = 0xff;
    cb->configure.cfg.bytes[12] = 0x00;
    cb->configure.cfg.bytes[13] = 0x3f;
    cb->configure.cfg.bytes[14] = 0x00;
    cb->configure.cfg.bytes[15] = 0x00;

    /* add our own customizations: */

    /* no frame CRC in memory: */
    cb->configure.cfg.field.disable_crc_in_memory = TRUE;

    /* src add already in buf: */
    cb->configure.cfg.field.no_src_addr_ins = TRUE;

    /* req. bus @ 1/4 space available: */
    cb->configure.cfg.field.fifo_limit = 12;

    cb->configure.cfg.field.loopback_mode =
      (ec->ec_if.if_flags & IFF_LOOPBACK) ? 3 : 0;
    cb->configure.cfg.field.promiscious_mode =
      (ec->ec_if.if_flags & IFF_PROMISC)/IFF_PROMISC;

    /* don't accept all multicast frames: */
    cb->configure.cfg.field.disable_mc_all =
      1 - (ec->ec_if.if_flags & IFF_ALLMULTI)/IFF_ALLMULTI;
#ifdef DEBUG
    if (ecdebug & EC_DEBUG_PRINT) {
	printf("ec596cfg: loopback/promisc/disable_mc_all is %d/%d/%d\n",
	       cb->configure.cfg.field.loopback_mode,
	       cb->configure.cfg.field.promiscious_mode,
	       cb->configure.cfg.field.disable_mc_all);
    } /* if */
#endif
    ec->if_flags = ec->ec_if.if_flags & IFF_CHIPCONFIG;	/* save chip config */
    return ec596cbstart(ec, cb);
} /* ec596cfg */


/*
 * Reset chip.
 */
void
ec596reset(int unit)
{
    rbd_t *last_rbd;
    tcb_t *last_tcb;
    int s;
    ecsi_t *ec = &ecsi[unit];
    rbd_t *rbd;
    tcb_t *tcb;
    io_req_t ior;
    int i;

#ifdef DEBUG
    if (ecdebug & EC_DEBUG_PRINT) {
	printf("ec596reset: ...\n");
    } /* if */
    ecto = 0;
#endif
    ec->initialized = FALSE;
    ec->scberrs = FALSE;
    ec->ec_if.if_flags &= ~IFF_RUNNING;
    ec->ec_if.if_timer = 0;

    /* reset chip: */
    ec596portaccess(unit, 0, port_reset);

    /* get ROM signature, self-test: */
    if (ec->signature == 0 || ec->force_selftest) {
	selftest_t selftest_area[sizeof(selftest_t) + DCACHE_LINE_SIZE];
        selftest_t *selftest = (selftest_t*)CACHE_ALIGNED(selftest_area);

        selftest->signature = 0;
        selftest->result.all = -1;
        FLUSH(selftest, sizeof(*selftest));
        ec596portaccess(unit, pmap_extract(kernel_pmap, selftest),
			port_selftest);

	/* wait 2 seconds for self-test to finish: */
        i = 2000;
        do {
	    /* self-test didn't finish */
            if (--i < 0) {
		/*
		 * We probably got here because the cable isn't
		 * connected.  Do not print a message because, if the
		 * user has no network, this is acceptable and we
		 * don't want to alarm the user.  Note that the TDR
		 * check in "ec596init()" will print out a cable
		 * diagnostic later anyway.  Here, we only complain if
		 * the watchdog could not re-initialize the chip or
		 * we know that we are not a standalone machine.
		 */
		break;
	    } /* if */
            delay(1000);		/* wait 1 ms */
            PURGE(selftest, sizeof(*selftest));
	} while (selftest->result.of.zeros);
        if (ec->force_selftest) {
	    if (i < 0) {
		printf("%s%d: 82596 time-out during self-test\n", ecname,
		       unit);
		KDB_BREAK("ec596reset: self-test time-out");
	    } else if (selftest->result.all) {
		printf("%s%d: 82596 failed self-test (result = %.8x)\n",
		       ecname, unit, selftest->result.all);
		KDB_BREAK("ec596reset: self-test failed");
            } else {
		/* don't repeat self-test */
		ec->force_selftest = FALSE;
	    } /* if */
	} /* if */
	/*
	 * Get a non-zero check sum so that we don't repeat the
	 * self-test solely to get the signature.
	 */
	ec->signature = selftest->signature;
	if (!ec->signature) {
	    ec->signature = 0xbad0bad;
	} /* if */
#ifdef DEBUG
        if (ecdebug & EC_DEBUG_PRINT) {
	    printf("82596 signature is %.8x\n", ec->signature);
	} /* if */
#endif
	/* reset chip */
        ec596portaccess(unit, 0, port_reset);
    } /* if */
    ec->retransmit = 0;
    ec->q_op = FALSE;
    ec->recover = FALSE;
    ec->txactive = FALSE;	/* no transmit is active */

    /* wait 0.9s before a "Transmit" times out: */
    ec->txtimelap = 90 * PAGE0->mem_10msec;

    tcb = ec->ecsc->tcb_list;
    last_tcb = &tcb[NTCBBUF - 1];
    for (; tcb <= last_tcb; ++tcb) {
	ior = tcb->tcb.ior;
	if (ior) {
	    /* deallocate "Transmit" buffer */
	    s = splimp();
	    ior->io_error = D_DEVICE_DOWN;
	    ior->io_op |= IO_ERROR;
	    ior->io_residual = ior->io_count;
	    iodone(ior);
	    IF_DROP(&ec->ec_if.if_snd);
	    splx(s);
	    tcb->tcb.ior = (io_req_t)0;
	} /* if */
    } /* for */
    FLUSH(ec->ecsc->tcb_list, sizeof(ec->ecsc->tcb_list));

    rbd = ec->ecsc->rbd_list;
    last_rbd = &rbd[NRBDBUF - 1];
    for (; rbd <= last_rbd; ++rbd) {
        if (rbd->rbd.kmsg) {
	    /* deallocate receive buffer */
            net_kmsg_put(rbd->rbd.kmsg);
            rbd->rbd.kmsg = (ipc_kmsg_t)0;
	} /* if */
        SET_EC596_NIL(&rbd->rbd.buffer);
        ++rbd;
        SET_EC596_NIL(&rbd->rbd.buffer);
    } /* for */
    FLUSH(ec->ecsc->rbd_list, sizeof(ec->ecsc->rbd_list));

    /* drop the output queue: */
    s = splimp();
    do {
	IF_DEQUEUE(&ec->ec_if.if_snd, ior);
	if (ior) {
#ifdef DEBUG
	    if (ecdebug & EC_DEBUG_PRINTALL) {
		printf("ec596reset: dropping packet\n");
	    } /* if */
#endif
	    ior->io_error = D_DEVICE_DOWN;
	    ior->io_op |= IO_ERROR;
	    ior->io_residual = ior->io_count;
	    iodone(ior);
	    IF_DROP(&ec->ec_if.if_snd);
	} /* if */
    } while (ior);
    splx(s);

    /* release any thread in "ec596cbstart()" */
    thread_wakeup((void *)ec);
} /* ec596reset */


/*
 * Initialize interface.
 */
static int
ec596init(int unit)
{
    ecsi_t *ec = &ecsi[unit];
    int status = D_IO_ERROR;

#ifdef DEBUG
    if (ecdebug & EC_DEBUG_PRINT) {
	printf("ec596init: unit %d\n", unit);
    } /* if */
#endif
    lock_write(&ec->init_lock);
    aspitab(INT_LAN, SPLLAN, ec596intr, unit, 0);
    ec596reset(unit);		/* reset the controller */
    /*
     * NOTE: For HP's PA-RISC architecture, we can't simply call
     * "lock_init()" in the following line, since this
     * implementation's "lock_data_t" variables refer to
     * dynamically-allocated "simple_lock_t" variables and
     * "lock_init()" would (A) reallocate new simple locks and (B)
     * destroy references to the old ones, thereby creating a memory
     * "leak".  We therefore let the "q_lock_init" field hold the
     * unique, initial value of its corres- ponding "q_lock" field,
     * and we use "struct"-assignment to overwrite "q_lock" when
     * necessary.
     */
    ec->q_lock = ec->q_lock_init;	/* (re)initialize the queue lock */
    simple_unlock(&ec->q_lock.interlock);

    /* configure chip: */
    if (ec596cfg(ec)) {
	if (ec596setupaddr(ec) && ec596loadmulticast(ec)) {
	    /* wait until we can allocate all RBDs: */
	    while (!ec596rxstart(ec)) {
		assert_wait((int)ec+ 1, FALSE);
		thread_block((void (*)())0);
            } /* while */
	    ec596txsetup(unit);		/* start any output */
	    ec->ec_if.if_timer = 1;	/* allow watchdog to run */
	    ec596tdr(ec);		/* perform cable diagnosis */
	    ec->initialized = TRUE;
	    status = D_SUCCESS;
	} else {
	    ec->ec_if.if_flags &= ~IFF_RUNNING;
        } /* if */
    } /* if */
    lock_done(&ec->init_lock);
#ifdef DEBUG
    if (ecdebug & EC_DEBUG_PRINT) {
	printf("ec596init: returns %d\n", status);
    } /* if */
#endif
    return status;
} /* ec596init */


int
ec596open(int dev, dev_mode_t mode, io_req_t ior)
{
    int status, unit;

    unit = minor(dev);
    if ((unsigned) unit >= NEC || !ecalive[unit]) {
	return D_NO_SUCH_DEVICE;
    } /* if */
    status = ec596init(unit);
    if (!status) {
	ecsi[unit].ec_if.if_flags |= IFF_UP;	/* interface is available */
    } /* if */
    return status;
} /* ec596open */


int
ec596close(int dev, dev_mode_t mode)
{
    int unit;
    ecsi_t *ec;

    unit = minor(dev);
    if ((unsigned) unit >= NEC || !ecalive[unit]) {
	return D_NO_SUCH_DEVICE;
    } /* if */
    ec = &ecsi[unit];
    if (!(ec->ec_if.if_flags & IFF_UP)) {
	return D_DEVICE_DOWN;
    } /* if */
    ec596reset(unit);
    ec->force_selftest = TRUE;	/* force self-test upon subsequent "open" */
    ec->ec_if.if_flags &= ~IFF_UP;
    return D_SUCCESS;
} /* ec596close */


int
ec596output(int dev, io_req_t ior)
{
    int unit;

    unit = minor(dev);
    if ((unsigned) unit >= NEC || !ecalive[unit]) {
	return D_NO_SUCH_DEVICE;
    } /* if */
    return net_write(&ecsi[unit].ec_if, ec596txsetup, ior);
} /* ec596_output */


int
ec596getstat(int dev, int flavor, dev_status_t status, unsigned int *count)
{
    int unit;

    unit = minor(dev);
    if ((unsigned) unit >= NEC || !ecalive[unit]) {
	return D_NO_SUCH_DEVICE;
    } /* if */
    return net_getstat(&ecsi[unit].ec_if, flavor, status, count);
} /* ec596getstat */


int
ec596setstat(int dev, int flavor, dev_status_t status, unsigned int count)
{
    ecsi_t *ec;
    int unit;

    unit = minor(dev);
    if ((unsigned) unit >= NEC || !ecalive[unit]) {
	return D_NO_SUCH_DEVICE;
    } /* if */
    ec = &ecsi[unit];
    switch (flavor) {
	union {
	    eaddr_t	addr;
	    int		word[2];
} convert;
        int old_flags;

#define NS ((struct net_status *)status)
      case NET_STATUS:
	if (count < NET_STATUS_COUNT) {
	    return D_INVALID_SIZE;
	} /* if */
	old_flags = ec->ec_if.if_flags;
	ec->ec_if.if_flags = old_flags & IFF_CANTCHANGE
	  | NS->flags & ~IFF_CANTCHANGE;
	if (ec->ec_if.if_flags & IFF_DEBUG) {
	    printf("ecset_status: NET_STATUS to %.4x\n", NS->flags);
	} /* if */
	if (NS->flags & IFF_UP) {
	    /* bring interface up */
	    if (!(old_flags & IFF_RUNNING) || (NS->flags & IFF_CHIPCONFIG)
		!= ec->if_flags)
	    {
		/* change interface configuration */
		return ec596init(ec->ec_if.if_unit);
	    } /* if */
	} else {
	    /* take down interface */
	    ec596reset(ec->ec_if.if_unit);
	    ec->force_selftest = FALSE;		/* no selftest msg */
	} /* if */
	return D_SUCCESS;
#undef NS

      case NET_ADDRESS:
	if (count < sizeof(convert)/sizeof(int)) {
	    return D_INVALID_SIZE;
	} /* if */
	convert.word[0] = ntohl(status[0]);
	convert.word[1] = ntohl(status[1]);
	if (ec->ec_if.if_flags & IFF_DEBUG) {
	    printf("ecset_status: NET_ADDRESS to [%s]\n",
		   ether_sprintf(convert.addr));
	} /* if */
	if (INVALIDADDR(convert.addr)) {
	    return D_INVALID_OPERATION;
	} /* if */
	bcopy(convert.addr, ec->ec_addr, sizeof(ec->ec_addr));
	return (ec->ec_if.if_flags & IFF_RUNNING)
	  ? (ec596setupaddr(ec) ? D_SUCCESS : D_IO_ERROR)
	    : ec596init(ec->ec_if.if_unit);

      case NET_ADD_MULTICAST_ADDRESS:
	/*
	 * NOTE: There is currently no machine-independent interface
	 * to the following functions in the OSF/1-compatible server,
	 * so this is mostly "just for the record" -- a reminder to
	 * add this support someday.  We assume that Multi-cast
	 * addresses will eventually be passed in the same form as the
	 * interface address above.  --Scott Marovich, HP
	 * Laboratories, April 1992.
	 */
	if (count < sizeof(convert)/sizeof(int)) {
	    return D_INVALID_SIZE;
	} /* if */
	convert.word[0] = ntohl(status[0]);
	convert.word[1] = ntohl(status[1]);
	if (ec->ec_if.if_flags & IFF_DEBUG) {
	    printf("ecset_status: NET_ADD_MULTICAST_ADDRESS [%s]\n",
		   ether_sprintf(convert.addr));
	} /* if */
	if (INVALIDADDR(convert.addr)) {
	    return D_INVALID_OPERATION;
	} /* if */
	unit = 0;
	while (!ECADDRMATCH(ec->mc[unit], convert.addr) && ++unit < ec->nmc);
	if (unit < ec->nmc) {
	    return D_SUCCESS;
	} /* if */
	if (ec->nmc >= NMULTICAST) {
	    return D_INVALID_OPERATION;
	} /* if */
	bcopy(convert.addr, ec->mc[ec->nmc++], sizeof(convert.addr));
	return (ec->ec_if.if_flags & IFF_RUNNING)
	  ? (ec596loadmulticast(ec) ? D_SUCCESS : D_IO_ERROR)
	    : ec596init(ec->ec_if.if_unit);

      case NET_DELETE_MULTICAST_ADDRESS:
	if (count < sizeof(convert)/sizeof(int)) {
	    return D_INVALID_SIZE;
	} /* if */
	convert.word[0] = ntohl(status[0]);
	convert.word[1] = ntohl(status[1]);
	if (ec->ec_if.if_flags & IFF_DEBUG) {
	    printf("ecset_status: NET_DELETE_MULTICAST_ADDRESS [%s]\n",
		   ether_sprintf(convert.addr));
	} /* if */
	unit = 0;
	while (!ECADDRMATCH(ec->mc[unit], convert.addr) && ++unit < ec->nmc);
	if (unit >= ec->nmc) {
	    /* address not found */
	    return D_INVALID_OPERATION;
	} /* if */
	if (unit < --ec->nmc) {
	    /* not last list entry */
	    bcopy(ec->mc[unit+1], ec->mc[unit],
		  ec->nmc-unit * sizeof(eaddr_t));
	} /* if */
	return (ec->ec_if.if_flags & IFF_RUNNING)
	  ? (ec596loadmulticast(ec) ? D_SUCCESS : D_IO_ERROR)
	    : ec596init(ec->ec_if.if_unit);
    } /* switch */
    return D_INVALID_OPERATION;
} /* ec596setstat */


int
ec596setinput(int dev, mach_port_t port, int priority,
	      filter_t *filter, unsigned int count)
{
    int unit;

    unit = minor(dev);
    if ((unsigned) unit >= NEC || !ecalive[unit]) {
	return D_NO_SUCH_DEVICE;
    } /* if */
    return net_set_filter(&ecsi[unit].ec_if, (ipc_port_t)port,
			  priority, filter, count);
} /* ec596setinput */


/*
 * Process packet trailer, which is completely insane for the MACH
 * "micro-kernel", but we do it anyway.  The basic idea in Berkeley's
 * original networking code was to use virtual-memory page-mapping
 * instead of memory-to-memory copying, by letting packet data be
 * page-aligned.  This was done by transmitting protocol headers
 * *after* the data, rather than before, so that device drivers could
 * DMA a packet onto a page boundary, then chop off the trailer into a
 * separate MBUF and re-link it to the head of the packet's MBUF list.
 * But in our case, the entire packet is placed in a MACH IPC message
 * that the "microkernel" exports to a protocol handler in some user
 * task, and the packet data is not page-aligned here because it makes
 * no sense until protocol processing has determined that it is
 * necessary.  Thus, the only thing we can do with a trailer is to
 * reformat the packet by swapping the protocol headers and packet
 * data (squashing out the trailer-descriptor in the process).  But
 * this implies memory-to-memory copying, which is the very thing that
 * trailer-packets were intended to avoid!
 *
 * A better solution to this entire mess is to configure the interface
 * such that no trailer packets are accepted from any remote host.
 * But in case we really must handle them, the following code should
 * keep this driver from falling flat on its fanny.
 *
 * Having said all of this, it should be clear, that we don't care
 * much about the performance of this function.  Relative to the
 * cost of the extra copy, it just doesn't matter much.  Plus,
 * who uses trailers anyway?
 */
static ipc_kmsg_t
ec596get_trailer(ecsi_t *ec, ipc_kmsg_t kmsg)
{
    int i;
    unsigned int datlen, traildatlen;
    struct ether_header *eh;
    struct packet_header *pkthdr;
    struct trailhdr_t {
	unsigned short ether_type;
	unsigned short len;
    } *trailhdr;

    eh = (struct ether_header*) net_kmsg(kmsg)->header;
    pkthdr = (struct packet_header *)net_kmsg(kmsg)->packet;

    datlen = pkthdr->type - ETHERTYPE_TRAIL;
    traildatlen = pkthdr->length - (datlen << 9);
    if (traildatlen >= sizeof(struct trailhdr_t)) {
	/* find start of trailer */
	char trailer[100]; /* MAGIC SIZE!!!  EXPLANATION FOLLOWS: */
	/*
	 * NOTE: In a world without MBUFs, the maximum size of a
	 * trailer is defined as a "magic number", calculated as
	 * follows:
	 *
	 * Max. length of 1 MBUF				: 128 bytes
	 * - sizeof(struct pkthdr) [1 ptr, 1 "long"]		: - 8 bytes
	 * - sizeof(struct m_hdr) [3 ptrs, 2 "short"s, 1 "long"]: -20 bytes
	 * 							  ---------
	 *							  100 bytes
	 */
	trailhdr = (struct trailhdr_t *)((char*)(pkthdr + 1) + datlen);
	if (trailhdr->len <= traildatlen) {
	    i = trailhdr->len - sizeof(struct trailhdr_t);
	    if (i > sizeof(trailer)) {
		panic("ec596get: insufficient space for trailer");
	    } /* if */
	    pkthdr->length -= sizeof(struct trailhdr_t);
	    pkthdr->type = eh->ether_type = trailhdr->ether_type;
	    bcopy(trailhdr + 1, trailer, i);
	    ovbcopy(pkthdr + 1, (char*)(pkthdr + 1) + i, datlen);
	    bcopy(trailer, pkthdr + 1, i);
	    return kmsg; /* -> packet message */
	} /* if */
    } /* if */
#ifdef DEBUG
    if (ecdebug & EC_DEBUG_RX) {
	printf("ec596get_trailer: packet discarded\n");
    } /* if */
#endif
    ++ec->ec_if.if_ierrors; /* error in trailer packet if we get here */
    if (ec->ec_if.if_flags & IFF_DEBUG) {
	printf("ec596get: if_ierrors++\n");
    } /* if */
    net_kmsg_put(kmsg);
    return (ipc_kmsg_t)0;
} /* ec596get_trailer */


static __inline ipc_kmsg_t
ec596get(ecsi_t *ec)
{
    rfd_t *rfd = ec->rfd_head;
    rbd_t *rbd = ec->rbd_head, *next;
    struct ether_header *eh;
    struct packet_header *pkthdr;
    ipc_kmsg_t kmsg;
    int i;

    /* check incoming packets for errors, and handle trailer packets. */

    PURGE(&rfd->rfd.command.cacheline.cb_cmd, DCACHE_LINE_SIZE);
    PURGE(rbd, sizeof(*rbd));
    next = rbd->rbd.next;
    PURGE(next, sizeof(*next));
#ifdef DEBUG
    if (ecdebug & EC_DEBUG_CHECK) {
        if (!rfd->rfd.command.cmd.ok) {
	    panic("ec596get: !OK in RFD");
	} /* if */

	/*
	 * These fields are never activated because we don't put
	 * received data in RFD.  The "Filled" bit does get set by the
	 * chip.
	 */
        if (rfd->rfd.eof || rfd->rfd.count) {
	    panic("ec596get: EOF or FILLED or COUNT != 0 in RFD");
	} /* if */
        if (!rbd->rbd.filled || rbd->rbd.count != rbd->rbd.size) {
            printf("ec596get: bad rbd count is %d\n", rbd->rbd.count);
            panic("ec596get: !FILLED or COUNT in first RBD");
	} /* if */
        if (rbd->rbd.eof) {
	    panic("ec596get: unexpected EOF");
	} /* if */
	/*
	 * Note: the RFD should point to the RBD, but we cannot
	 * guarantee that the the RBD pointer field did not get
	 * overwritten by us, which could have happened: If the chip
	 * was writing the RBD pointer field at the same same time
	 * that we were resetting the EL bit in that RFD while linking
	 * in a new RFD, we may have trashed it.  INTEL assures us
	 * that RBDs get taken from the RBD list in strict order, so
	 * we can do without that pointer.  As a sanity check, see if
	 * the RBD pointer field is odd, which means that it has not
	 * been set, otherwise compare the pointer to what we expect.
	 * This is ugly, but it is only for debugging . . ..
	 */
        if (   !(rfd->rfd.rbd_ptr.low & 1)
            && (   rbd->rbd.pa.low  != rfd->rfd.rbd_ptr.low
                || rbd->rbd.pa.high != rfd->rfd.rbd_ptr.high
		)
	    )
        {
            printf(
	     "ec596get: mismatch rbd->rbd.pa=%.4x%.4x rfd->rbd_ptr=%.4x%.4x\n",
		   rbd->rbd.pa.high, rbd->rbd.pa.low,
		   rfd->rfd.rbd_ptr.high, rfd->rfd.rbd_ptr.low);
            panic("ec596get: 1st RBD out of sync");
	} /* if */
        if (!rbd->rbd.kmsg) {
	    panic("ec596get: kmsg is nil");
	} /* if */
    } /* if */
#endif /* DEBUG */
    kmsg = rbd->rbd.kmsg;
    rbd->rbd.kmsg = (ipc_kmsg_t)0;
    SET_EC596_NIL(&rbd->rbd.buffer);
    FLUSH(&rbd->cacheline.cmd, sizeof(rbd->cacheline.cmd));
    SET_EC596_NIL(&next->rbd.buffer);
    FLUSH(&next->cacheline.cmd, sizeof(next->cacheline.cmd));
    /*
     * When multi-casting, verify the destination address (filter is
     * not exact).
     */
    eh = (struct ether_header*)net_kmsg(kmsg)->header;
    if (ec->nmc && (eh->ether_dhost[0] & 1)
        && !(ec->if_flags & (IFF_ALLMULTI|IFF_PROMISC)))
    {
        if (ECADDRMATCH(eh->ether_dhost, etherbroadcastaddr)) {
	    /* our address */
#ifdef DEBUG
            ++ec4all;
#endif
	} else {
	    i = 0;
	    do {
		if (ECADDRMATCH(ec->mc[i], eh->ether_dhost)) {
		    /* acceptable address */
#ifdef DEBUG
		    ++ec4multi;
#endif
		} /* if */
	    } while (++i < ec->nmc);
	    if (i >= ec->nmc) {
#ifdef DEBUG
		++ec4other;
#endif
		net_kmsg_put(kmsg);
		return (ipc_kmsg_t) 0;
	    } /* if */
	} /* if */
    } /* if */
#ifdef DEBUG
    if (ECADDRMATCH(eh->ether_dhost, ec->ec_addr)) {
	++ec4me;
    } else {
	if (ECADDRMATCH(eh->ether_dhost, etherbroadcastaddr)) {
	    ++ec4all;
	} else {
	    if ((eh->ether_dhost[0] & 1) && (ec->if_flags & IFF_ALLMULTI)) {
		++ec4multi;
	    } else {
		if (ec->if_flags & IFF_PROMISC) {
		    ++ec4any;
		} else {
		    ++ec4who;
		} /* if */
	    } /* if */
	} /* if */
    } /* if */
    if (ecdebug & EC_DEBUG_PRINTALL) {
	printf("ec596get #1: %d bytes\n", rbd->rbd.count);
    } /* if */
    if (ecdebug & EC_DEBUG_CHECK) {
        if (!next->rbd.filled) {
	    panic("ec596get: !FILLED in second RBD");
	} /* if */
        if (  (unsigned) next->rbd.count - ETHERMIN
            > (ETHERMTU-ETHERMIN+sizeof(struct ether_header)))
        {
            printf("ec596get: bad rbd count is %d\n", next->rbd.count);
            panic("ec596get: received packet size is out of range");
	} /* if */
        if (   (rbd->rbd.link.low & 1)
            || (rbd->rbd.link.low & 0xffffl | rbd->rbd.link.high<<16)
	    != pmap_extract(kernel_pmap, &next->cmd.stat)) 
	{
	    panic("ec596get: 2nd RBD out of sync");
	} /* if */
        if (!next->rbd.eof) {
	    panic("ec596get: missing EOF");
	} /* if */
        if (next->rbd.kmsg) {
	    panic("ec596get: kmsg is non-nil");
	} /* if */
    } /* if */
    if (ecdebug & EC_DEBUG_PRINTALL) {
	printf("ec596get #2: %d bytes\n", next->rbd.count);
    } /* if */
#endif /* DEBUG */
    rbd = next;
    pkthdr = (struct packet_header *)net_kmsg(kmsg)->packet;
    pkthdr->length = rbd->rbd.count + sizeof(*pkthdr);
    pkthdr->type = eh->ether_type;
    if ((unsigned)(pkthdr->type - ETHERTYPE_TRAIL) < ETHERTYPE_NTRAILER) {
	kmsg = ec596get_trailer(ec, kmsg);
    } /* if */
#ifdef DEBUG
    if (kmsg && (ecdebug & EC_DEBUG_PRINT)) {
	printf(" R%d", rbd->rbd.count);
    } /* if */
#endif
    return kmsg; /* -> packet message */
} /* ec596get */


/*
 * Set up Receive Frame Area.
 */
static boolean_t
ec596rxsetup(ecsi_t *ec)
{
    rfd_t *rfd;
    rbd_t *rbd;
    /*
     * Allocate receive buffers and recycle RBD pairs, along with
     * their corresponding RFDs.  Since new receive buffer memory may
     * not always be available at any given time, try requesting it
     * and processing an RBD pair BEFORE recycling the corresponding
     * RFD.
     */
    while ((rfd = ec->rfd_tail->rfd.next) != ec->rfd_head) {
	ipc_kmsg_t kmsg;
	rbd_t *prev_rbd;

	prev_rbd = ec->rbd_tail->rbd.next;
	rbd = prev_rbd->rbd.next;
        if (rbd == ec->rbd_head || rbd->rbd.next != rbd + 1
	    || ec->rbd_head == rbd+1)
	{
	    panic("ec596rxsetup: RFD<->RBD-pair links out of sync");
	} /* if */
        if (!rbd->rbd.kmsg && !(kmsg = net_kmsg_get())) {
	    /* no buffer just now */
#ifdef DEBUG
            if (ecdebug & EC_DEBUG_RX) {
		printf("ec596rxsetup: no IPC buffer\n");
	    } /* if */
#endif
            return FALSE;
	} /* if */
	/*
	 * Purge the receive buffer don't touch it again until the
	 * chip has put data in it.  Then initialize other data
	 * structures for the chip and flush them to main memory.
	 */
        PURGE(kmsg, sizeof(struct net_rcv_msg));
        rbd->rbd.kmsg = kmsg;		/* reinitialize RBD pair */
        rbd->rbd.eof = FALSE;
        rbd->rbd.filled = FALSE;
        rbd->rbd.count = 0;
        POINT(&rbd->rbd.buffer, net_kmsg(kmsg)->header);
        ++rbd;
        rbd->rbd.eof = FALSE;
        rbd->rbd.filled = FALSE;
        rbd->rbd.el = TRUE;
        rbd->rbd.count = 0;
        POINT(&rbd->rbd.buffer,
	      net_kmsg(kmsg)->packet + sizeof(struct packet_header));
        FLUSH(rbd-1, 2*sizeof(*rbd));	/* push RBDs to main memory */

	/* reinitialize the corresponding RFD. */

    	rfd->rfd.command.cacheline.cb_stat = 0; /* Reinitialize RFD */
    	rfd->rfd.command.cmd.el = TRUE;
#ifdef DEBUG
    	rfd->rfd.count = 0; /* these fields are never supposed to be set */
    	rfd->rfd.eof = FALSE;
#endif
    	rfd->rfd.filled = FALSE; /* this IS set by controller... */
    	FLUSH(rfd, sizeof(*rfd));

	/* remove previous RFD's end-of-list marker. */

    	PURGE(&ec->rfd_tail->rfd.command.cacheline.cb_cmd, DCACHE_LINE_SIZE);
    	ec->rfd_tail->rfd.command.cmd.el = FALSE;
    	FLUSH(&ec->rfd_tail->rfd.command.cacheline.cb_cmd, DCACHE_LINE_SIZE);

	/* remove previous RBD's end-of-list marker. */

        PURGE(&prev_rbd->cacheline.cmd, DCACHE_LINE_SIZE);
        prev_rbd->rbd.el = FALSE;
        FLUSH(&prev_rbd->cacheline.cmd, DCACHE_LINE_SIZE);
#ifdef DEBUG
        if (ecdebug & EC_DEBUG_RX) {
	    printf("ec596rxsetup: rfd=%.8x, rbd=(%.8x, %.8x), kmsg=%.8x\n",
		   rfd, rbd-1, rbd, kmsg);
	} /* if */
#endif
	/*
	 * Now advance the two parallel list pointers, thereby
	 * recycling the corresponding elements that we have just
	 * reinitialized and advancing to the next list elements.
	 */
        ec->rbd_tail = rbd - 1;
    	ec->rfd_tail = rfd;
    } /* while */
    return TRUE;
} /* ec596rxsetup */


/*
 * Handle "Frame Received" (FR) interrupt.
 */
static void
ec596frint(ecsi_t *ec)
{
    int n = 0;
    /*
     * Basically, this routine simply loops to receive and route
     * incoming packets.  In order to defend against a "hung" receiver
     * due to data corruption by the chip (cf. second BIG FAT NOTICE
     * above), we must insure that we don't get stuck with a missing
     * "Complete" bit in an RFD.  Since the huge receive buffers we
     * employ should guarantee that there is never more than 1 RFD per
     * packet, if the "Complete" bit isn't set, we assume that the
     * chip bug has occurred and reinitialize.
     */
    while (!ec->recover) {
	/* check if we received a packet */
	ipc_kmsg_t kmsg;
	rfd_t *rfd = ec->rfd_head;

        PURGE(&rfd->rfd.command.cacheline.cb_stat,
	      sizeof(rfd->rfd.command.cacheline.cb_stat));
        if (!rfd->rfd.command.cmd.completion || !I82596RFDVALID(rfd)) {
	    /* bug! */
            if (n == 0) {
		/* check if next RFD has completion bit set */
                PURGE(&rfd->rfd.command.cacheline.cb_cmd,
		      sizeof(rfd->rfd.command.cacheline.cb_cmd));
                if (!rfd->rfd.command.cmd.el) {
		    /* there is a successor RFD */
                    rfd = rfd->rfd.next;
                    PURGE(&rfd->rfd.command.cacheline.cb_stat,
			  sizeof(rfd->rfd.command.cacheline.cb_stat));
                    if (rfd->rfd.command.cmd.completion && I82596RFDVALID(rfd))
                    {
                        ++ec->reason.rfd_not_completed;
			KDB_BREAK("ec596frint: RFD completion bit not set");
                        ec->recover = TRUE;
		    } /* if */
		} /* if */
	    } /* if */
            return;
	} /* if */
        ++n;
	kmsg = ec596get(ec);
        if (kmsg) {
#ifdef DEBUG
            ecreceive[ecrindex] = kmsg;	/* save I/O request for debugging */
            if (++ecrindex >= ECNMBUF) {
		ecrindex = 0;
	    } /* if */
#endif
            net_packet(&ec->ec_if, kmsg,
		     ((struct packet_header *)net_kmsg(kmsg)->packet)->length,
		       ethernet_priority(kmsg));
            ++ec->ec_if.if_ipackets;
#ifdef USELEDS
	    if (inledcontrol == 0) {
		ledcontrol(0, 0, LED_LANRCV);
	    } /* if */
#endif
	} /* if */
        ec->rfd_head = rfd->rfd.next;
        ec->rbd_head = ec->rbd_head->rbd.next->rbd.next;
        ec596rxsetup(ec);		/* allocate receive buffers */
    } /* while */
} /* ec596frint */


/*
 * Handle "Transmit" Command interrupt.
 */
static __inline void
ec596txint(ecsi_t *ec)
{
    tcb_t *tcb;

#ifdef DEBUG
    if (ecto) {
	return;
    } /* if */
#endif
    while ((tcb = ec->tcb_head) != ec->tcb_tail) {
	boolean_t fatal;

    	PURGE(&tcb->tcb.command.cacheline.cb_stat,
	      sizeof(tcb->tcb.command.cacheline.cb_stat));
    	if (!tcb->tcb.command.cmd.completion) {
	    break;			/* packet xmit not done */
	} /* if */
        ec->txtimestamp = it_read();	/* keep watchdog at bay */
        PURGE(&tcb->tcb.command.cacheline.cb_cmd,
	      sizeof(tcb->tcb.command.cacheline.cb_cmd));
        if (LINKMISMATCH(tcb->tcb.command.save.link,
			 tcb->tcb.command.cmd.link))
        {
            ++ec->reason.tcb_corrupted;	/* chip-bug data corruption occurred */
	    KDB_BREAK("ec596txint: TCB forward pointer corrupted");
            ec->recover = TRUE;
            return;
	} /* if */
#ifdef DEBUG
	/* keep MBUF pointer for debugging */
        ecsend[ecsindex] =
	  ec->retransmit ? (io_req_t)0x22222222 : tcb->tcb.ior;
        if (++ecsindex >= ECNMBUF) {
	    ecsindex = 0;
	} /* if */
#endif
        if (tcb->tcb.command.cmd.ok) {
	    if (tcb->tcb.command.xmit.deferred) {
		++ec->xmit_status.deferred;
	    } /* if */
	    if (tcb->tcb.command.xmit.maxcoll) {
		ec->ec_if.if_collisions += tcb->tcb.command.xmit.maxcoll;
	    } /* if */
#ifdef DEBUG
            if (ecdebug & EC_DEBUG_PRINT) {
		printf(" T");
	    } /* if */
#endif
            ++ec->ec_if.if_opackets;		/* all is well */
            ec->request_tdr = 0;
#ifdef USELEDS
	    if (inledcontrol == 0) {
		ledcontrol(0, 0, LED_LANXMT);
	    } /* if */
#endif
	} else {
	    /*
	     * "Transmit" command was not OK.
	     */
	    fatal = FALSE;
	    if (tcb->tcb.command.xmit.late_collision) {
		++ec->xmit_status.late_collision;
		++ec->ec_if.if_collisions;
		fatal = TRUE;
	    } /* if */
	    if (tcb->tcb.command.xmit.no_carrier_sense) {
		++ec->xmit_status.no_carrier_sense;
		++ec->request_tdr;	/* more reason to check the cable */
		fatal = TRUE;
	    } /* if */
	    if (tcb->tcb.command.xmit.lost_clear_to_send) {
		++ec->xmit_status.lost_clear_to_send;
		++ec->request_tdr;	/* more reason to check the cable */
		fatal = TRUE;
	    } /* if */
	    if (tcb->tcb.command.xmit.dma_underrun) {
		++ec->xmit_status.dma_underrun;
		fatal = TRUE;
	    } /* if */
	    if (tcb->tcb.command.xmit.too_many_collisions) {
		++ec->xmit_status.too_many_collisions;
		ec->ec_if.if_collisions += 16;
		fatal = TRUE;
	    } /* if */
	    /*
	     * Record a number of funny cases where we do not have the
	     * OK bit set, but no fatal-error bits are set, either.
	     */
            if (!fatal) {
                ++ec->ufotransmit;
		KDB_BREAK("ec596txint: no error status on transmit");
	    } /* if */
            if (ec->retransmit++ == 0) {
		/* try to retransmit (once only!) */
                ++ec->nretransmit;
                tcb->tcb.command.cmd.el = TRUE;		/* this block only ! */
                tcb->tcb.command.cacheline.cb_stat = 0;	/* clear status */
                FLUSH(tcb, sizeof(*tcb));
		/* the interrupt handler will re-transmit the packet */
                return;
	    } /* if */
#ifdef DEBUG
            if (ecdebug & EC_DEBUG_PRINT) {
		printf(" X");
	    } /* if */
#endif
	    KDB_BREAK("ec596txint: retransmit failed");
            ++ec->ec_if.if_oerrors;
            if (ec->ec_if.if_flags & IFF_DEBUG) {
		printf("ec596txint: if_oerrors++\n");
	    } /* if */
#if defined(DEBUG) && defined(USELEDS)
	    if (inledcontrol == 0) {
		ledcontrol(0, 0, LED4);
	    } /* if */
#endif
	    /*
	     * Consecutive (or troublesome) errors cause us to ask the
	     * watchdog to perform time-domain reflectometry if the
	     * following value is incremented past a certain
	     * threshold.
	     */
	    ++ec->request_tdr;
	} /* if */
        ec->retransmit = 0;		/* no re-tries */

	/* do machine-independent completion processing: */
        iodone(tcb->tcb.ior);
        tcb->tcb.ior = (io_req_t)0;	/* record for "ec596reset()" */
        FLUSH(&tcb->tcb.ior, sizeof(tcb->tcb.ior));
        ec->tcb_head = ec->tcb_head->tcb.next;	/* we have a free slot */
    } /* while */
} /* ec596txint */


static void
ec596txsetup(int unit)
{
    tbd_t *tbd_head;
    char *data_addr;
    long data_size;
    int s;
    ecsi_t *ec = &ecsi[unit];
    io_req_t ior;
    tbd_t *tbd;
    tcb_t *tcb;
    int i;

    if (!ec->initialized) {
	return;
    } /* if */
    /*
     * Take the next packet from the send queue, set up to transmit
     * it, and hand it to the controller until we run out of
     * "Transmit" Command Blocks.
     */
    s = splimp();
    if ((ec->ec_if.if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)) {
	do {
	    IF_DEQUEUE(&ec->ec_if.if_snd, ior);
	    if (ior) {
#ifdef DEBUG
		if (ecdebug & EC_DEBUG_PRINTALL) {
		    printf("ec596txsetup: dropping packet\n");
		} /* if */
#endif
		ior->io_error = D_DEVICE_DOWN;
		ior->io_op |= IO_ERROR;
		ior->io_residual = ior->io_count;
		iodone(ior);
		IF_DROP(&ec->ec_if.if_snd);
	    } /* if */
	} while (ior);
	splx(s);
        return;
    } /* if */
    while (1) {
	if (ec->recover || ec->txsetup) {
	    /* chip stuck or another thread here */
	    splx(s);
	    return;
	} /* if */
	if (!ec->txactive && ec->q_op && ec->tcb_head != ec->tcb_tail) {
	    ec->txactive = TRUE;	/* start txmitting pending packets */
	    ec->txtimestamp = it_read();
#ifdef DEBUG
	    if (ecdebug & EC_DEBUG_PRINT) {
		printf(" C");
	    } /* if */
	    if (!ecto) {
		ec596custart(ec, (cb_t *)ec->tcb_head);
	    } /* if */
#else
	    ec596custart(ec, (cb_t *)ec->tcb_head);
#endif
	} /* if */
	tcb = ec->tcb_tail;
	if (tcb->tcb.next == ec->tcb_head) {
	    /* no TCBs: try to release some TCBs whose processing is done */
	    ec596txint(ec);
	    tcb = ec->tcb_tail;
	    if (tcb->tcb.next == ec->tcb_head) {
		splx(s);
		return;
	    } /* if */
	} /* if */
	IF_DEQUEUE(&ec->ec_if.if_snd, ior);
	if (!ior) {
	    /* no packet to set up */
	    splx(s);
	    return;
	} /* if */
	ec->txsetup = TRUE;	/* exclude other threads from this routine */
	splx(s);		/* allow interrupts */
#ifdef DEBUG
	if (ecdebug & EC_DEBUG_PRINTALL) {
	    printf("ec596txsetup: dequeued I/O req %.8x data@%.8x cnt=%d\n",
		   ior, ior->io_data, ior->io_count);
	} /* if */
#endif
	if ((unsigned)(ior->io_count - sizeof(struct ether_header)) > ETHERMTU)
	{
	    printf("%s%d: illegal output packet size\n", ecname, unit);
	    KDB_BREAK("ec596txsetup: bad output packet size");
	    s = splimp();
	    ior->io_error = D_INVALID_SIZE;
	    ior->io_op |= IO_ERROR;
	    ior->io_residual = ior->io_count;
	    iodone(ior);
	    IF_DROP(&ec->ec_if.if_snd);
	    ec->txsetup = FALSE;
	    continue;
	} /* if */
	tbd_head = tbd = &ec->tbd_list[tcb - ec->ecsc->tcb_list][0];
	if (LINKMISMATCH(tcb->tcb.tbd_ptr, tbd->pa)) {
	    panic("ec596txsetup: bad TBD pointer in TCB");
	} /* if */
#ifdef DEBUG
	if (ecdebug & EC_DEBUG_TX) {
	    printf("ec596txsetup: tcb=%.8x, tbd=(%.8x", tcb, tbd);
	} /* if */
#endif
	data_size = ior->io_count;
	data_addr = ior->io_data;
	/*
	 * The A Step chip has trouble with "Transmit" buffers that
	 * start on an odd address, so move 1 byte into the TBD in
	 * order to transmit it from there, then tranmit the remaining
	 * data from the next TBD.
	 */
	if ((long)data_addr & 1) {
	    *tbd->oddchar = *data_addr++;
	    --data_size;
	    tbd->count = 1;
	    tbd->eof = FALSE;
	    POINT(&tbd->buffer, tbd->oddchar);
	    ++tbd;
#ifdef DEBUG
	    if (ecdebug & EC_DEBUG_TX) {
		printf(",%.8x", tbd);
	    } /* if */
#endif
	} /* if */
	FLUSH(data_addr, data_size);
	tbd->count = data_size;
	tbd->eof = FALSE;
	POINT(&tbd->buffer, data_addr);
	++tbd;
	/*
	 * Pad out the packet to a minimum length.  Notice that the
	 * controller could do this automatically, *if* the Ethernet
	 * length/type field gets used as a length field *only*.  Of
	 * course, this isn't the case with the Internet protocol
	 * suite, so we have to do the padding manually.
	 */
	i = ETHERMIN + sizeof(struct ether_header) - ior->io_count;
	if (i > 0) {
	    static unsigned short pbuffer[(ETHERMIN+1)/sizeof(short)];

	    tbd->count = i;
	    tbd->eof = FALSE;
	    POINT(&tbd->buffer, pbuffer);
#ifdef DEBUG
	    if (ecdebug & EC_DEBUG_TX) {
		printf(",%.8x", tbd);
	    } /* if */
#endif
	    ++tbd;
	} /* if */
	(tbd-1)->eof = TRUE;			/* set end of list */
	FLUSH(tbd_head, (char*)tbd - (char*)tbd_head);
	tcb->tcb.command.cmd.el = TRUE;		/* last TCB on list */
	tcb->tcb.command.cacheline.cb_stat = 0;	/* clear status bits */
	/* keep request pointer so we can free it after xmit: */
	tcb->tcb.ior = ior;
	FLUSH(tcb, sizeof(*tcb));
#ifdef DEBUG
	if (ecdebug & EC_DEBUG_TX) {
	    printf(")\n");
	} /* if */
#endif

	if (ec->txactive) {
	    /* try appending to previous TCB */
	    tcb_t *prev = tcb->tcb.prev;
	    PURGE(&prev->tcb.command.cacheline.cb_cmd, DCACHE_LINE_SIZE);
	    prev->tcb.command.cmd.el = FALSE;
	    FLUSH(&prev->tcb.command.cacheline.cb_cmd, DCACHE_LINE_SIZE);
	} /* if */
#ifdef DEBUG
	if (ecdebug & EC_DEBUG_PRINT) {
	    printf(" T%d", ior->io_count);
	} /* if */
#endif
	s = splimp();
	ec->tcb_tail = tcb->tcb.next;
	ec->txsetup = FALSE;
    } /* while */
} /* ec596txsetup */


static int
ec596intr(int unit)
{
    ecsi_t *ec = &ecsi[unit];
    scb_t *scb = &ec->ecsc->scb;
    scb_int_t scb_int;

    if (!ecalive[unit] || !ec596ready(ec)) {
	/* SCB not ready, let watchdog restart chip */
	return 1;
    } /* if */
    PURGE(&scb->scb.status, sizeof(scb->scb.status));
    scb_int = scb->scb.status.stat;
    if (!scb_int) {
	/*
	 * The following weird situation may result from a race
	 * condition between events happening in the 82596 chip and
	 * the VSC interrupt handler for the ASP chip, to which the
	 * 82596 relays interrupt requests.  At this point, we have an
	 * interrupt but no apparent reason for it.  What happened is:
	 *
	 * 1. A previous 82596 event sent an interrupt request to ASP,
	 * and the VSC interrupt handler reset the ASP chip, allowing
	 * further interrupt requests.
	 *
	 * 2. After this, but before control arrived here last time,
	 * there was a brief time window during which a second 82596
	 * event in fact happened, and the previous invocation of this
	 * routine serviced BOTH events.
	 *
	 * 3. But there was no way to cancel the 2nd ASP-chip
	 * interrupt request, so we arrive here again, with a "dummy"
	 * interrupt that we have already serviced!
	 *
	 * The proper response is to do nothing but return a status
	 * indicating that the interrupt really has been serviced.
	 * Since there is only one 82596 using using particular ASP
	 * interrupt-request bit, we are assured there is no other
	 * cause of this odd situation.
	 */
	return 1;
    } /* if */
    if (*(short *)&scb->scb.status == I82596CORRUPTION) {
	/* chip bug, again! */
        ++ec->reason.scb_corrupted;
	KDB_BREAK("ec596intr: SCB status corrupted");
	/* reset the chip */
        ec->recover = TRUE;
        return 1;
    } /* if */

    /* acknowledge these interrupts: */
    scb->scb.command.cmd.ack = scb_int;
    FLUSH(&scb->scb.command, sizeof(scb->scb.command));
    CHANNEL_ATTN(unit);

    if ((ec->ec_if.if_flags & IFF_RUNNING) && !ec->recover) {
	/* handle interrupts: */
        if (scb_int & (SCB_CNA|SCB_CX)) {
	    /* CU completed command or has become inactive: */
            ec->txactive = FALSE;
            ec596txint(ec);
            if (ec->q_op) {
		/* re-start output: */
		ec596txsetup(unit);
            } else {
		thread_wakeup((void *)ec);
	    } /* if */
	} /* if */
        if (scb_int & SCB_FR) {
	    /* Frame Received: */
	    ec596frint(ec);
	} /* if */
        if ((scb_int & SCB_RNR) && !ec->initialized) {
	    ec596rxstart(ec);
	} /* if */
    } /* if */

    /* interrupt handled: */
    return 1;
} /* ec596intr */


/*
 * Watchdog to keep an eye on 82596 operation.
 */
static void
ec596watchdog(int unit)
{
    int s, warn;
    ecsi_t *ec = &ecsi[unit];

    timeout(ec596watchdog, unit, hz);
    s = splimp();
    if (!(ec->ec_if.if_flags & IFF_UP) || ec->ec_if.if_timer == 0
	|| --ec->ec_if.if_timer)
    {
        splx(s);
        return;
    } /* if */
    ++ec->dogwalks;	/* no. of seconds we are doing this */

    /*
     * Check completion bit in our special, "no operation" CB (see
     * second BIG FAT NOTICE above).
     */
    PURGE(&ec->ecsc->nop, sizeof(short));
    if (*(short *)&ec->ecsc->nop & 0x8000) {
        ++ec->reason.nop_completed;
	KDB_BREAK("ec596watchdog: \"no-operation\" CB completed");
        ec->recover = TRUE;
    } /* if */
    if (ec->tcb_head != ec->tcb_tail || ec->txactive) {
        ec596txint(ec);		/* check if any more xmits have finished */
        if (it_read() - ec->txtimestamp > ec->txtimelap) {
            ++ec->reason.xmit_timeout;
	    KDB_BREAK("ec596watchdog: transmit time-out");
            ec->recover = TRUE;
	} /* if */
    } /* if */

    /* wait until SCB is clear: */
    ec596ready(ec);

    /* re-initialize the chip: */
    if (ec->recover) {
	int i;

        ec->ec_if.if_flags &= ~IFF_RUNNING;
        ec596portaccess(unit, pmap_extract(kernel_pmap, ec->ecsc->dump),
			port_dump);
	/* ..to see with debugger: */
        PURGE(ec->ecsc->dump, sizeof(ec->ecsc->dump));
#ifdef DEBUG
        if (ecdebug & EC_DEBUG_PRINT) {
	    printf("ecrecover: ec%d was stuck, 82596 dump @ %.8x\n", unit,
		   ec->ecsc->dump);
	} /* if */
#endif
        i = 0;
	/* try to re-initialize twice if interface has not been shut down: */
        do {
            if (!(ec->ec_if.if_flags & IFF_UP) || ec596init(unit) == D_SUCCESS)
	    {
		warn = FALSE;
		break;
	    } /* if */
	    /* force self-test on second pass: */
            warn = ec->force_selftest = TRUE;
	} while (++i < 2);
        if (warn) {
	    /*
	     * By now, we have reset the chip and tried to execute the
	     * self-test, which failed.  We could "panic()" here, but
	     * instead let's warn the user that network access has
	     * failed and allow a graceful shut-down.
	     */
	    printf("%s%d: interface has malfunctioned and is unavailable\n",
		   ecname, unit);
	    printf("\tsystem shutdown and diagnosis is recommended\n");
	} /* if */
    } else {
	if (ec->request_tdr >= 4 || ec->last_warning) {

	    /*
	     * "Transmit" commands are failing or we have pending
	     * network trouble.  Perform TDR, but refrain from other
	     * checks.  If the user has stand- alone node, there is no
	     * need to try resetting the chip every few Seconds.  Note
	     * that the smaller the TDR-request threshold above, the
	     * sooner we will look for even the smallest disturbences
	     * on the cable.
	     */
	    ec596tdr(ec);
	    ec->request_tdr = 0;
	} /* if */

	/* synchronize the error counters: */
	if (ec->initialized) {
	    scb_t *scb = &ec->ecsc->scb;
#define ADDCNTR(new,old) \
	    { \
		if (new.low != old.low || new.high != old.high) { \
		    int add; \
		    add = (new.high<<16|new.low) - (old.high<<16|old.low); \
		    if ((unsigned) (add - 1) < 999 && ec->scberrs) { \
			ec->ec_if.if_ierrors+=add; \
			if (ec->ec_if.if_flags & IFF_DEBUG) { \
			    printf("ec596watchdog: %d if_ierrors\n", add); \
			} /* if */ \
		    } /* if */ \
		    old = new; \
		} /* if */ \
	    }
	    /*
	     * Record the number of frames lost due to lack of
	     * resources or over-run.  We let the watchdog do this
	     * instead of configuring the chip to save bad frames.
	     */
	    ADDCNTR(scb->scb.ovrnerrs, ec->ovrnerrs);
	    ADDCNTR(scb->scb.rscerrs, ec->rscerrs);
	    ec->scberrs = TRUE;
#undef ADDCNTR
	} /* if */
    } /* if */

    /* re-trigger in 1 second: */
    ec->ec_if.if_timer = 1;
    splx(s);
} /* ec596watchdog */


/*
 * This function "probes" or checks for the i82596 chip to see if it
 * is there.
 */
int
ec596probe(struct hp_device *dev)
{
    if ((unsigned) dev->hp_unit >= NEC) {
	return 0;
    } /* if */
    /*
     * Code for the on-board unit relies on knowing the data cache
     * line size, in order to position type definitions across cache
     * line boundaries.  With some extra, work it should be possible
     * to adapt the code to deal with varying line sizes and assume
     * the largest one for DCACHE_LINE_SIZE.
     */
    if (dcache_line_size != DCACHE_LINE_SIZE) {
        return 0;
    } /* if */
    return 1;
} /* ec596probe */


void
ec596attach(struct hp_device *dev)
{
    static char *my_name = "ec596attach";
    ecsi_t *ec = &ecsi[dev->hp_unit];
    int i;
    unsigned char *station_addr;

    /*
     * Tell machine-independent code to allocate at least as many
     * network message buffers as this driver is prepared to handle.
     */
    if (net_queue_free_min < NBUF) {
	net_queue_free_min = NBUF;
    } /* if */

    /* initialize "Transmit" Command Block template: */

    cb_template.transmit.tcb.command.cmd.el = TRUE; /* only 1 block on list */
    cb_template.transmit.tcb.command.cmd.sf = TRUE; /* use "flexible" mode */
    cb_template.transmit.tcb.command.cmd.cuc = cb_cuc_transmit;
    SET_EC596_NIL(&cb_template.transmit.tcb.command.cmd.link);

    station_addr = (unsigned char*) LAN_PROM_ADDR + OFFSET_STATION_ADDR;
    for (i = 0; i < sizeof(ec->ec_addr); ++i) {
	ec->ec_addr[i] = station_addr[i];
    } /* for */

    if (INVALIDADDR(ec->ec_addr)) {
        printf("%s: %s%d invalid MAC address\n",
	       my_name, ecname, dev->hp_unit);
        dev->hp_alive = '\000';
        return;
    } /* if */
    ec->ec_if.if_unit = dev->hp_unit;
    ecaddr[dev->hp_unit] = dev->hp_addr;
    ec->ec_if.if_flags = IFF_BROADCAST;
    ec->ec_if.if_timer = 0;
    ec->ec_if.if_mtu = ETHERMTU;
    ec->ec_if.if_header_size = sizeof(struct ether_header);
    ec->ec_if.if_header_format = HDR_ETHERNET;
    ec->ec_if.if_address_size = sizeof(eaddr_t);
 /* ec->ec_if.if_alloc_size = ???; */
    ec->ec_if.if_address = ec->ec_addr;
    if_init_queues(&ec->ec_if);

    /* set up pointer to a page with the lower 16 bits set to 0x4000: */

#ifdef DEBUG
    if (ecdebug & EC_DEBUG_CHECK) {
	printf("ec596attach: got memory @ va=%x, pa=%x\n", 
	       (long) ecsc_area, (long) pmap_extract(kernel_map, ecsc_area));
    } /* if */
#endif /* DEBUG */
    ec->ecsc = &ecsc_area[dev->hp_unit];
    lock_init(&ec->init_lock, TRUE);
    lock_init(&ec->q_lock_init, TRUE);	/* get initial value for queue lock */

    /* indicate that no receive buffers have been allocated: */
    i = 0;
    do {
	ec->ecsc->rbd_list[i].rbd.kmsg = (ipc_kmsg_t)0;
    } while (++i < NRBDBUF);
    ec596reset(dev->hp_unit);		/* reset the controller */
    aspitab(INT_LAN, SPLLAN, ec596intr, dev->hp_unit, 0);
    printf("%s%d: %s, address [%s]\n", ecname, dev->hp_unit,
	   DISP_LANJMPR(ASP_CTRL->iostat.lan_jmpr),
	   ether_sprintf(ec->ec_addr));
    timeout(ec596watchdog, dev->hp_unit, hz);	/* start "watchdog" timer */
    ecalive[dev->hp_unit] = TRUE;
} /* ec596attach */


#if defined(MACH_KDB) || defined(KGDB)
/*
 * Kernel-debugger "$HE" command runs this routine.
 */
void
ec596trace(void)
{
    ecsi_t *ec = &ecsi[0];
    scb_t *scb;
    int i;

    PURGE(ec->ecsc, sizeof(ecsc_t));

    scb = &ec->ecsc->scb;
    printf("scb: stat=%x,cus=%d,rus=%d,throad=%d\n",
	   scb->scb.status.stat, scb->scb.status.cus,
	   scb->scb.status.rus, scb->scb.status.throad);
#define UNSWAP(x)	((x).low & 0xfffl | ((x).high << 16))
    printf("scb: crcerrs=%d, alnerrs=%d, rscerrs=%d, ovrnerrs=%d\n",
	   UNSWAP(scb->scb.crcerrs), UNSWAP(scb->scb.alnerrs),
	   UNSWAP(scb->scb.rscerrs), UNSWAP(scb->scb.ovrnerrs));
    printf("scb: rcvcdt=%d, shrtfrm=%d\n",
	   UNSWAP(scb->scb.rcvcdt), UNSWAP(scb->scb.shrtfrm));
#undef UNSWAP

    PURGE(ec->tbd_list, sizeof(ec->tbd_list));
    printf(
"ecsi@%.8x ecsc@%.8x rfd_list@%.8x(%d) rbd_list@%.8x(%d) tcb_list@%.8X(%d)\n",
	   ecsi, ec->ecsc, ec->ecsc->rfd_list, NRFDBUF, ec->ecsc->rbd_list,
	   NRBDBUF, ec->ecsc->tcb_list, NTCBBUF);
    printf("rfd=%.8x/%.8x rbd=%.8x/%.8x tcb=%.8x/%.8x\n",
	   ec->rfd_head, ec->rfd_tail, ec->rbd_head, ec->rbd_tail,
	   ec->tcb_head, ec->tcb_tail);
    printf("txactive=%d q_op=%d txsetup=%d\n",
	   ec->txactive, ec->q_op, ec->txsetup);
    if (ec->ru_noresources) {
	printf("ru_noresources=%d\n", ec->ru_noresources);
    } /* if */
    if (ec->nretransmit) {
	printf("nretransmit=%d\n", ec->nretransmit);
    } /* if */
    printf(
      "recover: %d %d %d %d %d %d %d (scbn/r, tcb, scb, nop, xmit, rfd, ru)\n",
	   ec->reason.scb_not_ready, ec->reason.scb_corrupted,
	   ec->reason.tcb_corrupted, ec->reason.nop_completed,
	   ec->reason.xmit_timeout, ec->reason.rfd_not_completed,
	   ec->reason.ru_not_ready);
    printf("xmit: %d %d %d %d %d %d (lcol, nocr, ncts, dmau, defr, tcol)\n",
	   ec->xmit_status.late_collision, ec->xmit_status.no_carrier_sense,
	   ec->xmit_status.lost_clear_to_send, ec->xmit_status.dma_underrun,
	   ec->xmit_status.deferred, ec->xmit_status.too_many_collisions);
    printf("iflags: %.4x, ifq: %d %d %d (oerrors, rcvdrops, ierrors)\n",
	   ec->ec_if.if_flags, ec->ec_if.if_oerrors, ec->ec_if.if_rcvdrops,
	   ec->ec_if.if_ierrors);
    /* forward link checks */
    i = 0;
    do {
	if (LINKMISMATCH(ec->ecsc->rfd_list[i].rfd.command.save.link,
			 ec->ecsc->rfd_list[i].rfd.command.cmd.link))
	{
	    printf("bad link address in RFD @ %.8x\n", &ec->ecsc->rfd_list[i]);
	} /* if */
    } while (++i < NRFDBUF);
    i = 0;
    do {
	if (LINKMISMATCH(ec->ecsc->rbd_list[i].save.link,
			 ec->ecsc->rbd_list[i].rbd.link))
	{
	    printf("bad link address in RBD @ %.8x\n",
		   &ec->ecsc->rbd_list[i]);
	} /* if */
    } while (++i < NRBDBUF);
    i = 0;
    do {
	if (LINKMISMATCH(ec->ecsc->tcb_list[i].tcb.command.save.link,
			 ec->ecsc->tcb_list[i].tcb.command.cmd.link))
	{
	    printf("bad link address in TCB @ %.8x\n",
		   &ec->ecsc->tcb_list[i]);
	} /* if */
    } while (++i < NTCBBUF);
} /* ec596trace */
#endif /* MACH_KDB */


void
ec596_special_page(vm_offset_t start, vm_offset_t end, vm_offset_t *va)
{
    vm_offset_t pa;

    /*
     * In order to work around the INTEL-chip data-corruption bug (see
     * the BIG FAT NOTICEs above) we need a special memory page, such
     * that the least-significant 16 bits of the physical page
     * number are 0x4000, to hold an "ecsc_t" structure for each
     * Ethernet controller.
     */
#define SPECIAL_SIZE (round_page(NEC*sizeof(ecsc_t)))
    
    /* setup pointer to a page with the lower 16 bits set to 0x4000: */
    pa = start + 0xffff & ~0xffff | I82596CORRUPTION;
    if (pa < start) {
	panic("ec_special_page: ecsc address too low");
    } /* if */
    if (pa + SPECIAL_SIZE >= end) {
	panic("ec_special_page: ecsc address too high");
    } /* if */
    /*
     * See the second BIG FAT NOTICE above: Since the forward link
     * addresses of the "Transmit" Command Blocks can be truncated
     * (least-significant 16 bits sets to 0x4000), we keep all TCBs in
     * memory after the special "no operation" TCB in the ecsc_t
     * structure, with the most-significant 16 bits unchanged.
     */
    if ((pa & ~0xffff) != (pa + SPECIAL_SIZE & ~0xffff)) {
	panic("ec_special_page: ecsc area spans 64KB boundary");
    } /* if */
    pmap_define_hole(pa, pa + SPECIAL_SIZE);
    ecsc_area = (ecsc_t*) *va;
    *va = pmap_map(ecsc_area, pa, pa + SPECIAL_SIZE,
		   VM_PROT_READ|VM_PROT_WRITE);
#undef SPECIAL_SIZE
}
# endif	/* NEC > 0 */

			/*** end of ec596.c ***/
